/**
 * Version.java
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.app;

import org.apache.oro.text.regex.MalformedPatternException;
import org.apache.oro.text.regex.MatchResult;
import org.apache.oro.text.regex.Pattern;
import org.apache.oro.text.regex.PatternCompiler;
import org.apache.oro.text.regex.PatternMatcher;
import org.apache.oro.text.regex.Perl5Compiler;
import org.apache.oro.text.regex.Perl5Matcher;

import com.eXtropia.config.ConfigException;

/**
 * Version.java
 * Version provides an abstract class which all others extend to provide
 * version information.
 * 
 * @author <a href="st_lim@eXtropia.com">Lim Swee Tat</a>
 * @created Sun Aug 18 02:30:53 2002
 * @version $Id: ExtropiaVersion.java,v 1.6 2002/11/16 08:45:08 rayfd Exp $
 */
public abstract class ExtropiaVersion {
    private static String TAGSTRING = "Name: (([^0-9]*)-([0-9_]*))";

    public final static int RELEASE = 0;
    public final static int STAGING = 1;
    public final static int TESTING = 2;

    public abstract String getCvsName();

    private final static int TAGNAME_PART = 1;
    private final static int APPNAME_PART = 2;
    private final static int VERSION_PART = 3;

    private Pattern _tagPattern = null;

    public ExtropiaVersion() throws ConfigException {
        // Create Perl5Compiler and Perl5Matcher instances.
        PatternCompiler compiler = new Perl5Compiler();

        // Attempt to compile the pattern.  If the pattern is not valid,
        // report the error and exit.
        try {
            _tagPattern = compiler.compile(TAGSTRING);
        } catch (MalformedPatternException e) {
            throw new ConfigException(e.getMessage());
        }
    }

    /**
     * Method getVersion.
     * @return String
     */
    public String getTagVersion() {
        return matchPart(TAGNAME_PART);
    }

    /**
     * Method getAppName.
     * @return String
     */
    public String getAppName() {
        return matchPart(APPNAME_PART);
    }

    /**
     * Method getVersion.
     * @return String
     */
    public String getVersion() {
        return matchPart(VERSION_PART);
    }

    /**
     * Method matchPart.
     * @return String
     */
    private String matchPart(int part) {
        return matchPart(getCvsName(), part);
    }

    private String matchPart(String s, int part) {
        PatternMatcher matcher = new Perl5Matcher();

        if (matcher.contains(s, _tagPattern)) {
            MatchResult matchResult = matcher.getMatch();
            return matchResult.group(part);
        }
        return "";
    }

    /**
     * Method isRelease.
     * @return boolean
     */
    public boolean isRelease() {
        return (!(isStaging() || isTesting()));
    }

    /**
     * Method isStaging.
     * @return boolean
     */
    public boolean isStaging() {
        if (getCvsName().indexOf("STAGING") > -1) {
            return true;
        }
        return false;
    }

    /**
     * Method isTesting.
     * @return boolean
     */
    public boolean isTesting() {
        if (getCvsName().indexOf("TESTING") > -1) {
            return true;
        }
        return false;
    }

    public String showVersion() {
        StringBuffer result = new StringBuffer();
        result.append("ADT CVS Tag: " + getAdtCvsName() + "\n");
        result.append("ADT Version: " + getAdtVersion() + "\n\n");
        result.append("CVS Tag: " + getCvsName() + "\n");
        result.append("Application: " + getAppName() + "\n");
        result.append("Version: " + getVersion() + "\n");
        result.append("This version is a ");
        if (isTesting()) {
            result.append("testing");
        } else if (isStaging()) {
            result.append("staging");
        } else {
            result.append("production");
        }
        result.append(" version\n");
        return result.toString();
    }

    public String showHTMLVersion() {
        StringBuffer result = new StringBuffer();
        result.append("<HTML>\n<HEAD>\n");
        result.append("<TITLE>" + getAppName() + " Version</TITLE>\n</HEAD>\n<BODY>\n");
        result.append("ADT CVS Tag: " + getAdtCvsName() + "<BR>\n");
        result.append("ADT Version: " + getAdtVersion() + "<BR><HR>\n");
        result.append("CVS Tag: " + getCvsName() + "<BR>\n");
        result.append("Application: " + getAppName() + "<BR>\n");
        result.append("Version: " + getVersion() + "<BR>\n");
        result.append("This version is a ");
        if (isTesting()) {
            result.append("testing");
        } else if (isStaging()) {
            result.append("staging");
        } else {
            result.append("production");
        }
        result.append(" version.<BR>\n");
        result.append("</BODY>\n</HTML>");
        return result.toString();
    }

    public String toString() {
        return ("Version: " + getCvsName() + "\nADT Version: " + getAdtCvsName());
    }

    public String getAdtCvsName() {
        return "$Name: toolkit-1_20 $";
    }

    public String getAdtVersion() {
        return matchPart(getAdtCvsName(), VERSION_PART);
    }
}
