/* 
 * Action.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.app.servlet;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import javax.servlet.ServletRequest;

import com.eXtropia.app.UserGesture;
import com.eXtropia.app.UserGestureResponse;
import com.eXtropia.auth.User;
import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.datahandler.DataHandlerManager;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.util.ExtropiaException;

/** 
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 1.0
 */
public abstract class Action implements Configurable {
    protected DataHandlerManager _dataHandlerManager;
    protected Doctor _doctor;
    protected boolean _requiresAuth = true;
    protected String[] _authGroups;
    private HashSet _listeners = new HashSet();
    protected String _view;
    protected String _targetView;

    public static final String CFG_BASE = "Action";
    public static final String CFG_REQUIRES_AUTH = "RequiresAuth";
    public static final String CFG_AUTH_GROUPS = "AuthGroups";
    public static final String CFG_AUTH_GROUP = "Group";
    public static final String CFG_VIEW = "View";
    public static final String CFG_TARGET_VIEW = "TargetView";

    public Action() {

    }

    public void addListener(ActionListener listener) {
        synchronized (_listeners) {
            _listeners.add(listener);
        }
    }

    public void configure(Config config) throws ConfigException {
        if (!config.getName().equals(CFG_BASE))
            throw new ConfigException("Incorrect config passed to " + CFG_BASE + ": " + config.getName());
        if (config.hasChild(CFG_REQUIRES_AUTH))
            _requiresAuth = config.getChild(CFG_REQUIRES_AUTH).getBooleanValue();
        if (config.hasChild(CFG_AUTH_GROUPS)) {
            if (!_requiresAuth)
                throw new ConfigException(CFG_AUTH_GROUPS + " specified for an action that does not require authentication");
            List list = new ArrayList();
            Config[] groups = config.getChild(CFG_AUTH_GROUPS).getChildren();
            for (int i = 0; i < groups.length; i++) {
                if (groups[i].getName().equals(CFG_AUTH_GROUP))
                    list.add(groups[i].getValue());
            }
            _authGroups = (String[]) list.toArray(new String[list.size()]);
        }
        Config[] children = config.getChildren();
        int listenerCount = 0;
        ActionListener listener;
        for (int i = 0; i < children.length; i++) {
            if (children[i].getName().equals(ActionListener.CFG_BASE)) {
                try {
                    listener = (ActionListener) children[i].createAndConfigureInstance();
                    addListener(listener);
                    listenerCount++;
                } catch (ClassCastException cce) {
                    throw new ConfigException("Config type " + config.getType() + " is not a " + ActionListener.class.getName());
                }
            }
        }
        if (config.hasChild(CFG_VIEW))
            _view = config.getChild(CFG_VIEW).getValue();
        if (config.hasChild(CFG_TARGET_VIEW))
            _targetView = config.getChild(CFG_TARGET_VIEW).getValue();
    }

    public abstract UserGestureResponse execute(UserGesture gesture, ServletRequest request) throws ExtropiaException;

    public DataHandlerManager getDataHandlerManager() {
        return _dataHandlerManager;
    }

    public boolean getRequiresAuthentication() {
        return _requiresAuth;
    }

    public void setDataHandlerManager(DataHandlerManager dataHandlerManager) {
        _dataHandlerManager = dataHandlerManager;
    }

    public void setDoctor(Doctor doctor) {
        _doctor = doctor;
    }

    public void setRequiresAuthentication(boolean requiresAuth) {
        _requiresAuth = requiresAuth;
    }

    public boolean canExecute(User user) {
        if (_requiresAuth) {
            if (user == null)
                return true; //assume they can: lazy authentication
            else if (_authGroups != null) {
                for (int i = 0; i < _authGroups.length; i++) {
                    if (user.isMemberOf(_authGroups[i]))
                        return true;
                }
                return false;
            } else //just being logged in is enough
                return true;
        } else
            return true;
    }

    public void removeListener(ActionListener listener) {
        synchronized (_listeners) {
            _listeners.remove(listener);
        }
    }

    protected void notifyListeners(ActionEvent actionEvent) {
        HashSet listeners;
        synchronized (_listeners) {
            listeners = (HashSet) _listeners.clone();
        }
        for (Iterator i = listeners.iterator(); i.hasNext();) {
            ((ActionListener) i.next()).onActionEvent(actionEvent);
        }
    }

    /**
     * @return the view that should be returned to upon datahandler exception or null if not relevant
     * If this action requires that upon successful execution, a certain view must be displayed, this
     * method will return that view. Otherwise, the view to be displayed mmay be determined by the gesture
     */
    public String getTargetView() {
        return _targetView;
    }

    /**
     * @return the view that this action assumes that data has been sent from
     */
    public String getView() {
        return _view;
    }
}