/*
 * ActionManager.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.app.servlet;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.ServletContext;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.eXtropia.app.UserGesture;
import com.eXtropia.app.UserGestureResponse;
import com.eXtropia.auth.AuthException;
import com.eXtropia.auth.AuthManager;
import com.eXtropia.auth.User;
import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.datahandler.DataHandlerException;
import com.eXtropia.datahandler.DataHandlerManager;
import com.eXtropia.datahandler.Rule;
import com.eXtropia.datahandler.ServletRequestDataHandlerManager;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.util.ExtropiaException;
import com.eXtropia.view.View;
import com.eXtropia.view.ViewManager;

/**
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 1.0
 */
public abstract class ActionManager implements Configurable {
    public static final String CFG_BASE = "ActionManager";
    public static final String CFG_DEFAULT_VIEW = "DefaultView";
    public static final String CFG_DEFAULT_ACTION = "DefaultAction";
    public static final String CFG_VIEW_ACTIONS = "ViewActions";
    public static final String CFG_ACTIONS = "Actions";
    public static final String CFG_ACTION_VIEW = "View";
    public static final String CFG_RULES = "Rules";
    public static final String CFG_RULE_ORDER = "RuleOrder";
    public static final String CFG_RULE_ORDER_NAME = "Name";
    public static final String CFG_USER_GESTURE_FACTORY = "UserGestureFactory";
    public static final String CFG_LOG_ACTION_EXCEPTIONS = "LogActionExceptions";

    public static final String WIDGET_RULE_NAME = "widget";

    public static final Integer DEFAULT_START_OF_DAY = new Integer(8);

    private Map _actions;
    private Map _viewActions;
    private String _defaultView = Views.NO_SUCH_VIEW;
    private String _defaultAction = null; // was = Views.NO_SUCH_ACTION; but we want to skip if not defined...
    protected Doctor _doctor;
    protected String _appInstance;
    protected String _appName;
    protected Map _rules;
    protected UserGestureFactory _userGestureFactory;
    protected ActionAccessControlList _acl;
    protected boolean _logActionExceptions = false;

    protected ViewManager _viewManager;
    protected AuthManager _authManager;
    protected ServletContext _servletContext;

    public abstract void configureConcrete(Config config) throws ConfigException;

    public ActionManager(String appName, String appInstance, AuthManager authManager, ViewManager viewManager) {
        _appInstance = appInstance;
        _appName = appName;
        _authManager = authManager;
        _viewManager = viewManager;
        _doctor = Diagnostics.getDoctor(new String[] { appInstance, appName, "ActionManager" });
        _acl = new ActionAccessControlList(this);
    }

    public final void configure(Config config) throws ConfigException {
        if (!config.getName().equals(CFG_BASE))
            throw new ConfigException("Invalid Config passed to " + CFG_BASE + ": " + config.getName());

        if (config.hasChild(CFG_USER_GESTURE_FACTORY)) {
            try {
                _userGestureFactory = (UserGestureFactory) config.getChild(CFG_USER_GESTURE_FACTORY).createAndConfigureInstance();
            } catch (ClassCastException cce) {
                throw new ConfigException("Config " + CFG_USER_GESTURE_FACTORY + " is not of type " + CFG_USER_GESTURE_FACTORY);
            }
        } else
            throw new ConfigException("No " + CFG_USER_GESTURE_FACTORY + " specified in the " + CFG_BASE + " Config");

        configureConcrete(config);
        configureActions(config);

        //get the default action
        if (config.hasChild(CFG_DEFAULT_VIEW))
            _defaultView = config.getChild(CFG_DEFAULT_VIEW).getValue();

        if (config.hasChild(CFG_DEFAULT_ACTION))
            _defaultAction = config.getChild(CFG_DEFAULT_ACTION).getValue();

        if (config.hasChild(CFG_LOG_ACTION_EXCEPTIONS))
            _logActionExceptions = config.getChild(CFG_LOG_ACTION_EXCEPTIONS).getBooleanValue();
    }

    protected void configureActions(Config config) throws ConfigException {
        Action action;
        if (config.hasChild(CFG_RULES))
            _rules = createRules(config.getChild(CFG_RULES));
        Config[] actionConfigs;
        _actions = new HashMap();
        if (config.hasChild(CFG_ACTIONS)) {
            actionConfigs = config.getChild(CFG_ACTIONS).getChildren();
            for (int i = 0; i < actionConfigs.length; i++) {
                if (actionConfigs[i].getName().equals(Action.CFG_BASE)) {
                    _doctor.log(new LogEvent(this, LogEvent.INFO, "Creating action: " + actionConfigs[i].getValue()));
                    action = createAction(actionConfigs[i]);
                    _actions.put(actionConfigs[i].getValue(), action);
                    _doctor.log(new LogEvent(this, LogEvent.INFO, CFG_ACTIONS + " " + actionConfigs[i].getValue() + " => " + action.getClass().getName()));
                }
            }
        } else
            throw new ConfigException("No " + CFG_ACTIONS + " specified in the " + CFG_BASE + " Config");

        _viewActions = new HashMap();
        if (config.hasChild(CFG_VIEW_ACTIONS)) {
            actionConfigs = config.getChild(CFG_VIEW_ACTIONS).getChildren();
            for (int i = 0; i < actionConfigs.length; i++) {
                if (actionConfigs[i].getName().equals(Action.CFG_BASE)) {
                    _doctor.log(new LogEvent(this, LogEvent.INFO, "Creating action: " + actionConfigs[i].getValue()));
                    action = createAction(actionConfigs[i]);
                    _viewActions.put(actionConfigs[i].getValue(), action);
                    _doctor.log(new LogEvent(this, LogEvent.INFO, CFG_VIEW_ACTIONS + " " + actionConfigs[i].getValue() + " => " + action.getClass().getName()));
                }
            }
        } else
            throw new ConfigException("No " + CFG_VIEW_ACTIONS + " specified in the " + CFG_BASE + " Config");
    }

    public Action createAction(Config config) throws ConfigException {
        String[] diagFilters;
        Rule[] logicRules;

        Rule widgetRule = null, tmpRule;
        View view = null;
        String[] widgetOrder;
        DataHandlerManager dataHandlerManager;
        Action action = null;

        try {
            action = (Action) config.createAndConfigureInstance();
        } catch (ClassCastException cce) {
            throw new ConfigException("Action " + config.getValue() + " is not of the type " + Action.class.getName());
        }

        diagFilters = new String[] { _appInstance, _appName, "Action", config.getValue()};
        action.setDoctor(Diagnostics.getDoctor(diagFilters));

        //now determine the datahandler manager for the action
        logicRules = getRuleOrder(config);
        if (action.getView() != null) {
            view = _viewManager.getView(action.getView());
            if (view != null) {
                widgetOrder = view.getWidgetOrder();
                widgetRule = new Rule(WIDGET_RULE_NAME);
                for (int j = 0; j < widgetOrder.length; j++) {
                    tmpRule = view.getWidget(widgetOrder[j]).getRule();
                    if (tmpRule != null) {
                        widgetRule.add(tmpRule);
                    }
                }
            }
        } else
            widgetRule = null;
        Rule overallRule = new Rule("rule");
        if (widgetRule != null) {
            overallRule.add(widgetRule);
        }
        if (logicRules != null) {
            Rule logicRule = new Rule("logic");
            for (int j = 0; j < logicRules.length; j++)
                logicRule.add(logicRules[j]);
            overallRule.add(logicRule);
            overallRule.setMaxErrors(1); //only report logic errors if widget rule executes OK
        }
        dataHandlerManager = new ServletRequestDataHandlerManager();
        dataHandlerManager.setRule(overallRule);
        action.setDataHandlerManager(dataHandlerManager);
        return action;
    }

    protected Map createRules(Config config) throws ConfigException {
        Map rules = new HashMap();
        Config[] children = config.getChildren();
        Rule rule;
        for (int i = 0; i < children.length; i++) {
            if (children[i].getName().equals(Rule.CFG_BASE)) {
                try {
                    rule = new Rule();
                    rule.configure(children[i]);
                    rules.put(rule.getName(), rule);
                } catch (ClassCastException e) {
                    throw new ConfigException(config.getType() + " is not of type Rule");
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
        return rules;
    }

    protected Rule[] getRuleOrder(Config config) throws ConfigException {
        if (config.hasChild(CFG_RULE_ORDER)) {
            if (_rules == null)
                throw new ConfigException("You have specified a RuleOrder but no Rules in ViewManager");
            Config[] children = config.getChild(CFG_RULE_ORDER).getChildren();
            List rules = new ArrayList();
            for (int i = 0; i < children.length; i++) {
                if (_rules.containsKey(children[i].getValue())) {
                    rules.add(_rules.get(children[i].getValue()));
                } else
                    throw new ConfigException("No Rule defined by " + CFG_RULE_ORDER_NAME + ": " + children[i].getValue());
            }
            return (Rule[]) rules.toArray(new Rule[rules.size()]);
        } else
            return null;
    }

    /**
     * Executes the user specific action specified in the <CODE>UserGesture</CODE>
     * @return the view that should be displayed to the user based on the action's results
     */
    public UserGestureResponse execute(HttpServletRequest request, HttpServletResponse response) {
        if (_actions == null)
            throw new IllegalStateException("UserActionManager has not been configured");

        response.setContentType("text/html");

        //need to get the action and the view action out first and determine if they
        //are authenticatable actions

        Action action;
        UserGestureResponse gestureResponse;
        DataHandlerException dataHandlerException = null;

        UserGesture gesture = _userGestureFactory.create(request);

        // If there is a default action and no action nor view is set,
        // then we set the action to the default action
        //
        if (_defaultAction != null && gesture.getAction() == null && gesture.getView() == null)
            gesture.setAction(_defaultAction);

        if (gesture.getAction() != null) {
            //System.out.println("Executing action: " + gesture.getAction());
            //then the user is actually performing some real function
            action = (Action) _actions.get(gesture.getAction());
            if (action == null) {
                gestureResponse = new UserGestureResponse(Views.NO_SUCH_ACTION);
                showView(gesture, gestureResponse, request, response);
                return gestureResponse;
            }
            if (action.getRequiresAuthentication()) {
                User user = getUser(request, response);
                if (user == null)
                    return null;
                else if (!action.canExecute(user)) {
                    gestureResponse = new UserGestureResponse();
                    gestureResponse.setView(Views.UNAUTHORIZED);
                    showView(gesture, gestureResponse, request, response);
                    return gestureResponse;
                }
                gesture.setUser(user);
            } else if (_authManager != null)
                gesture.setUser(_authManager.getUserIfLoggedIn(request));

            DataHandlerManager dhm = action.getDataHandlerManager();
            try {
                if (request.getContentType() != null && request.getContentType().indexOf(UserGesture.MULTIPART_FORM_DATA) >= 0)
                    action.setDataHandlerManager(null);
                else if (dhm != null) {
                    Map validatedData = dhm.execute(request);
                    gesture.setData(validatedData);
                }
                if (_doctor.doTrace(TraceEvent.VERBOSE))
                    _doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Executing action: " + gesture.getAction()));
                gestureResponse = action.execute(gesture, request);
                if (gestureResponse != null) {
                    showView(gesture, gestureResponse, request, response);
                    return gestureResponse;
                } else if (gesture.getTargetView() != null) {
                    //System.out.println("Using specified target view");
                    gesture.setView(gesture.getTargetView());
                    gesture.setTargetView(null);
                } else {
                    //System.out.println("No specified target view");
                }
            } catch (DataHandlerException dhe) {
                if (_doctor.doTrace(TraceEvent.VERBOSE))
                    _doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Failed to validate data for action: " + gesture.getAction()));
                dataHandlerException = dhe;
                if (action.getView() != null)
                    gesture.setView(action.getView());
            } catch (ExtropiaException ee) {
                if (_logActionExceptions)
                    _doctor.log(new LogEvent(this, LogEvent.CRIT, "Caught Exception executing action " + gesture.getAction() + ": " + ee.getMessage()));
                gestureResponse = new UserGestureResponse();
                gestureResponse.setError(ee);
                gestureResponse.setView(Views.APPLICATION_ERROR);
                showView(gesture, gestureResponse, request, response);
                return gestureResponse;
            }
        }

        //get data from the underlying datasource that is required for the view
        if (gesture.getView() == null)
            gesture.setView(_defaultView);

        action = (Action) _viewActions.get(gesture.getView());

        if (action == null) {
            gestureResponse = new UserGestureResponse(Views.NO_SUCH_VIEW);
            showView(gesture, gestureResponse, request, response);
            return gestureResponse;
        } else {
            //System.out.println("Executing view action: " + gesture.getView());
            if (_doctor.doTrace(TraceEvent.DEVELOPER))
                _doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Executing view action: " + gesture.getView()));
            if (action.getRequiresAuthentication()) {
                User user = getUser(request, response);
                if (user == null) {
                    return null;
                } else if (!action.canExecute(user)) {
                    gestureResponse = new UserGestureResponse();
                    gestureResponse.setView(Views.UNAUTHORIZED);
                    showView(gesture, gestureResponse, request, response);
                    return gestureResponse;
                }
                gesture.setUser(user);
            } else if (_authManager != null)
                gesture.setUser(_authManager.getUserIfLoggedIn(request));

            DataHandlerManager dhm = action.getDataHandlerManager();

            if (dhm != null) {
                try {
                    Map validatedData = dhm.execute(request);
                    gesture.setData(validatedData);
                } catch (DataHandlerException dhe) {
                    dataHandlerException = dhe;
                }
            }

            try {
                gestureResponse = action.execute(gesture, request);
                if (gestureResponse.getView() != null) {
                    if (dataHandlerException != null)
                        gestureResponse.setDataHandlerException(dataHandlerException);
                    showView(gesture, gestureResponse, request, response);
                } else
                    _doctor.trace(new TraceEvent(this, TraceEvent.CRIT, "No destination view specified by View Action: " + gesture.getView()));
                return gestureResponse;
            } catch (ExtropiaException ee) {
                if (_logActionExceptions)
                    _doctor.log(new LogEvent(this, LogEvent.CRIT, "Caught Exception executing action " + gesture.getAction() + ": " + ee.getMessage()));
                gestureResponse = new UserGestureResponse();
                gestureResponse.setError(ee);
                gestureResponse.setView(Views.APPLICATION_ERROR);
                showView(gesture, gestureResponse, request, response);
                return gestureResponse;
            }
        }
    }

    protected boolean showView(UserGesture gesture, UserGestureResponse gestureResponse, ServletRequest request, ServletResponse response) {
        if (gestureResponse.getView() == null)
            return false;
        request.setAttribute(UserGesture.REQUEST_PARAM_NAME, gesture);
        request.setAttribute(UserGestureResponse.REQUEST_PARAM_NAME, gestureResponse);
        request.setAttribute(ActionAccessControlList.REQUEST_ATTRIBUTE_NAME, _acl);
        try {
            _viewManager.showView(gestureResponse.getView(), request, response);
            return true;
        } catch (Exception e) {
            e.printStackTrace();
            if (_doctor.doTrace(TraceEvent.CRIT)) {
                _doctor.trace(new TraceEvent(this, TraceEvent.CRIT, "Failed to show view (" + gestureResponse.getView() + "): " + e.getMessage(), e));
            }
            return false;
        }
    }

    protected User getUser(HttpServletRequest request, HttpServletResponse response) {
        User user = null;
        try {
            if (_doctor.doTrace(TraceEvent.DEVELOPER))
                _doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Authenticating User"));
            user = _authManager.authenticate(request, response);
            if (_doctor.doTrace(TraceEvent.DEVELOPER))
                _doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Authenticated the user"));
            if (response.isCommitted()) {
                //		if(_doctor.doTrace(TraceEvent.DEVELOPER))
                //				_doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "After authenticate, response is committed"));
                return null;
            } else if (user == null) {
                if (_doctor.doTrace(TraceEvent.DEVELOPER))
                    _doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "After authenticate, no user found"));
                throw new AuthException();
            } else {
                if (_doctor.doTrace(TraceEvent.DEVELOPER))
                    _doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Got the user: " + user.getId()));
            }

        } catch (AuthException ae) {
            showView(new UserGesture(), new UserGestureResponse(Views.UNAUTHORIZED), request, response);
        } catch (ExtropiaException ee) {
            UserGestureResponse gestureResponse = new UserGestureResponse();
            gestureResponse.setError(ee);
            gestureResponse.setView(Views.APPLICATION_ERROR);
            showView(new UserGesture(), gestureResponse, request, response);
        }
        return user;
    }

    Action getAction(String action) {
        return (Action) _actions.get(action);
    }

    Action getViewAction(String action) {
        return (Action) _viewActions.get(action);
    }

    public void setServletContext(ServletContext sc) {
        _servletContext = sc;
    }
}