/*
 * ExtropiaServlet.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.app.servlet;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.eXtropia.auth.AuthManager;
import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.mail.Mailer;
import com.eXtropia.util.Utils;
import com.eXtropia.view.ViewManager;

/**
 * <code>ExtropiaServlet</code> is now an abstract class that has a static class version.
 * All applications should extend this class and obtain a handle on ExtropiaVersion
 * which should be able to show the version of the ADT and the version of the application.
 *
 * @author  <a href="tim.parker@eXtropia.com">Tim Parker</a>
 * @author  <a href="st_lim@eXtropia.com">Lim Swee Tat</a>
 * @version $Id: ExtropiaServlet.java,v 1.22 2002/11/16 09:08:05 rayfd Exp $
 */
public abstract class ExtropiaServlet extends HttpServlet implements Configurable {
    public static final String CONFIG_FILENAME = "ConfigFile";
    public static final String CONFIG_RELOADABLE = "Reloadable";
    private static com.eXtropia.app.ExtropiaVersion _version;
    private Doctor _doctor;
    protected ActionManager _actionManager;
    protected boolean _reloadable = false;
    protected File _configFile;
    protected long _lastConfigured;

    /* (non-Javadoc)
     * @see java.lang.Object#Object()
     */
    public ExtropiaServlet() {
    }

    /**
     * Method getAppVersion.
     * @return ExtropiaVersion
     * @throws ConfigException
     */
    protected abstract com.eXtropia.app.ExtropiaVersion getAppVersion() throws ConfigException;

    /* (non-Javadoc)
     * @see javax.servlet.Servlet#init(javax.servlet.ServletConfig)
     */
    public void init(ServletConfig servletConfig) throws ServletException {
        super.init(servletConfig);

        try {
            _version = getAppVersion();
        } catch (ConfigException e) {
            throw new ServletException("No version found");
        }

        System.out.println(_version.showVersion());

        String fileName = servletConfig.getInitParameter(CONFIG_FILENAME);

        // Check if the fileName has a . in front of it. If it does
        // then we know that we want a relative path to the servlet context
        //
        if (fileName.startsWith(".")) {
            fileName = Utils.getFullPathName(getServletContext(), fileName);
        }

        try {
            _configFile = new File(fileName);
        } catch (Exception e) {
            throw new ServletException("Specified file: " + fileName + " is invalid");
        }

        _configFile = new File(fileName);

        if (!_configFile.exists()) {
            throw new ServletException("Specified config file does not exist: " + fileName);
        }

        String reloadable = servletConfig.getInitParameter(CONFIG_RELOADABLE);

        if (reloadable != null) {
            _reloadable = Boolean.valueOf(reloadable).booleanValue();
        }

        configureServlet();
        _doctor.trace(new TraceEvent(this, TraceEvent.INFO, _version.showVersion()));
    }

    /**
     * Method configureServlet.
     * @throws ServletException
     */
    private void configureServlet() throws ServletException {
        _lastConfigured = System.currentTimeMillis();

        Config config = null;

        try {
            // GB: Mar 4, 2002 -- We no longer pass a file reader
            // because we need to be able to get information about the
            // path of the file within the Config
            // itself.
            //
            //FileReader fr= new FileReader(_configFile);
            config = new Config(_configFile);
        } catch (Exception e) {
            e.printStackTrace();
            throw new ServletException("Failed to open specified " + CONFIG_FILENAME + " (" + _configFile.getAbsolutePath() + "): " + e.getMessage());
        }

        try {
            configure(config);
        } catch (ConfigException e) {
            throw new ServletException("Failed to configure: " + e.getMessage());
        }
    }

    /* (non-Javadoc)
     * @see com.eXtropia.config.Configurable#configure(com.eXtropia.config.Config)
     */
    public void configure(Config config) throws ConfigException {
        ServletContext servletContext = getServletContext();

        String appName = config.getName();
        String appInstance = config.getValue();
        this.configureDoctor(config, appInstance, appName);

        if (config.hasChild(Mailer.CFG_BASE)) {
            Mailer.configureMailer(config.getChild(Mailer.CFG_BASE));
        }

        ViewManager viewManager;

        if (config.hasChild(ViewManager.CFG_BASE)) {
            try {
                viewManager = new ViewManager(appName, appInstance);
                viewManager.configure(config.getChild(ViewManager.CFG_BASE));
            } catch (Exception e) {
                throw new ConfigException("Failed to create ViewManager: " + e.getMessage());
            }
        } else {
            throw new ConfigException("No " + ActionManager.CFG_BASE + " specified in Config");
        }

        AuthManager authManager;

        if (config.hasChild(AuthManager.CFG_BASE)) {
            _doctor.log(new LogEvent(this, LogEvent.INFO, "Using custom AuthManager"));
            authManager = (AuthManager) config.getChild(AuthManager.CFG_BASE).createAndConfigureInstance();
        } else {
            authManager = (AuthManager) servletContext.getAttribute(AuthManager.SERVLET_CONTEXT_ATTRIBUTE_NAME);

            if (authManager == null) {
                _doctor.log(new LogEvent(this, LogEvent.INFO, "No AuthManager being used"));
            } else {
                _doctor.log(new LogEvent(this, LogEvent.INFO, "Using AuthManager from ServletContext"));
            }
        }

        if (config.hasChild(ActionManager.CFG_BASE)) {
            try {
                String type = config.getChild(ActionManager.CFG_BASE).getType();

                if (type == null) {
                    throw new ConfigException("No type attribute specified for: " + ActionManager.CFG_BASE);
                }

                Class ActionManagerClass = Class.forName(config.getChild(ActionManager.CFG_BASE).getType());
                Constructor c = ActionManagerClass.getConstructor(new Class[] { String.class, String.class, AuthManager.class, ViewManager.class });
                _actionManager = (ActionManager) c.newInstance(new Object[] { appName, appInstance, authManager, viewManager });
                _actionManager.setServletContext(this.getServletContext());
                _actionManager.configure(config.getChild(ActionManager.CFG_BASE));
            } catch (ConfigException ce) {
                throw ce;
            } catch (Exception e) {
                e.printStackTrace();
                throw new ConfigException("Failed to create ActionManager: " + e.getMessage());
            }
        } else {
            throw new ConfigException("No " + ActionManager.CFG_BASE + " specified in Config");
        }

        _doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Finished initializing " + appName));
    }

    /**
     * Method configureDoctor.
     * @param config
     * @param appInstance
     * @param appName
     * @throws ConfigException
     */
    protected void configureDoctor(Config config, String appInstance, String appName) throws ConfigException {
        try {
            Config diagnosticsConfig = config.getChild(Diagnostics.CFG_BASE);
            Diagnostics.configureFactory(diagnosticsConfig);
        } catch (ConfigException ce) {
            throw new ConfigException("Failed to configure diagnostics" + ce.getMessage());
        }

        _doctor = Diagnostics.getDoctor(new String[] { appInstance, appName });
        _doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Initialising " + appName + "..."));
    }

    /* (non-Javadoc)
     * @see javax.servlet.http.HttpServlet#doGet(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        doGetPost(request, response);
    }

    /* (non-Javadoc)
     * @see javax.servlet.http.HttpServlet#doPost(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    public void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        doGetPost(request, response);
    }

    /**
     * Method doGetPost.
     * @param request
     * @param response
     * @throws ServletException
     * @throws IOException
     */
    public void doGetPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        String query = request.getQueryString();

        if ((query != null) && !query.equals("") && "version".equalsIgnoreCase(query)) {
            if (_version != null) {
                response.setContentType("text/html");

                PrintWriter out = null;
                out = response.getWriter();
                out.println(_version.showHTMLVersion());
                out.flush();
            } else {
                System.err.println("Error getting the version.");
                response.sendError(HttpServletResponse.SC_NOT_FOUND, "This is not a release version");
            }
        } else {
            if (_reloadable) {
                synchronized (_configFile) {
                    if (_lastConfigured < _configFile.lastModified()) {
                        configureServlet();
                    }
                }
            }

            ServletUtils.updateLocale(request);
            _actionManager.execute(request, response);
        }
    }
} // ExtropiaServlet.java
