/* 
 * RegisterUtility.java
 * Copyright (C) 2000 - 2001 eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.auth;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.util.ExtropiaException;

public class RegisterUtility {
    public static void main(String args[]) throws IOException {
        if (args.length != 1) {
            printUsage();
        }

        String configFile = args[0];
        FileReader reader = null;
        Config config = null;

        try {
            reader = new FileReader(configFile);
        } catch (FileNotFoundException e) {
            System.err.println("Config file " + configFile + " not found.");
            System.exit(-1);
        }

        try {
            config = new Config(reader);
        } catch (ConfigException e) {
            System.err.println("Error in configuration: " + e.getMessage());
            e.printStackTrace();
            System.exit(-1);
        }

        Config authenticatorConfig = findAuthenticatorConfig(config);
        if (authenticatorConfig == null) {
            System.err.println("No " + Authenticator.CFG_BASE + " specified  in config");
            System.exit(-1);
        }

        List attributes = new ArrayList();
        if (!authenticatorConfig.hasChild(Authenticator.CFG_USERFAC_BASE)) {
            System.err.println("Need " + Authenticator.CFG_USERFAC_BASE + " in " + Authenticator.CFG_BASE);
            System.exit(-1);
        }

        Config[] attributesConfig = authenticatorConfig.getChild(Authenticator.CFG_USERFAC_BASE).getChildren();
        for (int i = 0; i < attributesConfig.length; i++)
            attributes.add(attributesConfig[i].getValue());

        Authenticator authenticator = null;
        try {
            authenticator = (Authenticator) authenticatorConfig.createAndConfigureInstance();
        } catch (ConfigException e) {
            System.err.println("Could not create " + Authenticator.CFG_BASE + ": " + e.getMessage());
            System.exit(-1);
        }

        config = null;
        authenticatorConfig = null;

        boolean registerUser = true;
        String another;
        BufferedReader userInput = new BufferedReader(new InputStreamReader(System.in));
        while (registerUser) {
            registerUser(attributes, userInput, authenticator);
            another = null;
            while (another == null || !(another.toLowerCase().equals("y") || another.toLowerCase().equals("n"))) {
                System.out.print("Register another user? [Y/N]: ");
                another = readNullLine(userInput, true);
            }
            registerUser = another.toLowerCase().equals("y");
        }
    }

    protected static void registerUser(List attributes, BufferedReader userInput, Authenticator authenticator) throws IOException {
        System.out.print("Username: ");
        String uid = readNullLine(userInput, true);

        // does such a user exist already?
        try {
            if (authenticator.getUser(uid) != null) {
                System.err.println("User " + uid + " already exists.");
                return;
            }
        } catch (ExtropiaException e) {
            System.err.println("Error while checking for user: " + uid + ": " + e.getMessage());
            System.exit(-1);
        }

        System.out.print("Password: ");
        String password = readNullLine(userInput, false);

        System.out.print("Confirm password: ");
        String confirmPassword = readNullLine(userInput, false);

        if (!password.equals(confirmPassword)) {
            System.out.println("Passwords Not The Same");
            return;
        }

        // ask sysadmin for values for the attribs
        Map attribs = getAttributes(attributes, userInput);

        try {
            User user = authenticator.register(uid, password, attribs);
            System.out.println("" + user.getId() + " registered");
        } catch (ExtropiaException e) {
            System.err.println("Application Error registering user: " + e.getMessage());
            System.exit(-1);
        }
    }

    private static Map getAttributes(List attributes, BufferedReader reader) throws IOException {
        boolean printedEnterAttributes = false;

        Map _map = new HashMap(attributes.size());

        String val;
        String attr;
        for (Iterator iter = attributes.iterator(); iter.hasNext();) {
            attr = iter.next().toString();
            if (attr.equals(User.USER_ATTR_ID) || attr.equals(User.USER_ATTR_PASSWORD))
                continue;
            else {
                if (!printedEnterAttributes) {
                    System.out.println("Enter the value for each of the following attributes:");
                    printedEnterAttributes = true;
                }
            }
            System.out.print(attr + ": ");
            val = readNullLine(reader, true);
            if (val != null)
                _map.put(attr, val);
        }
        return _map;
    }

    protected static String readNullLine(BufferedReader reader, boolean printValue) throws IOException {
        String val = reader.readLine().trim();
        if (val.length() == 0)
            val = null;
        if (printValue)
            System.out.println(" [" + (val == null ? "None" : val) + "]");
        else
            System.out.println("");
        return val;
    }

    protected static Config findAuthenticatorConfig(Config config) {
        if (config.getName().equals(Authenticator.CFG_BASE)) {
            return config;
        } else {
            Config[] children = config.getChildren();
            Config result;

            for (int i = 0; i < children.length; i++) {
                result = findAuthenticatorConfig(children[i]);
                if (result != null)
                    return result;
            }
        }
        return null;
    }

    private static void printUsage() {
        System.out.println("Usage: java " + RegisterUtility.class.getName() + " ConfigFile");
    }
}
