/*
 * DatasourceAuthenticator.java
 * Copyright (C) 2000  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.auth.authenticator;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.sql.RowSet;

import com.eXtropia.auth.AuthException;
import com.eXtropia.auth.Authenticator;
import com.eXtropia.auth.Encryptor;
import com.eXtropia.auth.User;
import com.eXtropia.auth.user.DataSourceUser;
import com.eXtropia.auth.user.UserAttribute;
import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.datasource.Condition;
import com.eXtropia.datasource.Criterion;
import com.eXtropia.datasource.DataSource;
import com.eXtropia.datasource.Query;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.util.ExtropiaException;

/**
 * DatasourceAuthenticator: use a com.eXtropia.datasource for
 * authentication.
 *
 * @author Parijat Mishra (parijat@extropia.com)
 * @version 0.1
 *
 * Internal Revision $Revision: 1.11 $
 */
public class DatasourceAuthenticator implements Authenticator {
    private DataSource dsource1, dsource2;
    private Config userConfig;
    private HashMap userAttribs = new HashMap();
    // private HashMap userCache = new HashMap();
    private boolean passwordEncrypted = true;
    private String encryptionScheme;
    private Encryptor encryptor;

    private Doctor doctor;

    /**
     * Default constructor.
     */
    public DatasourceAuthenticator() {
    }

    /**
     * Implements Configurable
     */
    public void configure(Config config) throws ConfigException {

        doctor = Diagnostics.getDoctor(new String[] { CFG_BASE, "DataSourceAuthenticator" });

        // does the config belong to us?
        if (!config.getName().equals(Authenticator.CFG_BASE)) {
            throw new ConfigException("Incorrect Config passed to " + CFG_BASE);
        }

        if (!config.hasChild(CFG_USER_BASE))
            throw new ConfigException("Need " + CFG_USER_BASE + " to configure " + CFG_BASE);

        if (!config.hasChild(CFG_GROUP_BASE))
            throw new ConfigException("Need " + CFG_GROUP_BASE + " to configure " + CFG_BASE);

        Config config2 = config.getChild(CFG_USER_BASE);
        // get the datasource containing user attribs
        if (config2.hasChild(DataSource.CFG_BASE)) {
            try {
                dsource1 = (DataSource) config2.getChild(DataSource.CFG_BASE).createAndConfigureInstance();
            } catch (Exception e) {
                throw new ConfigException("Failed to configure " + CFG_USER_BASE + "::" + DataSource.CFG_BASE + ": " + e.getMessage());
            }
        } else {
            // someone forgot to configure a datasource for us
            throw new ConfigException("Need " + DataSource.CFG_BASE + " to configure " + CFG_USER_BASE);
        }

        // get the datasource containing group definitions
        // The datasource must contain two fields:
        //		group
        //		userid

        config2 = config.getChild(CFG_GROUP_BASE);
        if (config2.hasChild(DataSource.CFG_BASE)) {
            try {
                dsource2 = (DataSource) config2.getChild(DataSource.CFG_BASE).createAndConfigureInstance();
            } catch (Exception e) {
                throw new ConfigException("Failed to configure " + CFG_GROUP_BASE + "::" + DataSource.CFG_BASE + ": " + e.getMessage());
            }
        } /* else {
        			// someone forgot to configure a datasource for us
        			throw new ConfigException("Need " + DataSource.CFG_BASE +
        			" to configure " + CFG_GROUP_BASE);
        		}    */

        // what user attributes must we have?
        if (config.hasChild("Userfactory")) {
            userConfig = config.getChild("Userfactory");
        } else {
            throw new ConfigException("Need " + CFG_USERFAC_BASE + " to configure " + CFG_BASE);
        }

        // are passwords strored in cleartext?
        if (config.hasChild(CFG_PASSWORD_ENCRYPTION)) {
            String scheme = (String) config.getChild(CFG_PASSWORD_ENCRYPTION).getValue();
            if (scheme == null) {
                doctor.log(new LogEvent(this, LogEvent.INFO, "No scheme specified. Using default."));
                encryptor = new com.eXtropia.auth.crypto.JCrypt();
            } else if (scheme.equals("none")) {
                doctor.log(new LogEvent(this, LogEvent.INFO, "Using plaintext passwords"));
                passwordEncrypted = false;
            } else {
                try {
                    doctor.log(new LogEvent(this, LogEvent.INFO, "Trying custom Encryptor: " + scheme));
                    encryptor = (Encryptor) Class.forName(scheme).newInstance();
                } catch (Exception e) {
                    doctor.log(new LogEvent(this, LogEvent.CRIT, "Could not initialize Encryptor: " + scheme + ": " + e.getMessage()));
                    throw new ConfigException("Could not initialize class of type: " + scheme);
                }
            }
        } else {
            encryptor = new com.eXtropia.auth.crypto.JCrypt();
        }

        // convert userConfig to a convenient Map
        getUserAttribs();
    }

    /**
     * @param id a String containing the userid to be matched against the
     * User.USER_ATTR_ID field of the datasource.
     *
     * @param password a String containing the password (plaintext) to
     * be matched against the User.USER_ATTR_PASSWORD field of the
     * datasource.
     */
    public User authenticate(Object id, Object password) throws AuthException, ExtropiaException {
        if (id == null || !(id instanceof String) || password == null || !(password instanceof String))
            throw new AuthException("Arguments are null or non-String");

        Condition c1 = new Condition(User.USER_ATTR_ID, Condition.EQUAL, id);

        Condition c2 = new Condition(User.USER_ATTR_PASSWORD, Condition.EQUAL, password);
        Criterion c = new Criterion(Criterion.AND);
        c.add(c1);

        // DEBUG: print what is parading as auth tokens
        doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Userid = " + id + " Password = " + password));

        if (dsource1.count(c) == 1) {
            doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Got user"));
            // check password
            RowSet RS = dsource1.search(new Query(c));
            try {
                RS.next();
                String storedPass = RS.getString(User.USER_ATTR_PASSWORD);

                if (!passwordMatches(storedPass, (String) password)) {
                    doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Password incorrect"));
                    return null;
                }

            } catch (SQLException e) {
                throw new ExtropiaException("SQLException while authenticating:" + e.getMessage());
            }

            // password okay
            User _user = new DataSourceUser(userAttribs, dsource1, dsource2, c);

            // cache it
            // userCache.put(id, _user);
            return _user;
        }
        doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "No such user"));
        return null;
    }

    /**
     * @param id Object identifying the user.  If a user with this
     * id has been previously encountered then a cached user object will
     * be returned.
     * @param password the password in cleartext.
     * @param reginfo a  Map whose keys  are user attributes and whose
     * values  are Strings  containing  the values of the  attributes.
     * The special attributes--userid,  password,  and groups--may not
     * be present, and will be ignored.  Userid and password are to be
     * supplied as    in the  'id'  and  'password'  arguments. If  an
     * attribute is not supplied   and the datasource requires it,  an
     * AuthException will be thrown.  If  an attribute is supplied
     * which  is not a  recognized user attribute, an AuthException is
     * thrown.
     * @return a User object create using the reginfo.
     * @throws ExtropiaException if something went wrong with the
     * datasource
     * @throws AuthException if an unsupported attribute was supplied,
     * a required attribute was not supplied, or the user already exists
     */
    public User register(Object id, Object password, Map reginfo) throws AuthException, ExtropiaException {
        // is this user cached already?
        // 		if (userCache.containsKey(id)) {
        // 			doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE,
        // 										"register: "));
        // 			return (User) userCache.get(id);
        // 		}

        // does the user already exist?
        Criterion c = new Criterion(new Condition(User.USER_ATTR_ID, Condition.EQUAL, id));
        if (dsource1.count(c) > 0) {
            // oops, user already exists
            doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "User exists:" + (String) id));
            throw new AuthException("User exists");
        }

        Map persistentAttributes = new HashMap();
        Map transientAttributes = new HashMap();

        //split out persistent and non-persistent attributes
        Iterator i = reginfo.keySet().iterator();
        while (i.hasNext()) {
            String key = (String) i.next();
            if (userAttribs.containsKey(key))
                persistentAttributes.put(key, userAttribs.get(key));
            else
                transientAttributes.put(key, userAttribs.get(key));
        }

        // encrypt the password if necessary
        if (passwordEncrypted)
            password = encryptor.encrypt((String) password);

        doctor.log(new LogEvent(this, LogEvent.INFO, "register: adding user."));

        persistentAttributes.put(User.USER_ATTR_ID, id);
        persistentAttributes.put(User.USER_ATTR_PASSWORD, password);
        dsource1.add(persistentAttributes);

        // if all went well
        if (dsource1.count(c) > 0) {
            User user = new DataSourceUser(userAttribs, dsource1, dsource2, c);
            user.setAttributes(transientAttributes);
            doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "register: created user"));
            // cache it
            // userCache.put(id, user);
            return user;
        }
        doctor.log(new LogEvent(this, LogEvent.CRIT, "User creation unsuccessful"));
        throw new ExtropiaException("Could not create new user: datasource " + "returned no data");
    }

    /**
     * Implements Authenticator.
     */
    public User getUser(Object id) throws ExtropiaException {
        // 		if (userCache.containsKey(id))
        // 			return (User) userCache.get(id);
        // 		else {
        // try to get user info from datasource
        Condition cond = new Condition(User.USER_ATTR_ID, Condition.EQUAL, id);
        if (dsource1.count(cond) > 0) {
            User user = new DataSourceUser(userAttribs, dsource1, dsource2, new Criterion(cond));
            //				userCache.put(id, user);
            return user;
        }
        return null;
        //		}

    }

    /**
     * Implements Authenticator
     */
    public User[] getMembers(String group) throws ExtropiaException {
        Condition cond = new Condition(User.USER_ATTR_GROUPS, Condition.EQUAL, group);
        Query q = new Query(new Criterion(cond));
        RowSet rs = dsource2.search(q);
        Set users = new HashSet();

        try {
            if (rs != null) {
                while (rs.next()) {
                    Condition cond2 = new Condition(User.USER_ATTR_ID, Condition.EQUAL, rs.getString(User.USER_ATTR_ID));
                    users.add(new DataSourceUser(userAttribs, dsource1, dsource2, new Criterion(cond2)));
                }
                return (User[]) users.toArray(new User[0]);
            }
        } catch (SQLException e) {
            ExtropiaException ee = new ExtropiaException("SQLException while fetching Users");
            throw ee;
        }
        // no such group
        return null;
    }

    /* private methods here */

    // get the user attributes out from userConfig into userAttribs (a Map).
    private void getUserAttribs() throws ConfigException {
        Config[] children = userConfig.getChildren();
        for (int i = 0; i < children.length; i++) {
            UserAttribute u = new UserAttribute();
            u.configure(children[i]);
            userAttribs.put(u.getName(), u);

            // DEBUG: print the attributes we grok'ed
            doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Attribute: " + u.getName()));
        }
    }

    // compare the stored password with the supplied one, using encryption
    // if necessary
    private boolean passwordMatches(String stored, String supplied) {
        if (passwordEncrypted)
            return encryptor.match(supplied, stored);
        else
            return supplied.equals(stored);
    }

}
