/**
 * HTTPAuthManager.java
 *
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.auth.authmanager;

import java.io.IOException;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import com.eXtropia.auth.AuthException;
import com.eXtropia.auth.AuthManager;
import com.eXtropia.auth.Authenticator;
import com.eXtropia.auth.User;
import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.util.ExtropiaException;

/* An AuthManager that uses the HTTP Basic authentication scheme.
 * Created: Tue Jan  2 13:27:33 2001
 *
 * @author Parijat
 * @version 0.01
 */

public class HTTPAuthManager implements AuthManager, Configurable {

    public static final String CFG_BASE = "AuthManager";
    public static final String CFG_AUTHCTR_BASE = "Authenticator";
    public static final String CFG_AUTOREGISTER = "AutoRegister";

    public static final String HTTPAUTH_REALM = "HttpRealm";
    public static final String HTTPAUTH_DEFAULTREALM = "Basic";

    private Authenticator _authenticator;
    private boolean autoRegister = false;
    private String realm;

    private Doctor doctor;

    public HTTPAuthManager() {
    }

    /*
     * Implements Configurable
     */
    public void configure(Config config) throws ConfigException {
        doctor = Diagnostics.getDoctor(new String[] { CFG_BASE, "HTTPAuthManager" });
        //System.out.println("Start of Configuring HTTP AuthManager");
        doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Start of Configuring HTTP AuthManager"));
        if (!config.hasChild(CFG_AUTHCTR_BASE)) {
            throw new ConfigException("Need " + CFG_AUTHCTR_BASE + " to " + "configure " + CFG_BASE);
        }

        _authenticator = (Authenticator) config.getChild(CFG_AUTHCTR_BASE).createAndConfigureInstance();
        if (config.hasChild(CFG_AUTOREGISTER))
            autoRegister = true;
        if (config.hasChild(HTTPAUTH_REALM))
            realm = config.getChild(HTTPAUTH_REALM).getValue();
        else
            realm = HTTPAUTH_DEFAULTREALM;

        doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "End of Configuring HTTP AuthManager"));
        //System.out.println("End of Configuring HTTP AuthManager");
    }

    public User authenticate(Object request, Object response) throws AuthException, ExtropiaException {
        User _user;

        HttpServletRequest req = (HttpServletRequest) request;
        HttpServletResponse res = (HttpServletResponse) response;

        HttpSession session = req.getSession();
        try {
            if ((_user = (User) session.getAttribute(KEY_USER)) != null) {
                doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Using user object in session."));
                return _user;
            }
        } catch (ClassCastException cce) {
            doctor.log(new LogEvent(this, LogEvent.WARN, "Session object " + KEY_USER + " was " + "tied to a non-User object."));
            throw new ExtropiaException(CFG_BASE + ": Attribute " + KEY_USER + " not tied to a valid User object");
        }
        //if we get here, there is no user in the session
        //		try
        //		{
        doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "No user object in session."));

        // no User object.  Try getting the userid and password out
        String auth = req.getHeader("Authorization");
        if ((_user = allowUser(auth)) == null) {
            // TODO: autoRegister
            doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Sending WWW-Authenticate."));
            // For Internet Explorer, must set error AFTER sending the authenticate header...
            // or you will always get an unauthorized setting

            // -- yllee
            // rfc1945 states the following
            // The WWW-Authenticate response-header field must be included in 401
            // (unauthorized) response messages. The field value consists of at least one
            // challenge that indicates the authentication scheme(s) and parameters
            // applicable to the Request-URI.
            //
            // however, the original code does a sendError, which commits
            // the response, this causes problem when we want to
            // forward to a custom error page
            res.setHeader("WWW-Authenticate", "BASIC realm=\"" + realm + "\"");
            //res.res.sendError(res.SC_UNAUTHORIZED);
            res.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
            return null;
        } else {
            // allowed
            doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Returning verified user."));
            session.setAttribute(KEY_USER, _user);
            return _user;
        }
        //		}
        //		catch (IOException e)
        //		{
        //			doctor.log(new LogEvent(this, LogEvent.CRIT, "IOException caught during " + "authentication: " + e.getMessage()));
        //			throw new ExtropiaException(CFG_BASE + ": caught exception IOException: " + e.getMessage());
        //		}
    }

    public User getUserIfLoggedIn(Object r) {
        HttpServletRequest request = (HttpServletRequest) r;
        HttpSession session = request.getSession(false);
        if (session != null)
            return (User) session.getAttribute(KEY_USER);
        else
            return null;
    }

    public User register(Object id, Object password, Map reginfo) throws AuthException, ExtropiaException {
        return null; // don't know how to register
    }

    public User getUser(Object id) throws AuthException, ExtropiaException {
        return _authenticator.getUser(id);
    }

    public User[] getMembers(String group) throws AuthException, ExtropiaException {
        return _authenticator.getMembers(group);
    }

    private User allowUser(String authString) throws ExtropiaException {
        try {
            if (authString == null) {
                doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "No WWW-Auth data."));
                return null;
            }
            if (!authString.toUpperCase().startsWith("BASIC")) {
                doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "AuthType != BASIC."));
                return null; // we only do BASIC
            }

            doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Found WWW-Auth params."));
            String userpassEnc = authString.substring(6);
            sun.misc.BASE64Decoder dec = new sun.misc.BASE64Decoder();
            String userpassDecoded = new String(dec.decodeBuffer(userpassEnc));

            String username = userpassDecoded.substring(0, userpassDecoded.indexOf(":"));
            String password = userpassDecoded.substring(userpassDecoded.indexOf(":") + 1);
            doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Authenticating against database"));

            User _user = _authenticator.authenticate(username, password);
            if (_user == null)
                doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Database verification failed."));
            else
                doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Database verification succeeded."));
            return _user;
        } catch (IOException e) {
            doctor.log(new LogEvent(this, LogEvent.CRIT, "IOException during db access: " + e.getMessage()));
            throw new ExtropiaException(CFG_BASE + ": caught IOException: " + e.getMessage());
        }
    }
} // HTTPAuthManager
