/*
 * DataSourceUser.java
 *
 * Copyright (C) 2000  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package com.eXtropia.auth.user;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import javax.sql.RowSet;

import com.eXtropia.auth.AuthException;
import com.eXtropia.auth.User;
import com.eXtropia.datasource.Condition;
import com.eXtropia.datasource.Criterion;
import com.eXtropia.datasource.DataSource;
import com.eXtropia.datasource.Query;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.util.ExtropiaException;

/**
 * An implementation of the User interface which uses a DataSource
 * to get its data from.
 *
 * @author  Parijat Mishra (parijat@extropia.com)
 * @version
 */
public class DataSourceUser implements User {

    // DataSourceUser inherits the id, password and group fields from
    // the User interface.  The other attributes are defined in the
    // config file.
    // The special attributes User.USER_ATTR_ID, User.USER_ATTR_PASSWORD,
    // and User.USER_ATTR_GROUPS must be present in the configuration

    // special attributes
    private String userid;
    private String password;
    private HashSet groups = new HashSet();

    // a mapping of attribute name => UserAttribute object
    // the UserAttribute objects define the properties of each attribue
    // (for e.g. if an attribute is read-only/read-write).
    private Map attribConf;

    // the particular values of this user's attibutes
    // String => String.  The keys are the same as in #attribConf
    private HashMap attribs = new HashMap();

    // the DataSource containing this user's attributes
    // the field names are the names of the attributes
    private DataSource userDS;

    // the DataSource containing this user's groups
    // the field names are User.USER_ATTR_ID and User.USER_ATTR_GROUPS
    private DataSource groupDS;

    // the Criterion which identifies this user's data from others
    // from within the DataSource.  It is constructed by the Authenticator
    // which produces this object.
    private Criterion crit;

    private Doctor doctor;

    /**
     * @param attribConf A Map whose keys are Strings naming the
     * the attributes this User object must have and
     * whose values are UserAttribute objects containing
     * information about the attributes.  The attributes,
     * User.{USER_ATTR_ID, USER_ATTR_PASSWORD, USER_ATTR_GROUPS}
     * MUST be present.  Other attributes are app specific.
     *
     * @param ds1  A DataSOurce containing  the values  of the
     * attribs  in   attribConf.  The  fieldnames    in  the
     * DataSource  must be   the  keys in attribConf.    The
     * fields User.USER_ATTR_ID and  User.USER_ATTR_PASSWORD
     * must be present.
     *
     * @param ds2  A  DataSOurce  containing the user's  group
     * information.  This   class    requires   the   fields
     * User.USER_ATTR_ID and User.USER_ATTR_GROUP
     *
     * @param crit A Criterion  that identifies the  user's data within
     * all the data potentially available through the DataSource.  The
     * User    object will  be     reponsible  for  ensuring  that  it
     * accesses/changes   only data belonging   to  it (as defined  by
     * 'crit') to/from the DataSources.
     *
     */
    public DataSourceUser(Map attribConf, DataSource ds1, DataSource ds2, Criterion crit) throws ExtropiaException {
        // -------------------------------------------------------------
        // N.B. We could have used RowSet.  But it is not possible
        // for a JDBC 1.1 RowSet to update its contents back to the
        // database; that requires JDBC 2.0.  To enable the use of
        // older (err... current, I think) drivers, we use a DataSource
        // as it can update its contents back to the DB using JDBC 1.1
        // -------------------------------------------------------------
        this.attribConf = attribConf;
        this.userDS = ds1;
        this.groupDS = ds2;
        this.crit = crit;

        doctor = Diagnostics.getDoctor(new String[] { CFG_BASE, "DataSourceUser" });
        /*
         * get user attribs out from the DataSource into a hash
         */
        Query q = new Query(crit);
        RowSet RS = userDS.search(q);
        String[] fields = userDS.getFields();

        try {
            if (!RS.next()) {
                // Heh! No such user. Complain bitterly and die.
                doctor.log(new LogEvent(this, LogEvent.CRIT, "No such user. This should not have " + "been allowed by the Authenticator"));
                throw new ExtropiaException("No user information in RowSet");
            }

            String temp = "User: Attributes:\n"; // for debug
            for (int i = 0; i < fields.length; i++) {
                String val = RS.getString(fields[i]);
                temp += fields[i] + " = " + val + "\n";
                if (fields[i].equals(User.USER_ATTR_ID))
                    userid = val;
                else if (fields[i].equals(User.USER_ATTR_PASSWORD))
                    password = val;
                else
                    attribs.put(fields[i], val);
            }

            doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, temp));

            /*
             * get group information
             */
            doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Groups:\n"));
            RS = groupDS.search(q);
            while (RS.next()) {
                String grp = RS.getString(USER_ATTR_GROUPS);
                temp += grp + " ";
                groups.add(grp);
                doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Added group: " + grp + "\n"));

            }
        } catch (SQLException e) {
            doctor.log(new LogEvent(this, LogEvent.CRIT, "SQLException while fetching user attribs" + e.getMessage()));
            throw new ExtropiaException("SQLException while fetching user attribs" + e.getMessage());
        }

    }

    // NOTE: groups cannot be modified from here
    public void setAttribute(String attribute, Object value) throws AuthException, ExtropiaException {
        // is this a pre-defined attribute?
        doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "updating attibute: " + attribute + " => " + value));
        if (attribConf.containsKey(attribute)) {

            // yes. so is this writable?
            if (!((UserAttribute) attribConf.get(attribute)).isWritable()) {
                throw new AuthException(CFG_BASE + ": Attribute " + attribute + " is not modifiable.");
            }
            // yes. so change it here and...
            attribs.put(attribute, value);

            // ...update the DataSource
            HashMap _map = new HashMap();
            _map.put(attribute, value);
            userDS.update(_map, crit);
            doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "New persistent attribute added to user: " + userid));
        } else {
            // no such key, so this is a temp attribute that won't be
            // sent back to the datasource
            attribs.put(attribute, value);
            doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "New transient attribute added to user: " + userid));
        }
    }

    // NOTE: groups can't be extracted from here
    public Object getAttribute(String attribute) {
        //if (! attribs.containsKey(attribute) )
        //	throw new AuthException(CFG_BASE + ": No such attribute: " + attribute + ".");
        return attribs.get(attribute);
    }

    /**
     * @param attributes a Map whose keys are Strings containing the names of
     * the user's properties, and whose values are objects whose
     * toString() methods give Strings containing the
     * values of the user's properties
     */
    public void setAttributes(Map attributes) throws AuthException, ExtropiaException {
        Iterator i = attributes.entrySet().iterator();
        Map.Entry entry;
        while (i.hasNext()) {
            entry = (Map.Entry) i.next();
            setAttribute((String) entry.getKey(), entry.getValue());
        }
    }

    /**
     * @return A copy of the user's attributes as a Map.
     * Returning a copy ensures that
     * the user's properties are only modified
     * through its object.
     */
    public Map getAttributes() {
        return new HashMap(attribs);
    }

    // -----------------------------------------------------------------------
    // The special properties' accessor/mutators
    // -----------------------------------------------------------------------

    public Object getId() {
        // all User objects MUST have an 'id' attribute,
        // so we don't explicitly check for it here
        return userid;
    }

    // no setId()

    public Object getPassword() {
        return password;
    }

    public void setPassword(Object password) throws AuthException, ExtropiaException {
        // has password been marked as unmodifiable in the config
        if (((UserAttribute) attribConf.get(User.USER_ATTR_PASSWORD)).isWritable()) {
            // nope, it is modifiable
            try {
                this.password = (String) password;
            } catch (ClassCastException e) {
                throw new AuthException("Expected String as argument");
            }

            Map _map = new HashMap(1);
            _map.put(User.USER_ATTR_PASSWORD, this.password);

            userDS.update(_map, crit);
        } else {
            // oops, you are not allowed to modify the passwords
            throw new AuthException("Password in not modifiable");
        }
    }

    public boolean isMemberOf(String group) {
        // all User objects MUST have a group attribute,
        // so we don't explicitly check it here
        return groups.contains(group);
    }

    public void addGroupMembership(String group) throws AuthException, ExtropiaException {
        if (!((UserAttribute) attribConf.get(USER_ATTR_GROUPS)).isWritable())
            throw new AuthException(CFG_BASE + ": Groups are not modifiable.");
        groups.add(group);

        // update datasource
        HashMap _map = new HashMap(2);
        _map.put(USER_ATTR_ID, getAttribute(USER_ATTR_ID));
        _map.put(USER_ATTR_GROUPS, group);
        groupDS.add(_map);
    }

    public void removeGroupMembership(String group) throws AuthException, ExtropiaException {
        if (!((UserAttribute) attribConf.get(USER_ATTR_GROUPS)).isWritable())
            throw new AuthException(CFG_BASE + ": Groups are not modifiable.");
        groups.remove(group);

        // update datasource
        Condition cond[] = new Condition[2];
        cond[0] = new Condition(USER_ATTR_ID, Condition.EQUAL, (String) attribs.get(USER_ATTR_ID));
        cond[1] = new Condition(USER_ATTR_GROUPS, Condition.EQUAL, group);
        groupDS.delete(new Criterion(cond));
    }

    public String[] getGroupMemberships() {
        String[] arr = new String[groups.size()];
        return (String[]) groups.toArray(arr);
    }

    public void setGroupMemberships(String[] grps) throws AuthException, ExtropiaException {
        if (!((UserAttribute) attribConf.get(USER_ATTR_GROUPS)).isWritable())
            throw new AuthException(CFG_BASE + ": Groups are not modifiable.");

        // Clear local cache ===================
        groups.clear();
        // Delete all groups to which this user belonged =============
        Condition _cond = new Condition(USER_ATTR_ID, Condition.EQUAL, (String) attribs.get(USER_ATTR_ID));
        groupDS.delete(_cond);

        // now add the new groups
        for (int i = 0; i < grps.length; i++) {
            groups.add(grps[i]);
            addGroupMembership(grps[i]);
        }

    }

    /* private methods */

}
