/* 
 * UserAttribute.java
 * Copyright (C) 2000  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.auth.user;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;

/*
 * Serves to hold the attibutes in an UserAttribute tag
 *
 * A User object supplies its properties as attributes to clients.  A
 * UserAttribute tag identifies such properties, one tag per property.
 * Typically, these tags will be used by an Authenticator that creates
 * User objects to get the values of these properties from some source
 * of data, such as a database.
 * 
 * Each UserAttribute has three parameters:
 *	name:	this is the name of the attribute in the User object
 *		that the client code can get/set
 *	field:	this is an internal name which the Authenticator which
 *		produced the User object uses 
 *	permission:
 *		this has a value of either "ro" meaning read-only, or
 *		"rw" meaning read-write. An "ro" UserAttribute is a
 *		read-only item, and its value cannot be changhed by
 *		client.  An "rw" UserAttribute can be changed by the
 *		cleint and such change will reflect in  the source
 *		of the User authentication data.
 *
 * @author Parijat Mishra <parijat@extropia.com)
 * @version 0.01
 */
public class UserAttribute implements Configurable {

    public static final String CFG_BASE = "UserAttribute";
    public static final String CFG_PERM_BASE = "UserAttributePermission";

    protected String name;
    protected boolean writable = true;

    public UserAttribute() {
    }

    public void configure(Config config) throws ConfigException {
        // get our attribues
        if (!(config.getName().equals(CFG_BASE)))
            throw new ConfigException("Improper config object passwed to " + CFG_BASE);

        name = config.getValue();
        Config perm = config.getChild(CFG_PERM_BASE);
        if (perm != null)
            writable = perm.getValue().equals("rw");
        // could have more fields here in future
    }

    public String getName() {
        return name;
    }

    public boolean isWritable() {
        return writable;
    }

}
