/**
 * Config.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.config;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringReader;
import java.io.Writer;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Properties;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import com.eXtropia.util.ExtropiaException;

/**
 * This class takes an XML document and transforms it into a Config object
 * tree after parsing it into a DOM tree using Xerces or Jaxp. This class
 * should be compatible with any version of Xerces or JAXP released since
 * July 2000.
 *
 * Take the following example XML (defining the DataSource parameters for a WebDB)
 *
 * <WebDB>
 * ...
 *   <DataSource type="com.eXtropia.datasource.jdbc.JDBCDataSource">
 *     <Driver type="oracle.jdbc.driver.OracleDriver"/>
 *     <Connection>
 *        jdbc:oracle:thin:@192.168.168.10:1521:ora1
 *        <user>testjava</user>
 *        <password>testjava</password>
 *     </Connection>
 *     <Table>
 *       Address
 *       <Column><Id>first_name</Id></Column>
 *       <Column><Id>last_name</Id></Column>
 *     </Table>
 *     <CopyRowSet>true</CopyRowSet>
 *   </DataSource>
 * ...
 * </WebDB>
 *
 * Then in order to get out all the required fields we would execute the following code:
 *
 *      java.io.File configFile = new File("/var/adm/conf/webdb.xml");
 *      // First creat the config
 *      try {
 * 	  Config config= new Config(configFile);
 *      } catch (ConfigException e) {
 *        System.err.println("Config Failed To Create: " + e.getMessage());
 *        e.printStackTrace();
 *        System.exit(1);
 *      }
 *
 *      // Then we would interet the config tree. Note that dealing with
 *      // a config tree is not monolithic but rather tends to be Factory-driven.
 *      //
 *      // In the above example, DataSource has a "type" attribute which
 *      // tells Config what type of object to create and to pass all further
 *      // children Config nodes to the configure() method of the object
 *      // that was just created.
 *      DataSource ds;
 *	try {
 *		config = config.getChild("WebDB").getChild("DataSource");
 *              DataSource ds = config.create(); // Factory to create a DataSource object
 *	} catch(ConfigException ce) {
 *		System.err.println("Arrgggghhhhh: Expected format not found");
 *		System.exit(1);
 *	}
 *
 *      //
 *      // Here is something similar to what would happen inside of
 *      // JDBCDataSource' config method...
 *      //
 *
 *      String    tableName;
 *      Arraylist columnNames = new ArrayList;
 *
 *      try {
 *        config = config.getChild("Table");
 *        tableName = config.getValue();
 *      } catch (ConfigException e) {
 *        System.err.println("Expected a <Table> parameter: " + e.getMessage();
 *        System.exit(1);
 *      }
 *
 *      try {
 *        Config [] children = config.getChildren();
 *	  System.out.println("The Following are columns in the table:");
 *	  for(int i=0; i<children.length; i++) {
 *		if (children[i].getName().equals("Column")) {
 *			System.out.println("  " + children[i].getChild("Id").getValue(currentKey));
 *		}
 *	} ...
 *
 * Additional usage of Config allows us to include files and set up "macros"
 * whose value may be processed elsewhere in the XML file.
 *
 * Note that the concept of a macro can also be accomplished via DTD Entities. However,
 * DTDs pose technical limitations (no inheritance, only one DTD file allowed to be included)
 * so we have build out own macro system. XML Schemas could also support such
 * similar features but given that the XML parsers have only recently started
 * adding support, we felt it prudent to be as backwards compatible as possible.
 *
 * The macro/XML file include mechanism uses standard XML allowing the macros
 * and include files themselves to be validated by an XML parser. This allows
 * common errors such as unbalanced tags to be caught and diagnosed more
 * quickly.
 *
 * INCLUDE FILES
 *
 * To create an include file, you must surround the XML file with a single tag
 * whose name indicates the type of data to include. An "id" attribute on that
 * tag indicates the specific name of that included XML tag type. Finally,
 * "value_only" attribute indicates how the include is processed.
 *
 * INCLUDE FILE EXAMPLE
 *
 * <Connection id="addressbook_connection" value_only="false">
 *   jdbc:oracle:thin:192.168.168.10:1521:ora1
 *   <User>testjava</User>
 *   <Password>testjava</Password>
 * </Connection>
 *
 * Then, to include this file inside a larger XML file you would use the
 * following code:
 *
 * <DataSource type="...">
 *   <Driver type="..."/>
 *   <Connection id="addressbook_connection" include="include.xml"/>
 *   <Table>
 *     ....
 *   </Table>
 * </DataSource>
 *
 * Notice that the outermost tag in the include file and the tag with the
 * include attribute both are named "Connection". Also notice that the id's
 * are both "addressbook_connection". This provides two double checks so that
 * mistakes in including the wrong file in the wrong part of the main XML
 * file do not happen.
 *
 * VALUE_ONLY ATTRIBUTE
 *
 * The value_only attribute controls how the macro or include file is processed.
 * If value_only is false, then the entire set of tags including the outermost
 * tag is imported into the calling XML file. Only the "id" and "value_only"
 * attributes are removed from the inserted data.
 *
 * When value_only equals true, then the outermost tag is discarded and only
 * the "value" of the outermost tag (in other words, anything inside it) is
 * imported into the calling XML file.
 *
 * VALUE_ONLY EXAMPLE
 *
 * Let's assume that instead of making an entire include file for Connection
 * we were only interested in a single IP address. Yet, the IP address has
 * no tag surrounding it. Thus the following would NOT work:
 *
 * <Ip id="addressbook_ip" value_only="false">1.1.1.1</Ip>
 *
 * Where the main XML file had something like
 *
 * <DataSource type="...">
 *   <Driver type="..."/>
 *   <Connection>
 *     jdbc:oracle:thin:<Ip id="addressbook_ip" include="include.xml"/>:1521:ora1
 *     <User>testjava</User>
 *     <Password>testjava</Password>
 *   </Connection>
 *   <Table>
 *     ....
 *   </Table>
 * </DataSource>
 *
 * Why wouldn't this work? Well, if the outermost tag is included, the JDBC
 * URL will be wrong. It will contain XML tags instead of just the text
 * of the IP address as follows:
 *
 *     jdbc:oracle:thin:<Ip>1.1.1.1</Ip>:1521:ora1
 *
 * To force the macro system to discard the outermost tag and only use
 * the *value* inside the outermost tag, use value_only="true".
 *
 * This will provide the desired result of:
 *
 *     jdbc:oracle:thin:1.1.1.1:1521:ora1
 *
 * USING MACROS
 *
 * Using a macro is the same as including a file except that you simply
 * leave off the include attribute. Without an "include" attribute, the
 * Config logic will assume that the macro will be retrieved from memory
 * using the value of "id" as a unique key instead of reading a file.
 *
 * CREATING A MACRO
 *
 * Before using a macro, a macro must be saved in memory somehow. The trick
 * to doing this is to use the "value_only" attribute. When an XML tag
 * with an id="" exists and the value_only attribute is defined, Config
 * knows that this means that the data between the tags is meant to be
 * saved to memory and keyed on the value of "id" (similar to a HashMap).
 *
 * The following XML code inside of the XML file would define a macro called
 * "soap_uri"
 *
 * <Uri id="soap_uri" value_only="true">
 *   http://www.test.com:8080/soap/servlet/rpcrouter
 * </Uri>
 *
 * Recall that to USE a macro, you just need to use the ID such as
 *
 * <SoapConfig>
 *   <Uri id="soap_uri"/>
 * </SoapConfig>
 *
 * Because there is no "value_only" attribute in the second use of <Uri>
 * Config knows to RETRIEVE a macro from memory with the ID equal to "soap_uri"
 * instead of save the macro to memory that the first use demonstrates.
 *
 * MACRO SCOPE
 *
 * Macros are currently global. The first time they are defined, they will
 * remain defined in memory. If you redefined a macro with the same id,
 * the old one will be erased.
 *
 * Although the macros are "global" they are saved in depth-first search
 * order. That is, Config descends into the XML tree from top to bottom
 * in the order that the tags appear in a Config file and is processed in
 * a depth first manner.
 *
 * So the following tags:
 *
 * <1>
 *   <3>
 *     <4>
 *   <5>
 *     <6>
 * <2>
 *
 * Would be visited in 1,3,4,5,6,2 order.
 *
 * This is contrasted by a breadth-first definition of macros which would
 * define them using 1,2,3,5,4,6 order. Again, I stress that we use depth-first
 * XML parsing and macro definition.
 *
 * MACRO LIBRARIES
 *
 * The most likely way you will use include files and macros is to set them
 * up in a separate library file that is included at the start of the main
 * XML file.
 *
 * We recommend a structure like the following for the include file:
 *
 * <MacroLibrary id="webdb_app_macros"/ value_only="true">
 *   <Macro1 id="macro_1" value_only="true">
 *     ....
 *   </Macro1>
 *   <Macro2...>
 *    ...
 *   </<Macro2>
 *    ...etc...
 * </MacroLibrary>
 *
 * By using value_only="true", we assure that all the macros will be defined
 * at the top-most level of the XML file (early).
 *
 * BUILT-IN MACROS
 *
 * There is one built-in macro that you do not have to define. It is a
 * "value_only" default_path macro. It returns the path prefix to the
 * current Main xml file without the main XML file itself.
 *
 * This allows you to easily configure file values in Config using a
 * default path instead of absolute paths. For example:
 *
 * <Diagnostics>
 *   <LogListener type="com.eXtropia...FileListener">
 *     <File><Path id="default_path"/>../../../diag.log
 *   </LogListener>
 * </Diagnostics>
 *
 * To create a diagnostics log file relative but a few directories above
 * the current XML file.
 *
 * Note that the paths to include files are also relative to the
 * current main XML config file.
 *
 * @author <a href="mailto:rayfd@eXtropia.com">Ray Djajadinata</a>
 * @author <a href="mailto:gunther@eXtropia.com">Gunther Birznieks</a>
 * @author <a href="mailto:st_lim@eXtropia.com">Lim Swee Tat</a>
 * @author <a href="mailto:carlo@eXtropia.com">Carlo Ravagli</a>
 * @author <a href="mailto:tim@eXtropia.com">Tim Parker</a>
 * @version $Id: Config.java,v 1.43 2002/11/16 08:45:10 rayfd Exp $
 */
public class Config {
    //XML ATTRIBUTES
    protected static final String TYPE_ATTR = "type";
    protected static final String INCLUDE_ATTR = "include";
    protected static final String VALUE_ONLY_ATTR = "value_only";
    protected static final String ID_ATTR = "id";

    // The root XML element that corresponds to this configuration
    protected Element _element;

    /**
     * Creates a new <code>Config</code> instance from a File object. This is
     * the recommended method for creating san XML based Config object as
     * special features such as including other XML Files can resolve
     * path locations from the original passed in File object.
     *
     * @param is a <code>File</code> value
     * @exception ConfigException if an error occurs
     */
    public Config(File f) throws ConfigException {
        this(f, null, null);
    }

    /**
     * Creates a new <code>Config</code> instance from an empty
     * XML node. Use this to manually build up a configuration
     * tree instead of reading from a file.
     *
     * @param is a <code>String</code> value
     * @exception ConfigException if an error occurs
     */
    public Config(String rootName) throws ConfigException {
        this(new StringReader("<" + rootName + ">" + "</" + rootName + ">"));
    }

    /**
     * Creates a new <code>Config</code> instance.
     *
     * @param is an <code>InputStream</code> value
     * @exception ConfigException if an error occurs
     * @deprecated This is still used for backward compatibility, use Config(File) instead.
     * Config(Reader) is also usable but Config File provides more information
     */
    public Config(InputStream is) throws ConfigException {
        this(null, null, is);
    }

    /**
     * Creates a new <code>Config</code> instance from a Reader. Note that it is
     * recommended that you use Config(File) instead of Config(Reader)
     * because more advanced Config features rely on attributes passed from
     * a File object.
     *
     * @param is a <code>Reader</code> value
     * @exception ConfigException if an error occurs
     */
    public Config(Reader r) throws ConfigException {
        this(null, r, null);
    }

    /**
     * Creates a new <code>Config</code> instance from an InputStream, Reader,
     * or a File. Since the code for each of these is virtually identical,
     * we decided to use this method for all three.
     *
     * @param is a <code>File</code> value
     * @param is a <code>Reader</code> value
     * @param is an <code>InputStream</code> value
     * @exception ConfigException if an error occurs
     */
    private Config(File f, Reader r, InputStream is) throws ConfigException {
        _element = processXMLFile(f, r, is, true);
    }

    /**
     * Processes an XML file for later configuration
     */
    private Element processXMLFile(File f, Reader r, InputStream is, boolean processXMLTags) throws ConfigException {

        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        Document document = null;

        String absolutePath = null;
        String canonicalPath = null;
        String defaultPath = null;
        HashMap macros = null;

        try {
            builder = factory.newDocumentBuilder();
            if (f != null) {
                absolutePath = f.getAbsolutePath();
                canonicalPath = f.getCanonicalPath();

                // The following solves problems with the various XML parsers
                //
                // In particular, Xerces 1.2.0 cannot take a File interface
                // directly because it strips off the path portion of the file
                // path and loses all absolute references
                //
                // Jaxp that comes with JRun 3.1 requires a protocol in front
                // of the system identifier (eg file:)
                //
                // Xerces 1.4.4 is the most flexible of all the parsers encountered
                // so nothing special is taken into account here and in fact,
                // the following few lines that build up the URI is
                // based on Xerces' document builder code that builds up a
                // URI from a java.io.File object
                //
                String uri = "file:" + absolutePath;
                if (File.separatorChar == '\\') {
                    uri = uri.replace('\\', '/');
                }
                document = builder.parse(new InputSource(uri));

                // if the parent is null, then it's the current directory.
                String parent = f.getParent();
                defaultPath = ((parent != null) ? parent : ".") + File.separatorChar;
            } else if (r != null) {
                document = builder.parse(new InputSource(r));
            } else {
                document = builder.parse(is);
            }
        } catch (ParserConfigurationException pce) {
            ConfigException configException = new ConfigException("Parser configuration exception:" + pce.getMessage());
            configException.setNestedException(new ExtropiaException(pce));
            throw configException;
        } catch (IOException ioe) {
            ConfigException configException = new ConfigException("IO Exception:" + ioe.getMessage());
            configException.setNestedException(new ExtropiaException(ioe));
            throw configException;
        } catch (SAXParseException spe) {
            String extraFileInfo = "";
            if (f != null) {
                extraFileInfo = "[" + absolutePath + " or " + canonicalPath + "]";
            }
            ConfigException configException = new ConfigException("File " + extraFileInfo + " specified is not of the correct format: " + spe.getMessage() + " at line = " + spe.getLineNumber() + " and column = " + spe.getColumnNumber() + " and system id = " + spe.getSystemId() + " and public id = " + spe.getPublicId());
            configException.setNestedException(new ExtropiaException(spe));
            throw configException;
        } catch (SAXException se) {
            ConfigException configException = new ConfigException("File specified is not of the correct format: " + se.getMessage());
            configException.setNestedException(new ExtropiaException(se));
            throw configException;
        } catch (Exception e) {
            ConfigException configException = new ConfigException("Unexpected Exception Caught: " + e.getMessage());
            configException.setNestedException(new ExtropiaException(e));
            throw configException;
        }

        // The following code replaced code the originally simply retrieved
        // the first child node. This was done to allow DTDs to be expressed
        // in the XML config file. A DTD is usually the first child node and
        // therefore we need to skip it if we are looking for the actual
        // first node that corresponds to an XML element. -- GB (Feb 28, 2002)
        //
        Element element = null;
        NodeList nodes = document.getChildNodes();
        Node node;
        for (int i = 0; i < nodes.getLength(); i++) {
            node = nodes.item(i);
            if (node instanceof Element) {
                element = (Element) node;
                break;
            }
        }

        // If there were no XML elements, the Config file is still valid,
        // but we should throw an exception to let the author know there is
        // something wrong with the file.
        //
        if (element == null) {
            if (f == null) {
                throw new ConfigException("No XML elements exist in the config file.");
            } else {
                throw new ConfigException("No XML elements exists in the config file: " + f.getName());
            }
        }

        // The following processes all includes and macros on the config tree
        if (processXMLTags) {
            macros = new HashMap();
            processNodes(element, defaultPath, macros, document);
        }
        return element;
    } // end of processDocument

    /**
     * Processes a tree of XML nodes for include files and macro
     * substitution recursively.
     */
    private void processNodes(Element parentElement, String defaultPath, HashMap macros, Document rootDocument) throws ConfigException {
        NodeList nodes = parentElement.getChildNodes();
        Node node;
        for (int i = 0; i < nodes.getLength(); i++) {
            node = nodes.item(i);
            if (node instanceof Element) {
                // Case 1: Node Attribute is include
                Attr attr;
                Attr valueOnlyAttr = ((Element) node).getAttributeNode(VALUE_ONLY_ATTR);
                Attr idAttr = ((Element) node).getAttributeNode(ID_ATTR);
                if ((attr = ((Element) node).getAttributeNode(INCLUDE_ATTR)) != null) {
                    // This check is necessary--if it's absolute and
                    // this is not done, the includeFile path will
                    // be invalid -- RD
                    String includeFile = null;
                    if (new File(attr.getValue()).isAbsolute()) {
                        includeFile = attr.getValue();
                    } else {
                        includeFile = defaultPath + attr.getValue();
                    }
                    Element topIncludeElement = processXMLFile(new File(includeFile), null, null, false);

                    if ((valueOnlyAttr = ((Element) topIncludeElement).getAttributeNode(VALUE_ONLY_ATTR)) != null) {
                        boolean valueOnly = Boolean.valueOf(valueOnlyAttr.getValue()).booleanValue();
                        // The Node names must match...
                        if (!node.getNodeName().equals(topIncludeElement.getNodeName())) {
                            throw new ConfigException("Includor Element Name: " + node.getNodeName() + " did not match Included Element Name: " + topIncludeElement.getNodeName());
                        }
                        // The attribute ids must match...
                        Attr includorIdAttr;
                        Attr includedIdAttr;

                        if ((includedIdAttr = ((Element) topIncludeElement).getAttributeNode(ID_ATTR)) == null) {
                            throw new ConfigException("Included Element Id Attribute for Element: " + topIncludeElement.getNodeName() + " is required.");
                        }
                        if ((includorIdAttr = ((Element) node).getAttributeNode(ID_ATTR)) == null) {
                            throw new ConfigException("Includor Element Id Attribute for Element: " + node.getNodeName() + " is required.");
                        }

                        if (!includorIdAttr.getValue().equals(includedIdAttr.getValue())) {
                            throw new ConfigException("Includor Element Id (" + includorIdAttr.getValue() + ") must match included Element Id (" + includedIdAttr.getValue() + ")" + " for includor Element name: " + node.getNodeName());
                        }
                        // Then we have to strip away the original node and start processing over
                        // But also, we have to add the included values...
                        if (valueOnly) {
                            parentElement.removeChild(node);
                            topIncludeElement = (Element) importNode(rootDocument, topIncludeElement, true);
                            NodeList includeNodes = topIncludeElement.getChildNodes();
                            Node includeNode;
                            Node nextNode = nodes.item(i + 1);
                            while (includeNodes.getLength() > 0) {
                                Node includeNodeToAdd = includeNodes.item(0);
                                if (includeNodeToAdd.getNodeType() == Node.TEXT_NODE) {
                                    includeNodeToAdd.setNodeValue(includeNodeToAdd.getNodeValue().trim());
                                }
                                if (nextNode == null) {
                                    parentElement.appendChild(includeNodeToAdd);
                                } else {
                                    parentElement.insertBefore(includeNodeToAdd, nextNode);
                                }
                            }
                            i--;
                            continue;
                            // Else we remove the original node and then replace it with
                            // the new one minus the extra attributes..
                        } else {
                            parentElement.removeChild(node);

                            // The following would work with a DOM Level 2
                            // compliant parser, but this is recent stuff, so we need
                            // to hand roll our own for backwards-compatibility
                            topIncludeElement = (Element) importNode(rootDocument, topIncludeElement, true);
                            parentElement.appendChild(topIncludeElement);
                            topIncludeElement.removeAttribute(ID_ATTR);
                            topIncludeElement.removeAttribute(VALUE_ONLY_ATTR);
                            i--;
                            continue;
                        }
                    } else {
                        throw new ConfigException("Include File must have " + VALUE_ONLY_ATTR + " attribute defined.");
                    }

                    // Case 2: Node Attribute is an id with value_only (set macro)
                } else if (idAttr != null && valueOnlyAttr != null) {
                    String id = idAttr.getValue();
                    if (macros.get(id) != null) {
                        throw new ConfigException("Attempted to redefine existing macro id: " + id + " in node: " + node.getNodeName());
                    }
                    // Store the node away for later retrieval as a macro
                    // Note that this is a shallow not a deep copy
                    macros.put(id, node);
                    // Remove the node because we don't need to reprocess it
                    parentElement.removeChild(node);
                    i--;
                    continue;
                    // Case 3: Node Attribute is an id without value_only (get macro)
                } else if (idAttr != null && valueOnlyAttr == null) {
                    String id = idAttr.getValue();
                    Element macroElement = (Element) macros.get(id);
                    if (macroElement == null) {
                        throw new ConfigException("Attempted to retrieve nonexistent macro id: " + id + " in node: " + node.getNodeName());
                    }

                    if ((valueOnlyAttr = ((Element) macroElement).getAttributeNode(VALUE_ONLY_ATTR)) != null) {
                        boolean valueOnly = Boolean.valueOf(valueOnlyAttr.getValue()).booleanValue();
                        // The Node names must match...
                        if (!node.getNodeName().equals(macroElement.getNodeName())) {
                            throw new ConfigException("Macro Retrievor Element Name: " + node.getNodeName() + "for id: " + id + " did not match Retrieved Element Name: " + macroElement.getNodeName());
                        }
                        // The attribute ids must match...
                        Attr retrievorIdAttr;
                        Attr retrievedIdAttr;

                        if ((retrievedIdAttr = ((Element) macroElement).getAttributeNode(ID_ATTR)) == null) {
                            throw new ConfigException("Macro Retrieved Element Id Attribute for Element: " + macroElement.getNodeName() + " is required.");
                        }
                        if ((retrievorIdAttr = ((Element) node).getAttributeNode(ID_ATTR)) == null) {
                            throw new ConfigException("Macro Retrievor Element Id Attribute for Element: " + node.getNodeName() + " is required.");
                        }

                        if (!retrievorIdAttr.getValue().equals(retrievedIdAttr.getValue())) {
                            throw new ConfigException("Macro Retrievor Element Id (" + retrievorIdAttr.getValue() + ") must match retrieved Element Id (" + retrievedIdAttr.getValue() + ")" + " for retrievor Element name: " + node.getNodeName());
                        }
                        // Then we have to strip away the original node and start processing over
                        // But also, we have to add the included values...
                        if (valueOnly) {
                            // Deep copy with cloneNode(true) before using the data...
                            NodeList macroNodes = macroElement.getChildNodes();
                            Node macroNode;
                            Node nextNode = nodes.item(i + 1);

                            while (macroNodes.getLength() > 0) {
                                Node macroNodeToAdd = macroNodes.item(0);
                                if (macroNodeToAdd.getNodeType() == Node.TEXT_NODE) {
                                    macroNodeToAdd.setNodeValue(macroNodeToAdd.getNodeValue().trim());
                                }
                                if (nextNode == null) {
                                    parentElement.appendChild(macroNodeToAdd);
                                } else {
                                    parentElement.insertBefore(macroNodeToAdd, nextNode);
                                }
                            }
                            // Wait til the end to remove the child or the text elements
                            // will get merged
                            parentElement.removeChild(node);
                            i--;
                            continue;
                            // Else we remove the original node and then replace it with
                            // the new one minus the extra attributes..
                        } else {
                            parentElement.removeChild(node);
                            Element deepCopy = (Element) macroElement.cloneNode(true);
                            parentElement.appendChild(deepCopy);
                            deepCopy.removeAttribute(ID_ATTR);
                            deepCopy.removeAttribute(VALUE_ONLY_ATTR);
                            deepCopy.removeAttribute(INCLUDE_ATTR);
                            i--;
                            continue;
                        }
                    } else {
                        throw new ConfigException("Macro id: " + id + " must have " + VALUE_ONLY_ATTR + " attribute defined.");
                    }

                    // Case 4: Normal element so we recursively process and continue
                } else {
                    processNodes((Element) node, defaultPath, macros, rootDocument);
                }

            }
        }
    } // end of processNodes

    /**
     * Creates a new Properties object referencing a node in a DOM created using
     * one of the public methods.
     */
    private Config(Element element) {
        _element = element;
    }

    public void addChild(Config config) {
        //_element.appendChild(config._element);
        // wrong document exception -- has this never been tested?
        // anyway, I'm fixing it -- RD (March 21 2002)
        Document owner = _element.getOwnerDocument();
        Node toAdd = importNode(owner, config._element, true);
        _element.appendChild(toAdd);
    }

    public Config addChild(String name, String value) {
        Document owner = _element.getOwnerDocument();
        Element newElement = owner.createElement(name);
        newElement.appendChild(owner.createTextNode(value));
        _element.appendChild(newElement);
        return new Config(newElement);
    }

    /**
     * Instantiates an object of type <CODE>#getType()</CODE> and configures it
     * <P>
     * Requires that the type of object
     * <UL>
     *   <LI> Has a default constructor
     * </UL>
     *
     * This method can create either a Configurable instance or a bean instance.
     * The corresponding "bean" method has to contain child nodes that correspond
     * to writable properties of the bean (that is, if a child node names a read-only
     * property of the bean, this method throws a ConfigException).
     *
     * @deprecated This is still used for backward compatibility, use
     * Configurable createAndConfigureInstance() or Object createInstance()
     * instead.
     *
     * @return the constructed object
     */
    public Object create() throws ConfigException {
        Class clazz = null;
        try {
            clazz = Class.forName(getType());
        } catch (ClassNotFoundException cnfe) {
            throw new ConfigException("Unable to find specified class: " + getType());
        }

        // check whether it's a Configurable?
        if (Configurable.class.isAssignableFrom(clazz)) {
            return createAndConfigureInstance();
        } else {
            return createInstance();
        }
    }

    /**
     * Instantiates an object of type <CODE>#getType()</CODE>
     * <P>
     * Requires that the type of object
     * <UL>
     *   <LI> Has a default constructor
     * </UL>
     *
     * This method creates a bean instance. The corresponding "bean" method has
     * to contain child nodes that correspond to writable properties of the bean
     * (that is, if a child node names a read-only property of the bean, this method
     * throws a ConfigException).
     *
     * @return the constructed bean
     */
    public Object createInstance() throws ConfigException {
        try {
            Class clazz = Class.forName(getType());

            // If not, probably this is a bean, get all the properties
            BeanInfo bi = Introspector.getBeanInfo(clazz);
            PropertyDescriptor[] propDescs = bi.getPropertyDescriptors();

            // put 'em all into a Map--make it easier to check against the nodes
            HashMap propsMap = new HashMap();
            for (int i = 0; i < propDescs.length; i++) {
                propsMap.put(propDescs[i].getName(), propDescs[i]);
            }

            // create the instance
            Object bean = clazz.newInstance();

            // Get all the children of this config node,
            Config[] children = getChildren();
            for (int i = 0; i < children.length; i++) {
                // use decapitalize cos the property list entries
                // start with lowercase, while we usually have
                // capitalize the first letter of a Config node's name.
                String childName = Introspector.decapitalize(children[i].getName());

                // if a node is not among the properties--throw an exception
                if (!propsMap.containsKey(childName)) {
                    // invalid node! throw exception
                    throw new ConfigException("The following child node: " + children[i].getName() + " under this node: " + getName() + " is invalid because it doesn't correspond to any" + " property of this bean: " + getType());

                } else {
                    // get the setter method, if null, then the property is read-only.
                    Method method = ((PropertyDescriptor) propsMap.get(childName)).getWriteMethod();
                    if (method == null) {
                        throw new ConfigException("The property " + childName + " of bean: " + getType() + " is read-only.");
                    }

                    // first get the type of the child node
                    String type = children[i].getType();

                    // now get the type from the setter method
                    Class[] paramTypes = method.getParameterTypes();
                    // get the param type
                    String paramTypeName = paramTypes[0].getName();

                    // check the type! if there's no type attribute,
                    // just set it, but if there IS, ensure that it
                    // matches the type of the respective property.
                    if (type != null) {
                        // if they are not equal--throw ConfigException
                        if (!type.equals(paramTypeName)) {
                            throw new ConfigException("Incorrect type for property " + childName + " of bean: " + getType() + ", should be " + paramTypeName + " instead of " + type);
                        }
                    }

                    // now call the setter method, but before that
                    // convert the parameter from String to
                    // something if possible--that is, get the
                    // constructor of the property type which accepts
                    // a String and call it. That's why, if the type
                    // is a primitive type, it's replaced by the type
                    // of its wrapper.
                    if (paramTypes[0].isPrimitive()) {
                        paramTypes[0] = paramTypes[0].equals(Boolean.TYPE) ? Boolean.class : paramTypes[0].equals(Character.TYPE) ? Character.class : paramTypes[0].equals(Byte.TYPE) ? Byte.class : paramTypes[0].equals(Short.TYPE) ? Short.class : paramTypes[0].equals(Integer.TYPE) ? Integer.class : paramTypes[0].equals(Long.TYPE) ? Long.class : paramTypes[0].equals(Float.TYPE) ? Float.class : paramTypes[0].equals(Double.TYPE) ? Double.class : Void.class;
                    }
                    Constructor constructor = paramTypes[0].getConstructor(new Class[] { String.class });
                    Object param = constructor.newInstance(new Object[] { children[i].getValue()});
                    method.invoke(bean, new Object[] { param });
                }
            }
            return bean;
        } catch (IntrospectionException ipe) {
            throw new ConfigException("Unable to introspect object: " + getType());
        } catch (InvocationTargetException ite) {
            Throwable ex = ite.getTargetException();
            String message = (ex != null) ? ex.getMessage() : "N/A";
            throw new ConfigException("This exception occurred while configuring this bean: " + getType() + ", with the following message: " + message);
        } catch (NoSuchMethodException nsme) {
            throw new ConfigException("At least one property cannot be set using a String for bean: " + getType());
        } catch (InstantiationException ie) {
            throw new ConfigException("Failed to instantiate the specified object: " + getType());
        } catch (IllegalAccessException iae) {
            throw new ConfigException("Illegal access to specified class: " + getType() + ": " + iae.getMessage());
        } catch (ClassNotFoundException cnfe) {
            cnfe.printStackTrace();
            throw new ConfigException("Unable to find specified class: " + getType());
        }
    }

    /**
     * Instantiates an object of type <CODE>#getType()</CODE> and configures it
     * <P>
     * Requires that the type of object implements Configurable
     * <UL>
     *   <LI> Has a default constructor
     * </UL>
     *
     * This method creates a Configurable instance and configure it.
     *
     * @return the constructed and fully configured Configurable instance
     */
    public Configurable createAndConfigureInstance() throws ConfigException {
        try {
            Class clazz = Class.forName(getType());
            Configurable configurable = (Configurable) clazz.newInstance();
            configurable.configure(this);
            return configurable;
        } catch (InstantiationException ie) {
            throw new ConfigException("Failed to instantiate the specified object: " + getType());
        } catch (IllegalAccessException iae) {
            throw new ConfigException("Illegal access to specified class: " + getType() + ": " + iae.getMessage());
        } catch (ClassNotFoundException cnfe) {
            throw new ConfigException("Unable to find specified class: " + getType());
        }
    }

    /**
     * @return The FIRST Config tag matching the specified key
     */
    public Config getChild(String key) {
        NodeList nodes = _element.getChildNodes();
        Node node;
        for (int i = 0; i < nodes.getLength(); i++) {
            node = nodes.item(i);
            if (isConfigNode(node) && node.getNodeName().equals(key)) {
                return new Config((Element) node);
            }
        }
        return null;
    }

    /**
     * @return An iterator over the child Properties objects.
     */
    public Config[] getChildren() {
        NodeList nodes = _element.getChildNodes();
        Node node;
        ArrayList childrenList = new ArrayList();
        for (int i = 0; i < nodes.getLength(); i++) {
            node = nodes.item(i);
            if (isConfigNode(node)) {
                childrenList.add(new Config((Element) node));
            }
        }
        //now copy the ArrayList into a Config []
        Object[] objs = childrenList.toArray();
        Config[] children = new Config[objs.length];
        System.arraycopy(objs, 0, children, 0, objs.length);
        return children;
    }

    public String getName() {
        return _element.getNodeName();
    }

    private String getText(Element element, boolean trim) {
        // First consider whether we are in a CDATA section
        Node node = element.getFirstChild();

        if (node != null && node.getNodeType() == Node.CDATA_SECTION_NODE) {
            return node.getNodeValue();
        }

        // Else work with appending normal text nodes
        // together as one value...
        NodeList nodes = element.getChildNodes();
        String res = "";
        for (int j = 0; j < nodes.getLength(); j++) {
            node = nodes.item(j);
            if (node.getNodeType() == Node.TEXT_NODE) {
                res += node.getNodeValue();
            }
        }
        if (trim)
            res = res.trim();
        return res.equals("") ? null : res;
    }

    /**
     * @return Returns the type of the property if the type attribute is specified in the XML
     */
    public String getType() {
        String type = _element.getAttribute(TYPE_ATTR);
        return type.equals("") ? null : type;
    }

    public String getValue() {
        return getText(_element, true);
    }

    public boolean getBooleanValue() {
        return Boolean.valueOf(getValue()).booleanValue();
    }

    public int getIntValue() throws ConfigException {
        try {
            return Integer.parseInt(getValue());
        } catch (NumberFormatException nfe) {
            throw new ConfigException(getName() + " should be an integer");
        }
    }

    public String getRealValue() {
        return getText(_element, false);
    }

    public String getChildValue(String child) throws ConfigException {
        return getChildValue(child, true);
    }

    public String getChildValue(String child, boolean required) throws ConfigException {
        Config childConfig = getChild(child);
        String value = null;

        if (required && childConfig == null)
            throw new ConfigException("No <" + child + "> specifed in " + getName() + " config");

        if (childConfig != null) {
            value = childConfig.getValue();
            if (required && value == null)
                throw new ConfigException("No <" + child + "> value specifed in " + getName() + " config");
        }

        return value;
    }

    public boolean hasChild(String key) {
        return !(getChild(key) == null);
    }

    public boolean hasChildren() {
        NodeList nodes = _element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++) {
            if (isConfigNode(nodes.item(i)))
                return true;
        }
        return false;
    }

    private boolean isConfigNode(Node node) {
        return node.getNodeType() == Node.ELEMENT_NODE;
    }

    public void removeChild(Config oneToDelete) {
        _element.removeChild(oneToDelete._element);
    }

    public void save(OutputStream os) {
        save(new OutputStreamWriter(os));
    }

    public void save(Writer w) {
        try {
            TransformerFactory transformerFactory = TransformerFactory.newInstance();
            Transformer transformer = transformerFactory.newTransformer();

            DOMSource source = new DOMSource(_element);
            StreamResult result = new StreamResult(w);
            transformer.transform(source, result);

        } catch (TransformerConfigurationException e) {
            e.printStackTrace();
        } catch (TransformerException e) {
            e.printStackTrace();
        }
    }

    private void setAttribute(String attr, String value) {
        //get out the attribute and setValue
        if (value == null || value.equals("")) {
            if (_element.getAttributeNode(attr) != null) {
                _element.removeAttribute(attr);
            }
        } else {
            _element.setAttribute(attr, value);
        }
    }

    public void setType(String type) {
        if (type == null) {
            _element.removeAttribute(TYPE_ATTR);
        } else {
            _element.setAttribute(TYPE_ATTR, type);
        }
    }

    public void setValue(String value) {
        //remove all the old text nodes
        NodeList nodes = _element.getChildNodes();
        Node node;
        for (int i = 0; i < nodes.getLength(); i++) {
            node = nodes.item(i);
            if (node.getNodeType() == Node.TEXT_NODE) {
                _element.removeChild(node);
            }
        }
        Document owner = _element.getOwnerDocument();

        _element.appendChild(owner.createTextNode(value));
    }

    /**
     * @return The Config object expressed as a Properties object
     */
    public Properties toProperties() {
        Properties properties = new Properties();
        Config[] children = getChildren();
        for (int i = 0; i < children.length; i++) {
            properties.put(children[i].getName(), children[i].getValue());
        }
        return properties;
    }

    public String toString() {
        return getName();
    }

    // The following importNode routines were adapted from a recent version
    // of org.w3c.dom.CoreDocumentImpl. The reason for taking the code
    // and adapting it here is to make sure we are backwards compatible
    // with older XML parsers that do not support DOM level 2 properly
    //
    protected Node importNode(Document document, Node source, boolean deep) throws DOMException {
        return importNode(document, source, deep, null);
    }

    private Node importNode(Document document, Node source, boolean deep, HashMap reversedIdentifiers) throws DOMException {
        Node newnode = null;

        int type = source.getNodeType();

        switch (type) {
            case Document.ELEMENT_NODE :
                {
                    Element newElement;
                    boolean domLevel20 = source.getOwnerDocument().getImplementation().hasFeature("XML", "2.0");
                    // Create element according to namespace support/qualification.
                    newElement = document.createElement(source.getNodeName());

                    // Copy element's attributes, if any.
                    NamedNodeMap sourceAttrs = source.getAttributes();
                    if (sourceAttrs != null) {
                        int length = sourceAttrs.getLength();
                        for (int index = 0; index < length; index++) {
                            Attr attr = (Attr) sourceAttrs.item(index);

                            // Copy the attribute only if it is not a default.
                            if (attr.getSpecified()) {
                                Attr newAttr = (Attr) importNode(document, attr, true, reversedIdentifiers);

                                // Attach attribute according to namespace
                                // support/qualification.
                                newElement.setAttributeNode(newAttr);
                            }
                        }
                    }

                    // Register element identifier.
                    if (reversedIdentifiers != null) {
                        // Does element have an associated identifier?
                        Object elementId = reversedIdentifiers.get(source);
                    }

                    newnode = newElement;
                    break;
                }

            case Document.ATTRIBUTE_NODE :
                {

                    newnode = document.createAttribute(source.getNodeName());
                    // if source is an AttrImpl from this very same implementation
                    // avoid creating the child nodes if possible
                    // According to the DOM spec the kids carry the value.
                    // However, there are non compliant implementations out
                    // there that fail to do so. To avoid ending up with no
                    // value at all, in this case we simply copy the text value
                    // directly.
                    if (source.getFirstChild() == null) {
                        newnode.setNodeValue(source.getNodeValue());
                        deep = false;
                    } else {
                        deep = true;
                    }
                    break;
                }

            case Document.TEXT_NODE :
                {
                    newnode = document.createTextNode(source.getNodeValue());
                    break;
                }

            case Document.CDATA_SECTION_NODE :
                {
                    newnode = document.createCDATASection(source.getNodeValue());
                    break;
                }

            case Document.ENTITY_REFERENCE_NODE :
                {
                    break;
                }

            case Document.ENTITY_NODE :
                {
                    // We don't care about copying entity nodes actually
                    break;
                }

            case Document.PROCESSING_INSTRUCTION_NODE :
                {
                    newnode = document.createProcessingInstruction(source.getNodeName(), source.getNodeValue());
                    break;
                }

            case Document.COMMENT_NODE :
                {
                    newnode = document.createComment(source.getNodeValue());
                    break;
                }

            case Document.DOCUMENT_FRAGMENT_NODE :
                {
                    newnode = document.createDocumentFragment();
                    // No name, kids carry value
                    break;
                }

            case Document.NOTATION_NODE :
                {
                    // eXtropia doesn't care about a Notation node copy
                    break;
                }
            case Document.DOCUMENT_TYPE_NODE :
            case Document.DOCUMENT_NODE : // Can't import document nodes
            default :
                { // Unknown node type
                    throw new RuntimeException("Node type being imported is not supported");
                }
        }

        // If deep, replicate and attach the kids.
        if (deep) {
            for (Node srckid = source.getFirstChild(); srckid != null; srckid = srckid.getNextSibling()) {
                newnode.appendChild(importNode(document, srckid, true, reversedIdentifiers));
            }
        }
        return newnode;

    }
}
