/*
 * Rule.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.datahandler;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;

/**
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 1.0
 */
public class Rule implements Configurable {
    public static final String CFG_BASE = "Rule";
    public static final String CFG_MAX_ERRORS = "MaxErrors";
    public static final String CFG_OVERRRIDE_ERRORS = "OverrideErrors";

    private List _rules;
    private List _dataHandlers;

    private String _name;
    private int _maxErrors = 0;
    private boolean _overrideErrors = false;

    public Rule() {
    }

    public Rule(String name) {
        setName(name);
    }

    /**
     * Adds a datahandler to this rule
     * <P>
     * When validating / transforming this datahandler will be executed in the
     * order that it was added to the <CODE>Rule</CODE>
     */
    public void add(DataHandler dataHandler) {
        if (_dataHandlers == null) {
            _dataHandlers = new ArrayList();
        }
        _dataHandlers.add(dataHandler);
    }

    /**
     * Adds a datahandler to this rule
     * <P>
     * This rule will be executed in the order that it was added
     * to the <CODE>Rule</CODE>
     */
    public void add(Rule rule) {
        if (_rules == null) {
            _rules = new ArrayList();
        }
        _rules.add(rule);
    }

    public void configure(Config config) throws ConfigException {
        if (!config.getName().equals(CFG_BASE))
            throw new ConfigException("Incorrect config passed to " + CFG_BASE + ": " + config.getName());

        if (config.hasChild(CFG_OVERRRIDE_ERRORS)) {
            setOverrideErrors(Boolean.valueOf(config.getChild(CFG_OVERRRIDE_ERRORS).getValue()).booleanValue());
        }

        if (config.hasChild(CFG_MAX_ERRORS)) {
            try {
                setMaxErrors(Integer.parseInt(config.getChild(CFG_MAX_ERRORS).getValue()));
            } catch (NumberFormatException nfe) {
                throw new ConfigException("Invalid parameter in " + CFG_MAX_ERRORS + ": " + config.getChild(CFG_MAX_ERRORS).getValue());
            }
        }

        setName(config.getValue());

        Config[] children = config.getChildren();
        for (int i = 0; i < children.length; i++) {
            if (children[i].getName().equals(DataHandler.CFG_BASE)) {
                add((DataHandler) children[i].createAndConfigureInstance());
            } else if (children[i].getName().equals(CFG_BASE)) {
                add((Rule) children[i].createAndConfigureInstance());
            }
        }
    }

    public void execute(Map data) throws DataHandlerException, IllegalStateException {
        DataHandlerException failure = new DataHandlerException(_name);

        boolean executed = false;
        if (_rules != null)
            executeSubRules(data, failure);
        if (_dataHandlers != null)
            executeDataHandlers(data, failure);

        //will reach here if we got errors but aren't overridding the error message
        //or didn't reach the _maxError limit if set
        if (failure.hasChildren())
            throw failure;
    }

    protected void executeSubRules(Map data, DataHandlerException failure) throws DataHandlerException {
        for (Iterator ruleIterator = _rules.iterator(); ruleIterator.hasNext();) {
            try {
                ((Rule) ruleIterator.next()).execute(data);
            } catch (DataHandlerException dhe) {
                if (_overrideErrors) //we can throw as soon as we know we have an error
                    throw failure;
                else {
                    failure.addChild(dhe);
                    if (_maxErrors > 0 & failure.getChildren().length >= _maxErrors)
                        throw failure;
                }
            }
        }
    }

    protected void executeDataHandlers(Map data, DataHandlerException failure) throws DataHandlerException {
        DataHandler handler;
        String[] keys;
        Object[] values;
        Object transformedValue;
        Object id;

        for (Iterator dhIterator = _dataHandlers.iterator(); dhIterator.hasNext();) {
            handler = (DataHandler) dhIterator.next();
            keys = handler.getIds();
            values = new Object[keys.length];
            //Anderson
            Object[] orgValues = new Object[keys.length];

            for (int i = 0; i < keys.length; i++) {
                values[i] = data.get(keys[i]);
                //Anderson - make another copy
                orgValues[i] = data.get(keys[i]);
            }

            try {
                transformedValue = handler.validate(values); //throws DataHandlerException
                if (transformedValue != null) {
                    if (transformedValue instanceof Null)
                        data.remove(handler.getTransformId());
                    else
                        data.put(handler.getTransformId(), transformedValue);
                }
                //update the map & transform id
                for (int i = 0; i < values.length; i++) {
                    //if(values[i] == null){
                    if (values[i] == null && orgValues[i] != null) {
                        //Anderson - only remove the entry if the handler decides to change it from something to null.
                        //This might not be the best approach to solve the problem of 'clearing out an editable field during
                        //the sql update'.
                        //Some further consideration is needed in the future
                        data.remove(keys[i]);
                    } else
                        data.put(keys[i], values[i]);
                }
                if (transformedValue != null && handler.getTransformId() != null)
                    data.put(handler.getTransformId(), transformedValue);
            } catch (DataHandlerException dhe) {
                if (_overrideErrors) //we can throw as soon as we know we have an error
                    throw failure;
                else {
                    failure.addChild(dhe);
                    if (_maxErrors > 0 & failure.getChildren().length >= _maxErrors)
                        throw failure;
                }
            }
        }
    }

    public String getName() {
        return _name;
    }

    public DataHandler[] getDataHandlers() {
        if (_dataHandlers != null)
            return (DataHandler[]) _dataHandlers.toArray(new DataHandler[_dataHandlers.size()]);
        else
            return null;
    }

    public Rule[] getRules() {
        if (_rules != null)
            return (Rule[]) _rules.toArray(new Rule[_rules.size()]);
        else
            return null;
    }

    public boolean removeHandler(String name) {
        return remove(_dataHandlers, name);
    }

    public boolean removeRule(String name) {
        return remove(_rules, name);
    }

    protected boolean remove(List list, String name) {
        DataHandler handler;
        if (list != null) {
            for (Iterator iter = list.iterator(); iter.hasNext();) {
                handler = (DataHandler) iter.next();
                if (handler.getName().equals(name))
                    iter.remove();
            }
        }
        if ((_dataHandlers == null || _dataHandlers.size() == 0) && (_rules == null || _rules.size() == 0))
            return false;
        else
            return true;
    }

    public void setName(String name) {
        _name = name;
    }

    public void setMaxErrors(int maxErrors) {
        _maxErrors = maxErrors;
    }

    public void setOverrideErrors(boolean overrideErrors) {
        _overrideErrors = overrideErrors;
    }

    public void setReportAllErrors() {
        _maxErrors = 0;
    }
}
