/* 
 * DateTimeHandler.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.datahandler.handler;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.datahandler.DataHandler;
import com.eXtropia.datahandler.DataHandlerException;

/** 
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 1.0
 */
public class DateTimeHandler extends DataHandler {
    public static final String NAME = "date";

    public static final String ERROR_INVALID_MINUTE = "InvalidMinute";
    public static final String ERROR_INVALID_HOUR = "InvalidHour";
    public static final String ERROR_INVALID_DATE = "InvalidDate";
    public static final String ERROR_INVALID_MONTH = "InvalidMonth";
    public static final String ERROR_INVALID_YEAR = "InvalidYear";

    public static final int TYPE_DATE = 1;
    public static final int TYPE_DATE_TIME = 0;

    public static final String CFG_TYPE = "Type";
    public static final String CFG_TYPE_DATE = "date";
    public static final String CFG_TYPE_TIME = "time";
    public static final String CFG_RETURN_FORMAT = "TransformFormat";

    public static final int YEAR_INDEX = 0;
    public static final int MONTH_INDEX = 1;
    public static final int DATE_INDEX = 2;
    public static final int HOUR_INDEX = 3;
    public static final int MINUTE_INDEX = 4;

    protected int _type = TYPE_DATE_TIME;
    protected DateFormat _returnFormat;

    public DateTimeHandler() {
        super(NAME);
    }

    public DateTimeHandler(int type) {
        super(NAME);
        setType(type);
    }

    public void configure(Config config) throws ConfigException {
        if (config.hasChild(CFG_TYPE)) {
            String type = config.getChild(CFG_TYPE).getValue().toLowerCase();
            if (type.equals(CFG_TYPE_DATE))
                setType(TYPE_DATE);
            else
                setType(TYPE_DATE_TIME);
        }

        if (config.hasChild(CFG_RETURN_FORMAT))
            setReturnFormat(new SimpleDateFormat(config.getChild(CFG_RETURN_FORMAT).getValue()));
        super.configure(config);
    }

    public void setIds(String[] ids) {
        if (_type == TYPE_DATE_TIME && ids.length == 5)
            super.setIds(ids);
        else if (_type == TYPE_DATE && ids.length == 3)
            super.setIds(ids);
        else
            throw new IllegalArgumentException("Incorrect number of ids passed to DateTimeHandler: " + ids.length);
    }

    public void setType(int type) {
        _type = type;
    }

    public void setReturnFormat(DateFormat returnFormat) {
        _returnFormat = returnFormat;
    }

    public Object validate(Object[] values) throws DataHandlerException {
        //if all three are empty: we have no date
        /*
        System.out.println("date: [" + _name + "]");
        System.out.println("  hour: [" + values[HOUR_INDEX] + "]");
        System.out.println("  minute: [" + values[MINUTE_INDEX] + "]");
        System.out.println("  date: [" + values[DATE_INDEX] + "]");
        System.out.println("  month: [" + values[MONTH_INDEX] + "]");
        System.out.println("  year: [" + values[YEAR_INDEX] + "]");
        */

        if ((values[DATE_INDEX] == null || values[DATE_INDEX].toString().length() == 0) && (values[MONTH_INDEX] == null || values[MONTH_INDEX].toString().length() == 0) && (values[YEAR_INDEX] == null || values[YEAR_INDEX].toString().length() == 0))
            return null;

        if (_type == TYPE_DATE_TIME && (values[HOUR_INDEX] == null || values[HOUR_INDEX].toString().length() == 0) && (values[MINUTE_INDEX] == null || values[MINUTE_INDEX].toString().length() == 0))
            return null;

        int date = -1, month = -1, year = -1, hour = -1, minute = -1;
        try {
            if (_type == TYPE_DATE_TIME) {
                minute = Integer.parseInt(values[MINUTE_INDEX].toString());
                hour = Integer.parseInt(values[HOUR_INDEX].toString());
            }
            date = Integer.parseInt(values[DATE_INDEX].toString());
            month = Integer.parseInt(values[MONTH_INDEX].toString());
            year = Integer.parseInt(values[YEAR_INDEX].toString());
        } catch (Exception e) {
            throw new DataHandlerException(_name);
        }
        month--;
        if (year < 0)
            throw new DataHandlerException(_name, ERROR_INVALID_YEAR, _ids[YEAR_INDEX], values[YEAR_INDEX]);

        if (month < 0 || month > 12)
            throw new DataHandlerException(_name, ERROR_INVALID_MONTH, _ids[MONTH_INDEX], values[MONTH_INDEX]);

        if (date < 0 || date > 31)
            throw new DataHandlerException(_name, ERROR_INVALID_DATE, _ids[DATE_INDEX], values[DATE_INDEX]);

        if (_type == TYPE_DATE_TIME) {
            if (hour < 0 || hour > 23)
                throw new DataHandlerException(_name, ERROR_INVALID_HOUR, _ids[HOUR_INDEX], values[HOUR_INDEX]);

            if (minute < 0 || minute > 59)
                throw new DataHandlerException(_name, ERROR_INVALID_MINUTE, _ids[MINUTE_INDEX], values[MINUTE_INDEX]);
        }

        Calendar cal = Calendar.getInstance();
        cal.clear();

        cal.set(Calendar.YEAR, year);
        if (cal.get(Calendar.YEAR) != year)
            throw new DataHandlerException(_name, ERROR_INVALID_YEAR, _ids[YEAR_INDEX], values[YEAR_INDEX]);

        cal.set(Calendar.MONTH, month);
        if (cal.get(Calendar.MONTH) != month)
            throw new DataHandlerException(_name, ERROR_INVALID_MONTH, _ids[MONTH_INDEX], values[MONTH_INDEX]);

        cal.set(Calendar.DATE, date);
        if (cal.get(Calendar.DATE) != date)
            throw new DataHandlerException(_name, ERROR_INVALID_DATE, _ids[DATE_INDEX], values[DATE_INDEX]);

        if (_type == TYPE_DATE_TIME) {
            cal.set(Calendar.HOUR_OF_DAY, hour);
            if (cal.get(Calendar.HOUR_OF_DAY) != hour) {
                throw new DataHandlerException(_name, ERROR_INVALID_HOUR, _ids[HOUR_INDEX], values[HOUR_INDEX]);
            }

            cal.set(Calendar.MINUTE, minute);
            if (cal.get(Calendar.MINUTE) != minute)
                throw new DataHandlerException(_name, ERROR_INVALID_MINUTE, _ids[MINUTE_INDEX], values[MINUTE_INDEX]);
        }

        for (int i = 0; i < values.length; i++)
            values[i] = null;
        if (_returnFormat == null)
            return cal.getTime(); //return the date itself
        else
            return _returnFormat.format(cal.getTime()); //return a formatted string
    }
}