/* 
 * DynamicDateRangeHandler.java
 * Copyright (C) 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.datahandler.handler;

import java.util.Calendar;
import java.util.Date;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.datahandler.DataHandler;
import com.eXtropia.datahandler.DataHandlerException;

/** 
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 1.0
 */
public class DynamicDateRangeHandler extends DataHandler {
    public static final String ERROR_TOO_EARLY = "TooEarly";
    public static final String ERROR_TOO_LATE = "TooLate";

    public static final String CFG_START_RANGE_DAYS = "StartDays";
    public static final String CFG_START_RANGE_MONTHS = "StartMonths";
    public static final String CFG_START_RANGE_YEARS = "StartYears";
    public static final String CFG_END_RANGE_DAYS = "EndDays";
    public static final String CFG_END_RANGE_MONTHS = "EndMonths";
    public static final String CFG_END_RANGE_YEARS = "EndYears";

    protected int _startDays, _startMonths, _startYears, _endDays, _endMonths, _endYears;
    protected boolean _checkStartDate = false, _checkEndDate = false;

    public DynamicDateRangeHandler() {
        super("DateRange");
    }

    public void configure(Config config) throws ConfigException {
        super.configure(config);
        if (config.hasChild(CFG_START_RANGE_DAYS)) {
            _startDays = config.getChild(CFG_START_RANGE_DAYS).getIntValue();
            _checkStartDate = true;
        }

        if (config.hasChild(CFG_START_RANGE_MONTHS)) {
            _startMonths = config.getChild(CFG_START_RANGE_MONTHS).getIntValue();
            _checkStartDate = true;
        }

        if (config.hasChild(CFG_START_RANGE_YEARS)) {
            _startYears = config.getChild(CFG_START_RANGE_YEARS).getIntValue();
            _checkStartDate = true;
        }

        if (config.hasChild(CFG_END_RANGE_DAYS)) {
            _endDays = config.getChild(CFG_END_RANGE_DAYS).getIntValue();
            _checkEndDate = true;
        }

        if (config.hasChild(CFG_END_RANGE_MONTHS)) {
            _endMonths = config.getChild(CFG_END_RANGE_MONTHS).getIntValue();
            _checkEndDate = true;
        }

        if (config.hasChild(CFG_END_RANGE_YEARS)) {
            _endYears = config.getChild(CFG_END_RANGE_YEARS).getIntValue();
            _checkEndDate = true;
        }

        if (!(_checkStartDate || _checkEndDate)) {
            throw new ConfigException();
        }
    }

    public Object validate(Object[] values) throws DataHandlerException {
        Calendar cal = Calendar.getInstance();
        Date currentDate = cal.getTime();
        Date windowStart = null, windowEnd = null;
        Date dateToCheck;

        if (_checkStartDate) {
            //generate comparison date
            cal.add(Calendar.YEAR, _startYears);
            cal.add(Calendar.MONTH, _startMonths);
            cal.add(Calendar.DATE, _startDays);
            windowStart = cal.getTime();
        }

        if (_checkEndDate) {
            //generate comparison date
            cal.setTime(currentDate);
            cal.add(Calendar.YEAR, _endYears);
            cal.add(Calendar.MONTH, _endMonths);
            cal.add(Calendar.DATE, _endDays);
            windowEnd = cal.getTime();
        }

        for (int i = 0; i < values.length; i++) {
            if (values[i] != null) {
                dateToCheck = (Date) values[i];
                if (_checkStartDate && dateToCheck.compareTo(windowStart) == -1)
                    throw new DataHandlerException(_name, ERROR_TOO_EARLY, _ids[i], values[i]);
                if (_checkEndDate && dateToCheck.compareTo(windowEnd) == 1)
                    throw new DataHandlerException(_name, ERROR_TOO_LATE, _ids[i], values[i]);
            }
        }
        return null;
    }

    public void setWindowStart(int days, int months, int years) {
        _checkStartDate = true;
        _startDays = days;
        _startMonths = months;
        _startYears = years;
    }

    public void setWindowEnd(int days, int months, int years) {
        _checkEndDate = true;
        _endDays = days;
        _endMonths = months;
        _endYears = years;
    }
}