/**
 * RegexParser.java
 * Copyright (C) 2000 -2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.datahandler.handler;

import java.util.ArrayList;
import java.util.List;

import org.apache.oro.text.regex.MalformedPatternException;
import org.apache.oro.text.regex.MatchResult;
import org.apache.oro.text.regex.Pattern;
import org.apache.oro.text.regex.PatternCompiler;
import org.apache.oro.text.regex.PatternMatcher;
import org.apache.oro.text.regex.PatternMatcherInput;
import org.apache.oro.text.regex.Perl5Compiler;
import org.apache.oro.text.regex.Perl5Matcher;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;

/**
 * <code>RegexParser</code>
 * Created: Thu Dec 13 11:11:57 2001
 *
 * @author <a href="mailto: st_lim@eXtropia.com">Lim Swee Tat</a>
 * @version $Id: RegexParser.java,v 1.4 2002/11/16 08:45:11 rayfd Exp $
 */
public class RegexParser implements Configurable {
    public static final String CFG_BASE = "RegexParser";
    public static final String CFG_REGEX = "Regex";
    public static final String CFG_EXACT_MATCH = "ExactMatch";

    public static final String CFG_SUBEXPRESSION_MANY = "Subexpressions";
    public static final String CFG_SUBEXPRESSION = "Subexpression";
    public static final String CFG_SUBEXPRESSION_NAME = "Name";

    private String _regexString;
    private Pattern pattern = null;
    private boolean exactMatch = false;

    private String[] _subexpressionNames;
    private RegexParser[] _subexpressions;

    /**
     * Creates a new <code>RegexParser</code> instance.
     *
     */
    public RegexParser() {
    }

    /**
     * <code>configure</code> configures the concrete implementation
     * of the parser.
     *
     * @param config a <code>Config</code> value
     * @exception ConfigException if an error occurs
     */
    public void configure(Config config) throws ConfigException {

        // Checks if the soap connection is available
        if (!config.getName().equals(CFG_BASE))
            throw new ConfigException("Incorrect Config passed to " + CFG_BASE + ": " + config.getName());
        try {
            _regexString = config.getValue();
        } catch (Exception e) {
            throw new ConfigException("Failed to get regular expression for " + CFG_BASE + " : " + e.getMessage());
        }

        if (config.hasChild(CFG_EXACT_MATCH)) {
            exactMatch = (boolean) config.getChild(CFG_EXACT_MATCH).getBooleanValue();
        }

        // Create Perl5Compiler and Perl5Matcher instances.
        PatternCompiler compiler = new Perl5Compiler();

        // Attempt to compile the pattern.  If the pattern is not valid,
        // report the error and exit.
        try {
            pattern = compiler.compile(_regexString);
        } catch (MalformedPatternException e) {
            throw new ConfigException("Failed to create a regular expression on the parser.");
        }

        // Try to create the subexpressions that are in this expression
        Config[] subexpressionConfigs;

        // This should only be configured if there are subexpressions
        // Else the system should just use numbers, which is extremely kludgy
        if (config.hasChild(CFG_SUBEXPRESSION_MANY)) {
            subexpressionConfigs = config.getChild(CFG_SUBEXPRESSION_MANY).getChildren();

            // Use of array vs HashMap since the order of the subexpression in a
            // regex is ordered.
            _subexpressionNames = new String[subexpressionConfigs.length];
            _subexpressions = new RegexParser[subexpressionConfigs.length];

            for (int i = 0; i < subexpressionConfigs.length; i++) {
                if (subexpressionConfigs[i].getName().equals(CFG_SUBEXPRESSION)) {
                    Config subexpressionConfig = subexpressionConfigs[i];
                    _subexpressionNames[i] = subexpressionConfig.getChild(CFG_SUBEXPRESSION_NAME).getValue();
                    if (subexpressionConfig.hasChild(CFG_BASE)) {
                        _subexpressions[i] = (RegexParser) subexpressionConfig.getChild(CFG_BASE).createAndConfigureInstance();
                    } else
                        _subexpressions[i] = null;
                }
            }
        }
    }

    /**
     * <code>match</code> is just used to check if the String in is a match
     * for the regular expression.
     *
     * @param in a <code>String</code> value
     * @return a <code>boolean</code> value
     */
    public boolean match(String in) {
        PatternMatcher matcher = new Perl5Matcher();

        if (matcher.matches(in, pattern))
            return true;
        return false;
    }

    /**
     * Describe <code>parse</code> method here.
     *
     * @param in a <code>String</code> value
     * @param name a <code>String</code> value
     * @return an <code>String[]</code> value
     */
    public String[] parse(String in, String name) {
        int i;
        for (i = 0;(i < _subexpressionNames.length) && (!_subexpressionNames[i].equals(name)); i++);

        // Need to add one since the match is 0, then those in brackets are 1, 2...
        return parse(in, i + 1);
    }

    /**
     * Describe <code>parse</code> method here.
     *
     * @param in a <code>String</code> value
     * @param group an <code>int</code> value
     * @return an <code>String[]</code> value
     */
    public String[] parse(String in, int group) {
        List result = new ArrayList();
        PatternMatcherInput input = new PatternMatcherInput(in);
        PatternMatcher matcher = new Perl5Matcher();

        if (matcher.matches(in, pattern)) {
            MatchResult matchResult = matcher.getMatch();
            result.add(matchResult.group(group));
        }
        return (String[]) result.toArray(new String[0]);
    }

    /**
     * <code>parse</code> is used to break the input into multiple expressions.
     *
     * @param in a <code>String</code> value
     * @return an <code>String[]</code> value
     */
    public String[] parse(String in) {
        ArrayList result = new ArrayList();
        PatternMatcherInput input = new PatternMatcherInput(in);
        PatternMatcher matcher = new Perl5Matcher();

        if (exactMatch) {
            if (matcher.matches(input, pattern)) {
                MatchResult matchResult = matcher.getMatch();
                result = getGroups(result, matchResult);
            }
        } else {
            while (matcher.contains(input, pattern)) {
                MatchResult matchResult = matcher.getMatch();
                result = getGroups(result, matchResult);
            }
        }
        return (String[]) result.toArray(new String[0]);
    }

    /**
     * Describe <code>getSubexpression</code> method here.
     *
     * @param name a <code>String</code> value
     * @return a <code>RegexParser</code> value
     */
    public RegexParser getSubexpression(String name) {
        int i;
        for (i = 0;(i < _subexpressionNames.length) && (!_subexpressionNames[i].equals(name)); i++);

        return getSubexpression(i);
    }

    /**
     * Describe <code>getSubexpression</code> method here.
     *
     * @param group an <code>int</code> value
     * @return a <code>RegexParser</code> value
     */
    public RegexParser getSubexpression(int group) {
        return _subexpressions[group];
    }

    /**
     * The <code>actualParser</code> method is oro specific and
     * may be removed if updated to 1.4.
     *
     * @param list an <code>ArrayList</code> value
     * @param matchResult a <code>MatchResult</code> value
     * @return an <code>ArrayList</code> value
     */
    private ArrayList getGroups(ArrayList list, MatchResult matchResult) {
        int groups = matchResult.groups();
        for (int group = 1; group < groups; group++) {
            list.add(matchResult.group(group));
        }
        return list;
    }

    /**
     * Describe <code>getRegexString</code> method here.
     *
     * @return a <code>String</code> value
     */
    public String getRegexString() {
        return _regexString;
    }

    /**
     * <code>getSubexpressionNames</code> gets the names of the subexpressions.
     *
     * @return a <code>String[]</code> value
     */
    public String[] getSubexpressionNames() {
        return _subexpressionNames;
    }

    /**
     * <code>getSubexpressionParser</code> returns the parser for the subexpression "group".
     *
     * @param group an <code>int</code> value
     * @return a <code>RegexParser</code> value
     */
    public RegexParser getSubexpressionParser(int group) {
        return _subexpressions[group];
    }

} // RegexParser
