/* 
 * RequiredHandler.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.datahandler.handler;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.datahandler.DataHandler;
import com.eXtropia.datahandler.DataHandlerException;

/** 
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 1.0
 */
public class RequiredHandler extends DataHandler {
    public static final String NAME = "required";

    public static final String CFG_REQUIRE_AT_LEAST = "AtLeast";
    public static final String CFG_REQUIRE_AT_MOST = "AtMost";

    public static final String ERROR_REQUIRE_AT_LEAST = "AtLeast";
    public static final String ERROR_REQUIRE_AT_MOST = "AtMost";
    public static final String ERROR_REQUIRE_BETWEEN = "Between";

    private static final int REQUIRE_ALL = 0;
    private static final int CHECK_REQUIRE_AT_LEAST = 1;
    private static final int CHECK_REQUIRE_AT_MOST = 2;
    private static final int CHECK_REQUIRE_BETWEEN = 3;

    protected int _requireAtMost;
    protected int _requireAtLeast;

    protected int _checkLevels = 0;

    public RequiredHandler() {
        super(NAME);
    }

    public void configure(Config config) throws ConfigException {
        super.configure(config);
        if (config.hasChild(CFG_REQUIRE_AT_LEAST)) {
            try {
                setRequireAtLeast(Integer.parseInt(config.getChild(CFG_REQUIRE_AT_LEAST).getValue()));
            } catch (NumberFormatException e) {
                throw new ConfigException("The " + CFG_REQUIRE_AT_LEAST + " specified is not valid: " + config.getChild(CFG_REQUIRE_AT_LEAST).getValue());
            }
        }
        if (config.hasChild(CFG_REQUIRE_AT_MOST)) {
            try {
                setRequireAtMost(Integer.parseInt(config.getChild(CFG_REQUIRE_AT_MOST).getValue()));
            } catch (NumberFormatException e) {
                throw new ConfigException("The " + CFG_REQUIRE_AT_MOST + " specified is not valid: " + config.getChild(CFG_REQUIRE_AT_MOST).getValue());
            }
        }
    }

    public void setRequireAll() {
        _checkLevels = REQUIRE_ALL;
    }

    public void setRequireAtLeast(int level) {
        _requireAtLeast = level;
        _checkLevels = _checkLevels | CHECK_REQUIRE_AT_LEAST;
    }

    public void setRequireAtMost(int level) {
        _requireAtMost = level;
        _checkLevels = _checkLevels | CHECK_REQUIRE_AT_MOST;
    }

    public Object validate(Object[] values) throws DataHandlerException {
        int hits = 0;
        for (int i = 0; i < values.length; i++) {
            if (values[i] != null && values[i].toString().trim().length() > 0)
                hits++;
        }
        switch (_checkLevels) {
            case CHECK_REQUIRE_BETWEEN :
                if (hits > _requireAtMost || hits < _requireAtLeast)
                    throw new DataHandlerException(_name, ERROR_REQUIRE_BETWEEN, _ids, values);
                break;
            case CHECK_REQUIRE_AT_LEAST :
                if (hits < _requireAtLeast)
                    throw new DataHandlerException(_name, ERROR_REQUIRE_AT_LEAST, _ids, values);
                break;
            case CHECK_REQUIRE_AT_MOST :
                if (hits > _requireAtMost)
                    throw new DataHandlerException(_name, ERROR_REQUIRE_AT_MOST, _ids, values);
                break;
            default :
                if (hits < values.length)
                    throw new DataHandlerException(_name, _ids, values);
                break;
        }
        return null;
    }
}