/*
 * JDBCDataSource.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.datasource.jdbc;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.RowSet;

import sun.jdbc.rowset.CachedRowSet;
import sun.jdbc.rowset.RowSetMetaDataImpl;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.datasource.Condition;
import com.eXtropia.datasource.Criterion;
import com.eXtropia.datasource.DataSourceException;
import com.eXtropia.datasource.FieldOrder;
import com.eXtropia.datasource.Query;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.util.ExtropiaException;

/**
 *
 * @author  Carlo Ravagli (carlo.ravagli@extropia.com)
 * @version 0.1
 *
 * Modification History:
 * sequence - date - developer - notes
 * ========   ====   =========   =====
 *
 * [01] - 28 Nov 01 - stephan - Added JNDI Datasource resolution

 *
 *
 */
public class JDBCDataSource extends com.eXtropia.datasource.DataSource implements Configurable {
    public static final String CFG_DRIVER = "Driver";
    public static final String CFG_CONNECTION = "Connection";
    public static final String CFG_TABLE = "Table";
    public static final String CFG_ROWSET_TYPE = "CopyRowSet";
    public static final String CFG_JNDI_DATASOURCE = "JNDIResource"; //[01:+]

    public static final String CFG_COLUMN = "Column";
    public static final String CFG_COL_DEFAULT_VALUE = "DefaultValue";
    public static final String CFG_COL_ALIAS = "Id";
    public static final String CFG_COL_KEYWORD_SEARCHABLE = "KeywordSearchable";
    public static final String CFG_COL_NOCASE = "CaseInsensitive";
    public static final String CFG_COL_PRIMARY_KEY = "PrimaryKey";

    public static final String CFG_AUTOINC_COLUMN = "AutoIncColumn";
    public static final String CFG_AUTOINC_STORED_PROC = "StoredProc";
    public static final String CFG_AUTOINC_NEXTVAL = "NextValueFunc";
    public static final String CFG_AUTOINC_CURRVAL = "CurrentValueFunc";
    public static final String CFG_AUTOINC_SOURCE = "PseudoSourceTable";

    private static final String[] _operators = { "=", "!=", ">=", ">", "<=", "<" };

    private HashMap _columnMetaData;
    private String _table;
    private String[] _keyFields;
    private boolean _copyRowSet;
    private boolean _configured;

    private boolean _autoIncIsStoredProc = false;
    private String _autoIncNextValFunc;
    private String _autoIncCurrValFunc;
    private String _autoIncPseudoTable;
    private ColMetaData _autoIncColMetaData;

    private boolean _useJNDI = false; //[01:+]
    private Context _initialContext; //[01:+]
    private String _jndiDBPath; //[01:+]
    private javax.sql.DataSource _jndiDataSource; //[01:+]
    private Driver _driver;
    private String _connectionUrl;
    private Properties _connectionProps;

    protected Doctor doctor;

    public JDBCDataSource() throws ExtropiaException {
        _columnMetaData = new HashMap();
    }

    public synchronized void configure(Config config) throws ConfigException {
        if (_configured)
            throw new ConfigException("JDBCDataSource already configured");

        Config driverConfig = config.getChild(CFG_DRIVER);
        Config connectionConfig = config.getChild(CFG_CONNECTION);
        Config tableConfig = config.getChild(CFG_TABLE);
        Config rowsetConfig = config.getChild(CFG_ROWSET_TYPE);
        Config jndiConfig = config.getChild(CFG_JNDI_DATASOURCE); //[01:+]
        String myClassName = getClass().getName();

        //[01:>] The following needs to check and make sure we have at least one
        //       of either JNDI of JDBC Driver configured
        if ((driverConfig == null || driverConfig.getType() == null) && //[01:>] modified to include JNDI
         (jndiConfig == null || jndiConfig.getValue() == null)) //[01:+]
            throw new ConfigException("No <Driver> or <JNDIResource> sections was specified for JDBCDataSource");

        if ((connectionConfig == null || connectionConfig.getValue() == null) && //[01:>] modified to include JNDI
         (jndiConfig == null || jndiConfig.getValue() == null)) //[01:+]
            throw new ConfigException("No <Connection> url or <JNDIResource> specified for JDBCDataSource");

        if (tableConfig == null || tableConfig.getValue() == null)
            throw new ConfigException("No table specified in configuration for JDBCDataSource");

        try {
            if (connectionConfig != null) //[01:+]
                {
                _connectionUrl = connectionConfig.getValue();
                _connectionProps = connectionConfig.toProperties();
                Class.forName(driverConfig.getType());
                try {
                    _driver = DriverManager.getDriver(_connectionUrl);
                } catch (SQLException e) {
                    _driver = (Driver) Class.forName(driverConfig.getType()).newInstance();
                }
            } //connectionConfig != null
            //[01:+>]
            else if (jndiConfig != null) {

                try {
                    _useJNDI = true;
                    _initialContext = new InitialContext();
                    _jndiDBPath = jndiConfig.getValue();
                    _jndiDataSource = (javax.sql.DataSource) _initialContext.lookup(_jndiDBPath);

                } catch (NamingException ex) {
                    throw new ConfigException("Your JNDI Datasource is not properly configured: " + ex.getMessage());
                }
            }
            //[01:+<]

        } catch (Exception e) {
            throw new ConfigException("Unable to find specified JDBC driver class: " + driverConfig.getType());
        }

        _table = tableConfig.getValue();
        doctor = Diagnostics.getDoctor(new String[] { CFG_BASE, myClassName.substring(myClassName.lastIndexOf(".") + 1), _table });

        if (rowsetConfig != null && rowsetConfig.getValue() != null)
            _copyRowSet = Boolean.valueOf(rowsetConfig.getValue()).booleanValue();

        // Remember to take this out when ConnectedOffsetRowSet is completed
        if (!_copyRowSet)
            throw new ConfigException("Sorry connected JDBC rowset is not yet implemented.");

        getColsConfig(tableConfig);

        _configured = true;
    }

    //[01:+>]
    /* This method returns a JDBC Connection object depending on
       what the Config section in your <DataSource> looks like
    */
    private Connection getConnection() throws SQLException, ConfigException {
        Connection conn = null;

        if (_useJNDI && (_initialContext != null)) /*Check for JNDI*/ {

            if (_jndiDataSource != null)
                conn = _jndiDataSource.getConnection();
        } else /*assume JDBC driver datasource*/ {
            if (_driver != null)
                conn = _driver.connect(_connectionUrl, _connectionProps);
        }

        return conn;
    }
    //[01:+<]

    private void getColsConfig(Config tableConfig) throws ConfigException {
        Config[] colsConfig = tableConfig.getChildren();
        if (colsConfig == null || colsConfig.length == 0)
            throw new ConfigException("No columns specified for JDBC DataSource table: " + _table);

        for (int i = 0; i < colsConfig.length; i++) {
            if (colsConfig[i].getValue() == null)
                throw new ConfigException("No column name specified in column config for JDBC DataSource table: " + _table);

            ColMetaData metaData = new ColMetaData();
            metaData.setName(colsConfig[i].getValue());
            Config childConfig;

            List keys = new ArrayList();
            if (colsConfig[i].getName().equals(CFG_COLUMN)) {
                childConfig = colsConfig[i].getChild(CFG_COL_KEYWORD_SEARCHABLE);
                if (childConfig != null && childConfig.getValue() != null)
                    metaData.setKeywordSearchable(Boolean.valueOf(childConfig.getValue()).booleanValue());

                childConfig = colsConfig[i].getChild(CFG_COL_DEFAULT_VALUE);
                if (childConfig != null && childConfig.getValue() != null)
                    metaData.setDefaultValue((String) childConfig.getValue());

                childConfig = colsConfig[i].getChild(CFG_COL_NOCASE);
                if (childConfig != null && childConfig.getValue() != null)
                    metaData.setCaseInsensitive(Boolean.valueOf(childConfig.getValue()).booleanValue());

                childConfig = colsConfig[i].getChild(CFG_COL_PRIMARY_KEY);
                if (childConfig != null && childConfig.getValue() != null) {
                    keys.add(colsConfig[i].getValue());
                }
            } else if (colsConfig[i].getName().equals(CFG_AUTOINC_COLUMN)) {
                if (_autoIncColMetaData != null)
                    throw new ConfigException("Cannot have more that one auto incrementing column");

                _autoIncColMetaData = metaData;
                metaData.setAutoIncrement(true);

                childConfig = colsConfig[i].getChild(CFG_AUTOINC_STORED_PROC);
                if (childConfig != null && childConfig.getValue() != null)
                    _autoIncIsStoredProc = Boolean.valueOf(childConfig.getValue()).booleanValue();

                childConfig = colsConfig[i].getChild(CFG_AUTOINC_NEXTVAL);
                if (childConfig != null && childConfig.getValue() != null)
                    _autoIncNextValFunc = childConfig.getValue();
                else if (_autoIncIsStoredProc)
                    throw new ConfigException("No stored proc defined for auto increment");

                childConfig = colsConfig[i].getChild(CFG_AUTOINC_CURRVAL);
                if (childConfig != null && childConfig.getValue() != null)
                    _autoIncCurrValFunc = childConfig.getValue();

                childConfig = colsConfig[i].getChild(CFG_AUTOINC_SOURCE);
                if (childConfig != null && childConfig.getValue() != null)
                    _autoIncPseudoTable = childConfig.getValue();

                childConfig = colsConfig[i].getChild(CFG_COL_PRIMARY_KEY);
                if (childConfig != null && childConfig.getValue() != null) {
                    keys.add(colsConfig[i].getValue());
                }
            } else
                throw new ConfigException("Invalid config element in table config: " + colsConfig[i].getName());

            childConfig = colsConfig[i].getChild(CFG_COL_ALIAS);
            _columnMetaData.put(childConfig != null && childConfig.getValue() != null ? childConfig.getValue() : metaData.getName(), metaData);
            if (keys.size() > 0) {
                _keyFields = (String[]) keys.toArray(new String[keys.size()]);
            }
        }
        populateColMetaData();
    }
    private void populateColMetaData() throws ConfigException {
        if (doctor.doTrace(TraceEvent.VERBOSE))
            doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Getting column meta data for database table:" + _table));

        Connection conn = null;
        Statement stmt = null;
        ResultSet rs = null;
        try {
            System.out.println(_connectionUrl);
            System.out.println(_connectionProps);
            //[01:-] conn = _driver.connect(_connectionUrl, _connectionProps);
            conn = getConnection(); //[01:+]

            stmt = conn.createStatement();
            stmt.setMaxRows(1);
            rs = stmt.executeQuery(createMetaDataQuery());

            ResultSetMetaData rsmd = rs.getMetaData();

            for (int i = 1; i <= rsmd.getColumnCount(); i++) {

                // Doing hash lookup of alias at config to test if every thing is OK.
                ColMetaData metaData = (ColMetaData) _columnMetaData.get(rsmd.getColumnName(i));
                metaData.setSqlType(rsmd.getColumnType(i));
                boolean searchable = rsmd.isSearchable(i);
                if (metaData.isKeywordSearchable() && !searchable)
                    throw new ConfigException("Cannot make column <" + metaData.getName() + "> keyword searchable because it is not database searchable");
                metaData.setSearchable(rsmd.isSearchable(i));

                if (rsmd.isAutoIncrement(i) && !metaData.isAutoIncrement())
                    metaData.setAutoIncrement(true);
            }
        } catch (SQLException e) {
            String errmsg = "Failed to get column metadata from database (Table:" + _table + "): " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new ConfigException(errmsg);
        } finally {
            try {
                if (rs != null)
                    rs.close();
                if (stmt != null)
                    stmt.close();
                if (conn != null)
                    conn.close();
            } catch (SQLException e) {
                doctor.log(new LogEvent(this, LogEvent.CRIT, "Failed to close database resources after getting column metadata: " + e.getMessage()));
            }
        }
    }

    private String createMetaDataQuery() {
        StringBuffer s = new StringBuffer("SELECT ");
        boolean appendCommer = false;

        for (Iterator iter = _columnMetaData.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            ColMetaData metadata = (ColMetaData) entry.getValue();
            if (appendCommer)
                s.append(',');
            else
                appendCommer = true;

            // Aliasing col names to test validity at config time rather than execution.
            s.append(metadata.getName()).append(" AS \"").append(entry.getKey()).append('"');
        }
        s.append(" FROM ").append(_table);

        return s.toString();
    }

    private void addSubCriterion(Criterion criterion, StringBuffer sql, List values) throws DataSourceException {
        String type = criterion.getType() != Criterion.AND ? " OR " : " AND ";
        boolean appendType = false;
        Condition[] conditions = criterion.getConditions();

        for (int i = 0; conditions != null && i < conditions.length; i++) {
            ColMetaData metaData = (ColMetaData) _columnMetaData.get(conditions[i].getField());

            if (metaData == null)
                throw new DataSourceException("Field not defined in datasource (" + _table + "): " + conditions[i].getField());

            if (appendType)
                sql.append(type);
            else {
                sql.append('(');
                appendType = true;
            }
            if (metaData.isCaseInsensitive())
                sql.append("LOWER(").append(metaData.getName()).append(')');
            else
                sql.append(metaData.getName());

            if (conditions[i].getOperator() == Condition.EQUAL || conditions[i].getOperator() == Condition.NOT_EQUAL) {
                if (conditions[i].getValue() == null) {
                    sql.append(" IS");
                    if (conditions[i].getOperator() == Condition.NOT_EQUAL)
                        sql.append(" NOT");
                    sql.append(" NULL");
                } else {
                    if (conditions[i].getValue() instanceof String) {
                        String value = metaData.isCaseInsensitive() ? ((String) conditions[i].getValue()).toLowerCase() : (String) conditions[i].getValue();

                        boolean likeOperator = false;
                        boolean needsEscaping = false;
                        for (int j = 0; j < value.length() && !likeOperator; j++) {
                            switch (value.charAt(j)) {
                                case '*' :
                                case '?' :
                                    if (j == 0 || value.charAt(j - 1) != '\\')
                                        likeOperator = true;
                                    else
                                        needsEscaping = true;
                            }
                        }

                        if (likeOperator) {
                            if (conditions[i].getOperator() == Condition.NOT_EQUAL)
                                sql.append(" NOT");
                            sql.append(" LIKE \'").append(escapeLikeParttern(value)).append('\'');
                        } else {
                            sql.append(_operators[conditions[i].getOperator()]).append('?');
                            values.add(new SqlValue(needsEscaping ? escapeString(value) : value, metaData.getSqlType()));
                        }
                    } else {
                        sql.append(_operators[conditions[i].getOperator()]).append('?');
                        values.add(new SqlValue(conditions[i].getValue(), metaData.getSqlType()));
                    }
                }
            } else {
                if (conditions[i].getValue() == null)
                    throw new DataSourceException("No value specified for condition: " + conditions[i].getField() + _operators[conditions[i].getOperator()]);

                sql.append(_operators[conditions[i].getOperator()]).append('?');
                values.add(new SqlValue(conditions[i].getValue(), metaData.getSqlType()));

            }
        }

        Criterion[] subCriteria = criterion.getSubCriteria();
        for (int i = 0; subCriteria != null && i < subCriteria.length; i++) {
            if (subCriteria[i] != null && (subCriteria[i].getConditions() != null || subCriteria[i].getSubCriteria() != null)) {
                if (appendType)
                    sql.append(type);
                else {
                    sql.append('(');
                    appendType = true;
                }

                addSubCriterion(subCriteria[i], sql, values);
            }
        }

        if (appendType)
            sql.append(')');
    }

    protected String escapeLikeParttern(String likePattern) throws DataSourceException {
        StringBuffer retval = new StringBuffer(likePattern.length());

        //
        // if we need to escape the '%' or '_' and ask the db server
        // not to interpret them as wild cards.
        // There's no harm in appending "escape '/'", just that
        // the SQL statement will be longer ....
        //
        boolean escapeOn = false;
        if (likePattern.indexOf('%') != -1 || likePattern.indexOf('_') != -1) {
            escapeOn = true;
        }

        for (int i = 0; i < likePattern.length(); i++) {
            char c = likePattern.charAt(i);
            switch (c) {
                case '%' :
                case '_' :
                    retval.append('\\').append(c);
                    break;
                case '\'' :
                    retval.append('\'').append(c);
                    break;
                case '*' :
                    retval.append('%');
                    break;
                case '?' :
                    retval.append('_');
                    break;
                case '\\' :
                    if (++i < likePattern.length()) {
                        char escapedChar = likePattern.charAt(i);
                        switch (escapedChar) {
                            case '*' :
                            case '?' :
                                retval.append(escapedChar);
                                break;
                            case '\\' :
                                if (escapeOn) {
                                    //
                                    // since we are escaping the char '\'
                                    // we need to append two '\', the first
                                    // one to escape the second one
                                    //
                                    retval.append('\\').append('\\');
                                } else {
                                    retval.append('\\');
                                }
                                break;
                            default :
                                throw new DataSourceException("Invalid escape character in LIKE pattern: " + escapedChar);
                        }
                    } else {
                        //
                        // escape char at the end of the word needs not be escaped
                        // ie. "a\\" and "a\\\\" will have similar results
                        //
                        if (escapeOn) {
                            //
                            // since we are escaping the char '\'
                            // we need to append two '\', the first
                            // one to escape the second one
                            //
                            retval.append('\\').append('\\');
                        } else {
                            retval.append('\\');
                        }
                    }
                    break;
                default :
                    retval.append(c);
            }
        }
        if (escapeOn) {
            retval.append("' escape '\\");
        }

        return retval.toString();
    }

    protected String escapeString(String s) throws DataSourceException {
        StringBuffer retval = new StringBuffer(s.length());

        for (int i = 0; i < s.length(); i++) {
            char c = s.charAt(i);
            switch (c) {
                case '\\' :
                    if (++i < s.length()) {
                        char escapedChar = s.charAt(i);
                        switch (escapedChar) {
                            case '*' :
                            case '?' :
                            case '\\' :
                                retval.append(escapedChar);
                                break;
                            default :
                                throw new DataSourceException("Invalid escape character in LIKE pattern: " + escapedChar);
                        }
                    } else
                        retval.append('\\').append('\\');
                    break;
                default :
                    retval.append(c);
            }
        }

        return retval.toString();
    }

    protected Iterator generateWhereClause(Criterion criterion, StringBuffer sql) throws DataSourceException {
        if (criterion == null || (criterion.getConditions() == null && criterion.getSubCriteria() == null)) {
            return new Iterator() {
                public boolean hasNext() {
                    return false;
                }
                public Object next() {
                    throw new NoSuchElementException();
                }
                public void remove() {
                    throw new UnsupportedOperationException();
                }
            };
        }

        LinkedList values = new LinkedList();
        sql.append(" WHERE ");
        addSubCriterion(criterion, sql, values);

        return values.iterator();
    }

    public int add(Map rowValues) throws DataSourceException {
        // Added check for default values where none exists...
        for (Iterator iter = _columnMetaData.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            ColMetaData metaData = (ColMetaData) _columnMetaData.get(entry.getKey());
            String s = metaData.getDefaultValue();
            if (s != null && rowValues.get(metaData.getName()) == null) {
                rowValues.put(metaData.getName(), s);
            }
        }
        if (!_configured)
            throw new DataSourceException("Unable to perform add. JDBC data source has not been configured");

        if (rowValues.size() == 0)
            throw new DataSourceException("Failed to perform add: Unable to insert row where no values have been apecified.");

        if (doctor.doTrace(TraceEvent.INFO))
            doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Adding row to datasource."));

        StringBuffer colNames = new StringBuffer();
        StringBuffer placeHolders = new StringBuffer();

        //anderson - change the implementation to ignore the temporary columns
        /*
        SqlValue[] values = new SqlValue[ rowValues.size() ];
        
        int i = 0;
        for( Iterator iter = rowValues.entrySet().iterator(); iter.hasNext(); i++ )
        {
            Map.Entry entry = (Map.Entry)iter.next();
            ColMetaData metaData = (ColMetaData)_columnMetaData.get(entry.getKey());
        
            if( metaData == null )
                throw new DataSourceException("Field not defined not in datasource: " + entry.getKey());
        
            if( metaData.isAutoIncrement() )
                throw new DataSourceException("Cannot specify a value for auto incrementing field: " + entry.getKey());
        
            if( i > 0 )
            {
                colNames.append(',');
                placeHolders.append(',');
            }
        
            colNames.append( metaData.getName() );
            placeHolders.append('?');
            values[i] = new SqlValue( entry.getValue(), metaData.getSqlType() );
        }
        */
        ArrayList valuesList = new ArrayList(53);
        int i = 0;
        for (Iterator iter = rowValues.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            ColMetaData metaData = (ColMetaData) _columnMetaData.get(entry.getKey());

            if (metaData == null) {
                //Anderson
                //just ignore the error and ignore the entry instead of throwing the exception
                continue;
                //throw new DataSourceException("Field not defined not in datasource: " + entry.getKey());
            }

            if (metaData.isAutoIncrement())
                throw new DataSourceException("Cannot specify a value for auto incrementing field: " + entry.getKey());

            if (i > 0) {
                colNames.append(',');
                placeHolders.append(',');
            }

            colNames.append(metaData.getName());
            placeHolders.append('?');
            valuesList.add(new SqlValue(entry.getValue(), metaData.getSqlType()));
            //has to put the i++ here instead of the same of the for statement
            i++;
        }
        SqlValue[] values = new SqlValue[valuesList.size()];
        values = (SqlValue[]) valuesList.toArray(values);

        if (_autoIncNextValFunc != null) {
            colNames.append(',').append(_autoIncColMetaData.getName());
            if (_autoIncIsStoredProc)
                placeHolders.append(",?");
            else
                placeHolders.append(',').append(_autoIncNextValFunc);
        }

        if (doctor.doTrace(TraceEvent.VERBOSE))
            doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Insert SQL for prepared statement: INSERT INTO " + _table + '(' + colNames + ") VALUES (" + placeHolders + ')'));

        Connection conn = null;
        Statement stmt = null;
        PreparedStatement prepStmt = null;
        try {
            int autoIncCurrVal = 0;
            //[01:-] conn = _driver.connect(_connectionUrl, _connectionProps);
            conn = getConnection(); //[01:+]

            if (_autoIncIsStoredProc) {
                stmt = conn.createStatement();
                stmt.setMaxRows(1);

                if (doctor.doTrace(TraceEvent.DEVELOPER))
                    doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Getting auto increment next value"));

                ResultSet rs = stmt.executeQuery(_autoIncNextValFunc);

                if (rs.next())
                    autoIncCurrVal = ((Number) rs.getObject(1)).intValue();
                else
                    throw new DataSourceException("Failed to get auto increment value from stored proc");

                stmt.close();
                stmt = null;
            }
            prepStmt = conn.prepareStatement("INSERT INTO " + _table + " (" + colNames + ") VALUES (" + placeHolders + ')');

            for (i = 0; i < values.length; i++) {
                if (values[i].getValue() != null)
                    prepStmt.setObject(i + 1, values[i].getValue());
                else
                    prepStmt.setNull(i + 1, values[i].getType());
            }
            if (_autoIncIsStoredProc)
                prepStmt.setInt(values.length + 1, autoIncCurrVal);
            if (prepStmt.executeUpdate() != 1)
                throw new DataSourceException("Failed to add row into data source: " + _table);

            prepStmt.close();
            prepStmt = null;

            if (!_autoIncIsStoredProc && _autoIncCurrValFunc != null) {
                stmt = conn.createStatement();
                stmt.setMaxRows(1);
                String sql = "SELECT " + _autoIncCurrValFunc + " FROM " + (_autoIncPseudoTable != null ? _autoIncPseudoTable : _table);

                if (doctor.doTrace(TraceEvent.VERBOSE))
                    doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Getting auto increment current value: " + sql));

                ResultSet rs = stmt.executeQuery(sql);

                if (rs.next())
                    autoIncCurrVal = ((Number) rs.getObject(1)).intValue();
                else
                    throw new DataSourceException("Failed to get auto increment value after insert");
            }

            return autoIncCurrVal;
        } catch (SQLException e) {
            String errmsg = "Failed to add row. Database exception occured: " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new DataSourceException(errmsg);
        } catch (Exception e) {
            String errmsg = "Failed to add row. Unexpected runtime exception occured: " + e.getClass().getName() + " - " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new DataSourceException(errmsg);
        } finally {
            try {
                if (stmt != null)
                    stmt.close();
                if (prepStmt != null)
                    prepStmt.close();
                if (conn != null)
                    conn.close();
            } catch (SQLException e) {
                doctor.log(new LogEvent(this, LogEvent.CRIT, "Failed to close database resources after insert: " + e.getMessage()));
            }
        }
    }

    public int delete() throws DataSourceException {
        return delete((Criterion) null);
    }

    public int delete(Condition condition) throws DataSourceException {
        return delete(new Criterion(condition));
    }

    public int delete(Criterion criterion) throws DataSourceException {
        if (!_configured)
            throw new DataSourceException("Unable to perform delete. JDBC data source has not been configured");

        if (doctor.doTrace(TraceEvent.INFO))
            doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Deleting row(s) from datasource."));

        StringBuffer sql = new StringBuffer("DELETE FROM " + _table);
        Iterator valuesIter = generateWhereClause(criterion, sql);
        if (doctor.doTrace(TraceEvent.VERBOSE))
            doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Delete SQL for prepared statement: " + sql));

        Connection conn = null;
        PreparedStatement stmt = null;
        try {
            //[01:-] conn = _driver.connect(_connectionUrl, _connectionProps);
            conn = getConnection(); //[01:+]
            stmt = conn.prepareStatement(sql.toString());

            for (int i = 1; valuesIter.hasNext(); i++) {
                SqlValue conditionValue = (SqlValue) valuesIter.next();
                stmt.setObject(i, conditionValue.getValue());
            }

            return stmt.executeUpdate();
        } catch (SQLException e) {
            String errmsg = "Database exception occured while trying to delete row(s): " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new DataSourceException(errmsg);
        } catch (Exception e) {
            String errmsg = "Failed to delete. Unexpected runtime exception occured: " + e.getClass().getName() + " - " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new DataSourceException(errmsg);
        } finally {
            try {
                if (stmt != null)
                    stmt.close();
                if (conn != null)
                    conn.close();
            } catch (SQLException e) {
                doctor.log(new LogEvent(this, LogEvent.CRIT, "Failed to close database resources after delete: " + e.getMessage()));
            }
        }
    }

    public int update(Map rowValues, Condition condition) throws DataSourceException {
        return update(rowValues, new Criterion(condition));
    }

    public int update(Map rowValues, Criterion criterion) throws DataSourceException {
        if (!_configured)
            throw new DataSourceException("Unable to perform update. JDBC data source has not been configured");

        if (rowValues.size() == 0) {
            if (doctor.doTrace(TraceEvent.INFO))
                doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Tring to update row(s) with no new values. Doing nothing."));
            return 0;
        }

        if (doctor.doTrace(TraceEvent.INFO))
            doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Updating row(s) in datasource."));

        StringBuffer sql = new StringBuffer("UPDATE " + _table + " SET ");
        /*
        SqlValue[] updateValues = new SqlValue[ rowValues.size() ];
        int i = 0;
        for( Iterator iter = rowValues.entrySet().iterator(); iter.hasNext(); )
        {
            Map.Entry entry = (Map.Entry)iter.next();
            ColMetaData metaData = (ColMetaData)_columnMetaData.get(entry.getKey());
        
            if( metaData == null ) {
                throw new DataSourceException("Field not defined not in datasource: " + entry.getKey());
            }
        
            if( metaData.isAutoIncrement() )
                throw new DataSourceException("Cannot specify a value for auto incrementing field: " + entry.getKey());
        
            if( i > 0 ) sql.append(',');
        
            sql.append( metaData.getName() ).append("=?");
        
            updateValues[i++] = new SqlValue(entry.getValue(), metaData.getSqlType() );
        }
        */
        ArrayList updateValuesList = new ArrayList(53);
        int i = 0;
        for (Iterator iter = rowValues.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            ColMetaData metaData = (ColMetaData) _columnMetaData.get(entry.getKey());

            if (metaData == null) {
                //Anderson
                //just ignore the error instead of throwing the exception
                continue;
                //throw new DataSourceException("Field not defined not in datasource: " + entry.getKey());
            }

            if (metaData.isAutoIncrement())
                throw new DataSourceException("Cannot specify a value for auto incrementing field: " + entry.getKey());

            if (i > 0)
                sql.append(',');

            sql.append(metaData.getName()).append("=?");

            updateValuesList.add(new SqlValue(entry.getValue(), metaData.getSqlType()));
            i++;
        }
        SqlValue[] updateValues = new SqlValue[updateValuesList.size()];
        updateValues = (SqlValue[]) updateValuesList.toArray(updateValues);

        Iterator conditionValuesIter = generateWhereClause(criterion, sql);
        if (doctor.doTrace(TraceEvent.VERBOSE))
            doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Update SQL for prepared statement: " + sql));

        Connection conn = null;
        PreparedStatement stmt = null;
        try {
            //[01:-] conn = _driver.connect(_connectionUrl, _connectionProps);
            conn = getConnection(); //[01:+]
            stmt = conn.prepareStatement(sql.toString());

            for (i = 0; i < updateValues.length; i++) {
                if (updateValues[i].getValue() != null)
                    stmt.setObject(i + 1, updateValues[i].getValue());
                else
                    stmt.setNull(i + 1, updateValues[i].getType());
            }

            while (conditionValuesIter.hasNext()) {
                SqlValue conditionValue = (SqlValue) conditionValuesIter.next();
                stmt.setObject(++i, conditionValue.getValue());
            }

            return stmt.executeUpdate();
        } catch (SQLException e) {
            String errmsg = "Database exception occured while trying to update row(s): " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new DataSourceException(errmsg);
        } catch (Exception e) {
            String errmsg = "Failed to update. Unexpected runtime exception occured: " + e.getClass().getName() + " - " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new DataSourceException(errmsg);
        } finally {
            try {
                if (stmt != null)
                    stmt.close();
                if (conn != null)
                    conn.close();
            } catch (SQLException e) {
                doctor.log(new LogEvent(this, LogEvent.CRIT, "Failed to close database resources after update: " + e.getMessage()));
            }
        }
    }

    public RowSet search(Query query) throws DataSourceException {
        if (!_configured)
            throw new DataSourceException("Unable to perform search. JDBC data source has not been configured");

        if (query == null)
            new IllegalArgumentException();

        if (doctor.doTrace(TraceEvent.INFO))
            doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Querying datasource: " + _table));

        StringBuffer sql = new StringBuffer("SELECT ");

        if (query.isDistinct()) {
            sql.append("DISTINCT ");
        }

        String[] returnFields = query.getReturnFileds();
        boolean needCommer = false;

        if (returnFields == null || returnFields.length == 0) {
            for (Iterator iter = _columnMetaData.entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                ColMetaData metaData = (ColMetaData) entry.getValue();
                if (needCommer)
                    sql.append(',');
                else
                    needCommer = true;

                sql.append(metaData.getName()).append(" AS \"").append(entry.getKey()).append('"');
            }
        } else {
            for (int i = 0; i < returnFields.length; i++) {
                ColMetaData metaData = (ColMetaData) _columnMetaData.get(returnFields[i]);

                if (metaData == null)
                    throw new DataSourceException("Field not defined not in datasource: " + returnFields[i]);

                if (i > 0)
                    sql.append(',');

                sql.append(metaData.getName()).append(" AS \"").append(returnFields[i]).append('"');
            }
        }

        sql.append(" FROM ").append(_table);

        Iterator conditionValuesIter = generateWhereClause(query.getCriterion(), sql);

        FieldOrder[] orderBy = query.getFieldOrder();
        if (orderBy != null && orderBy.length > 0) {
            sql.append(" ORDER BY");
            for (int i = 0; i < orderBy.length; i++) {
                ColMetaData metaData = (ColMetaData) _columnMetaData.get(orderBy[i].getField());

                if (metaData == null)
                    throw new DataSourceException("Field not defined not in datasource: " + orderBy[i].getField());

                sql.append(i > 0 ? ',' : ' ');
                if (metaData.isCaseInsensitive())
                    sql.append("LOWER(").append(metaData.getName()).append(')');
                else
                    sql.append(metaData.getName());

                if (orderBy[i].getOrder() == FieldOrder.DESCEND)
                    sql.append(" DESC");
            }
        }
        if (doctor.doTrace(TraceEvent.VERBOSE))
            doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Query SQL for RowSet: " + sql));

        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet results = null;

        try {
            //[01:-] conn = _driver.connect(_connectionUrl, _connectionProps);
            conn = getConnection(); //[01:+]
            stmt = conn.prepareStatement(sql.toString());

            if (query.getLimit() > 0)
                stmt.setMaxRows(query.getOffset() + query.getLimit());

            for (int i = 1; conditionValuesIter.hasNext(); i++) {
                SqlValue conditionValue = (SqlValue) conditionValuesIter.next();
                if (doctor.doTrace(TraceEvent.VERBOSE)) {
                    String msg = "Where param " + i + ": " + conditionValue.getValue().getClass().getName();
                    msg += " = " + conditionValue.getValue().toString() + " SQLTYPE = " + conditionValue.getType();
                    doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, msg));
                }
                stmt.setObject(i, conditionValue.getValue());
            }
            results = stmt.executeQuery();
            CachedRowSet crs = new CachedRowSet();

            //crs.populate(results);
            //          if( query.getOffset() > 0 )
            //              results.absolute( query.getOffset() );
            //          Above is only supported by JDBC 2 replaced with the following
            for (int i = 0; i < query.getOffset() && results.next(); i++);
            // The following try/catch was added to fix a problem with
            // CachedRowSets and the Sybase JConnect drivers (both 4.2 and 5.2)
            //
            // In JConnect 4.2, a SQLException is thrown
            // In JConnect 5.2, a normal Exception is thrown with the
            // same message. Go Figure.
            //
            boolean areWeUsingSillySybase = false;
            try {
                crs.populate(results);
            } catch (SQLException e) {
                if (e.getMessage().indexOf("getSchemaName") < 0) {
                    e.printStackTrace();
                    throw e;
                }
                areWeUsingSillySybase = true;
            } catch (Exception e) {
                if (e.getMessage().indexOf("getSchemaName") < 0) {
                    e.printStackTrace();
                    throw e;
                }
                areWeUsingSillySybase = true;
            }
            if (areWeUsingSillySybase) {
                RowSetMetaDataImpl rsmd = new RowSetMetaDataImpl();
                ResultSetMetaData resultMeta = results.getMetaData();
                int count = resultMeta.getColumnCount();
                rsmd.setColumnCount(count);
                for (int i = 1; i <= count; i++) {
                    rsmd.setSchemaName(i, "");
                    rsmd.setColumnName(i, resultMeta.getColumnName(i));
                    rsmd.setColumnLabel(i, resultMeta.getColumnLabel(i));
                    rsmd.setColumnType(i, resultMeta.getColumnType(i));
                }
                crs.setMetaData(rsmd);
            }
            return crs;
        } catch (SQLException e) {
            String errmsg = "Database exception occured while trying to query datasource: " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new DataSourceException(errmsg);
        } catch (Exception e) {
            String errmsg = "Failed to perform search. Unexpected runtime exception occured: " + e.getClass().getName() + " - " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new DataSourceException(errmsg);
        } finally {
            try {
                if (results != null)
                    results.close();
                if (stmt != null)
                    stmt.close();
                if (conn != null)
                    conn.close();
            } catch (SQLException e) {
                doctor.log(new LogEvent(this, LogEvent.CRIT, "Failed to close database resources after query: " + e.getMessage()));
            }
        }
    }

    public int count() throws DataSourceException {
        return count((Criterion) null);
    }

    public int count(Condition condition) throws DataSourceException {
        return count(new Criterion(condition));
    }

    public int count(Criterion criterion) throws DataSourceException {
        if (!_configured)
            throw new DataSourceException("Unable to perform count. JDBC data source has not been configured");

        if (doctor.doTrace(TraceEvent.INFO))
            doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Counting row(s) from datasource."));

        StringBuffer sql = new StringBuffer("SELECT COUNT(*) FROM " + _table);
        Iterator valuesIter = generateWhereClause(criterion, sql);

        if (doctor.doTrace(TraceEvent.VERBOSE))
            doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "COUNT SQL for prepared statement: " + sql));

        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try {
            //[01:-] conn = _driver.connect(_connectionUrl, _connectionProps);
            conn = getConnection(); //[01:+]
            stmt = conn.prepareStatement(sql.toString());
            stmt.setMaxRows(1);

            for (int i = 1; valuesIter.hasNext(); i++) {
                SqlValue conditionValue = (SqlValue) valuesIter.next();
                stmt.setObject(i, conditionValue.getValue());
            }

            rs = stmt.executeQuery();

            if (!rs.next())
                throw new DataSourceException("Failed to get row count for query: " + sql.toString());

            return ((Number) rs.getObject(1)).intValue();

        } catch (SQLException e) {
            String errmsg = "Database exception occured while trying to count row(s): " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new DataSourceException(errmsg);
        } catch (Exception e) {
            String errmsg = "Failed to perform count. Unexpected runtime exception occured: " + e.getClass().getName() + " - " + e.getMessage();
            doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
            throw new DataSourceException(errmsg);
        } finally {
            try {
                if (rs != null)
                    rs.close();
                if (stmt != null)
                    stmt.close();
                if (conn != null)
                    conn.close();
            } catch (SQLException e) {
                doctor.log(new LogEvent(this, LogEvent.CRIT, "Failed to close database resources after count query: " + e.getMessage()));
            }
        }
    }

    public Criterion generateCriterion(String keyword) throws DataSourceException {
        return generateCriterion(keyword, null);
    }

    public Criterion generateCriterion(String keyword, String[] searchFields) throws DataSourceException {
        if (!_configured)
            throw new DataSourceException("Unable to genterate keyword search criterion. JDBC data source has not been configured");

        if (keyword.length() == 0)
            return null;

        Criterion criterion = new Criterion(Criterion.OR);
        if (searchFields == null || searchFields.length == 0) {
            for (Iterator iter = _columnMetaData.entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                ColMetaData metaData = (ColMetaData) entry.getValue();
                if (metaData.isKeywordSearchable())
                    criterion.add(new Condition((String) entry.getKey(), Condition.EQUAL, keyword));
            }
        } else {
            for (int i = 0; i < searchFields.length; i++) {
                ColMetaData metaData = (ColMetaData) _columnMetaData.get(searchFields[i]);
                if (metaData == null || !metaData.isKeywordSearchable())
                    throw new DataSourceException("Field is either not defined not in datasource or not keyword searchable: " + searchFields[i]);
                criterion.add(new Condition(searchFields[i], Condition.EQUAL, keyword));
            }
        }

        return criterion;
    }

    public Criterion generateCriterion(String[] keywords, int criterionType) throws DataSourceException {
        return generateCriterion(keywords, criterionType, null);
    }

    public Criterion generateCriterion(String[] keywords, int criterionType, String[] searchFields) throws DataSourceException {
        if (!_configured)
            throw new DataSourceException("Unable to genterate keyword search criterion. JDBC data source has not been configured");

        if (keywords.length == 0)
            return null;

        Criterion criterion = new Criterion(criterionType);
        if (searchFields == null || searchFields.length == 0) {
            for (int i = 0; i < keywords.length; i++) {
                if (keywords[i] != null && keywords[i].length() > 0) {
                    Criterion subCriterion = new Criterion(Criterion.OR);
                    for (Iterator iter = _columnMetaData.entrySet().iterator(); iter.hasNext();) {
                        Map.Entry entry = (Map.Entry) iter.next();
                        ColMetaData metaData = (ColMetaData) entry.getValue();
                        if (metaData.isKeywordSearchable())
                            subCriterion.add(new Condition((String) entry.getKey(), Condition.EQUAL, keywords[i]));
                    }
                    criterion.add(subCriterion);
                }
            }
        } else {
            for (int i = 0; i < keywords.length; i++) {
                if (keywords[i] != null && keywords[i].length() > 0) {
                    Criterion subCriterion = new Criterion(Criterion.OR);
                    for (int j = 0; j < searchFields.length; j++) {
                        ColMetaData metaData = (ColMetaData) _columnMetaData.get(searchFields[j]);
                        if (metaData == null || !metaData.isKeywordSearchable())
                            throw new DataSourceException("Field is either not defined not in datasource or not keyword searchable: " + searchFields[j]);
                        subCriterion.add(new Condition(searchFields[j], Condition.EQUAL, keywords[i]));
                    }
                    criterion.add(subCriterion);
                }
            }
        }
        return criterion;
    }

    public String[] getFields() throws DataSourceException {
        return (String[]) _columnMetaData.keySet().toArray(new String[_columnMetaData.size()]);
    }

    public String[] getKeyFields() throws DataSourceException {
        if (_keyFields == null) {
            // builds a map of database column name to its identifier
            // so that we can do a lookup later on
            HashMap map = new HashMap();
            String[] arr = getFields();
            for (int i = 0; i < arr.length; i++) {
                ColMetaData col = (ColMetaData) _columnMetaData.get(arr[i]);
                map.put(col.getName().toLowerCase(), arr[i]);
            }

            Connection conn = null;
            ResultSet rs = null;
            try {
                //[01:-] conn = _driver.connect(_connectionUrl, _connectionProps);
                conn = getConnection(); //[01:+]
                rs = conn.getMetaData().getPrimaryKeys("", "%", _table);
                LinkedList keys = new LinkedList();

                while (rs.next()) {
                    // doing the lookup for the identifier
                    String col = (String) map.get(rs.getString(4).toLowerCase());
                    short seq = rs.getShort(5);
                    if (seq > keys.size()) {
                        for (int i = keys.size(); i < seq; i++)
                            keys.addLast(null);
                    }
                    keys.set(seq - 1, col);
                }

                _keyFields = (String[]) keys.toArray(new String[keys.size()]);
            } catch (SQLException e) {
                String errmsg = "Database exception occured while trying to get primary key info: " + e.getMessage();
                doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
                throw new DataSourceException(errmsg);
            }
            //[01:+>]
            catch (ConfigException cex) {
                throw new DataSourceException(cex.getMessage());
            }
            //[01:+<]
            finally {
                try {
                    if (rs != null)
                        rs.close();
                    if (conn != null)
                        conn.close();
                } catch (SQLException e) {
                    doctor.log(new LogEvent(this, LogEvent.CRIT, "Failed to close database resources: " + e.getMessage()));
                }
            }
        }
        return _keyFields;
    }
}
