/*
 * SecureDataSource.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.datasource.secure;

import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.AbstractCollection;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.sql.RowSet;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.datasource.Condition;
import com.eXtropia.datasource.Criterion;
import com.eXtropia.datasource.DataSource;
import com.eXtropia.datasource.DataSourceException;
import com.eXtropia.datasource.Query;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;
import com.eXtropia.diagnostics.LogListener;

/**
 *
 * @author Carlo Ravagli
 * @version 1.0
 */
public class SecureDataSource extends DataSource {
    public static final String CFG_READABLE_FIELDS = "ReadableFields";
    public static final String CFG_WRITABLE_FIELDS = "WritableFields";
    public static final String CFG_FIELD_ID = "Id";

    public static final String CFG_ENABLE_ADD = "EnableAdd";
    public static final String CFG_ENABLE_UPDATE = "EnableUpdate";
    public static final String CFG_ENABLE_DELETE = "EnableDelete";
    public static final String CFG_ENABLE_SEARCH = "EnableSearch";

    public static final String CFG_LOG_ADDS = "LogAdditions";
    public static final String CFG_LOG_UPDATES = "LogUpdates";
    public static final String CFG_LOG_DELETES = "LogDeletes";

    public static final String CFG_QUOTA_CHECK = "QuotaCheck";
    public static final String CFG_QUOTA_MAX = "Maximum";
    public static final String CFG_QUOTA_FIELDS = "ConditionFields";

    protected HashSet _readableFields = new HashSet();
    protected HashSet _writableFields = new HashSet();
    protected String[] _readableFieldsArray;

    protected boolean _addEnabled = false;
    protected boolean _updateEnabled = false;
    protected boolean _deleteEnabled = false;
    protected boolean _searchEnabled = false;

    protected boolean _logAdds = true;
    protected boolean _logDeletes = true;
    protected boolean _logUpdates = true;

    protected int _maxQuota;
    protected String[] _quotaFields;

    protected DataSource _datasource;
    protected Doctor _doctor;

    public SecureDataSource() {

    }

    public int add(Map rowValues) throws DataSourceException {
        try {
            if (!_addEnabled)
                throw new SecureDataSourceException("Insufficient privileges to perform add");

            validateWriteFields(rowValues);

            if (_maxQuota > 0) {
                Criterion criterion = null;
                if (_quotaFields != null) {
                    criterion = new Criterion();
                    for (int i = 0; i < _quotaFields.length; i++) {
                        Object value = rowValues.get(_quotaFields[i]);
                        if (value == null)
                            throw new SecureDataSourceException("Quota condition field not specified: " + _quotaFields[i]);
                        criterion.add(new Condition(_quotaFields[i], Condition.EQUAL, value));
                    }
                }

                if (_datasource.count(criterion) >= _maxQuota)
                    throw new QuotaException("Quota reached: " + _maxQuota, _maxQuota);
            }

            int rv = _datasource.add(rowValues);

            if (_logAdds)
                _doctor.log(new LogEvent(this, LogEvent.INFO, "add", rowValues));

            return rv;
        } catch (QuotaException qe) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, qe.getMessage()));
            throw qe;
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public int delete() throws DataSourceException {
        return delete((Criterion) null);
    }

    public int delete(Condition condition) throws DataSourceException {
        return condition != null ? delete(new Criterion(condition)) : delete((Criterion) null);
    }

    public int delete(Criterion criterion) throws DataSourceException {
        try {
            if (!_deleteEnabled)
                throw new SecureDataSourceException("Insufficient privileges to perform delete");

            validateCriterion(criterion);

            RowSet deletedValues = null;
            if (_logDeletes)
                deletedValues = _datasource.search(new Query(criterion));

            int count = _datasource.delete(criterion);

            if (_logDeletes)
                logValues("delete", deletedValues);

            return count;
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public int update(Map rowValues, Condition condition) throws DataSourceException {
        return update(rowValues, condition != null ? new Criterion(condition) : null);
    }

    public int update(Map rowValues, Criterion criterion) throws DataSourceException {
        try {
            if (!_updateEnabled)
                throw new SecureDataSourceException("Insufficient privileges to perform update");

            validateWriteFields(rowValues);
            validateCriterion(criterion);

            RowSet updatedValues = null;
            if (_logUpdates)
                updatedValues = _datasource.search(new Query(criterion));

            int count = _datasource.update(rowValues, criterion);

            if (_logUpdates)
                logValues("update", updatedValues);

            return count;
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public RowSet search(Query query) throws DataSourceException {
        try {

            if (!_searchEnabled)
                throw new SecureDataSourceException("Insufficient privileges to perform search");

            validateCriterion(query.getCriterion());
            if (query.getReturnFileds() != null)
                validateReadFields(query.getReturnFileds());
            else {
                if (_readableFieldsArray == null)
                    throw new SecureDataSourceException("Insufficient privileges to read all fields");
                query.setReturnFields(_readableFieldsArray);
            }

            return _datasource.search(query);
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public int count() throws DataSourceException {
        try {
            if (!_searchEnabled)
                throw new SecureDataSourceException("Insufficient privileges to perform count");

            return _datasource.count();
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public int count(Condition condition) throws DataSourceException {
        return condition != null ? count(new Criterion(condition)) : count((Criterion) null);
    }

    public int count(Criterion criterion) throws DataSourceException {
        try {
            if (!_searchEnabled)
                throw new SecureDataSourceException("Insufficient privileges to perform count");

            validateCriterion(criterion);

            return _datasource.count(criterion);
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public String[] getKeyFields() throws DataSourceException {
        try {
            String[] keyFields = _datasource.getKeyFields();
            validateReadFields(keyFields);
            return keyFields;
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public String[] getFields() throws DataSourceException {
        return _readableFieldsArray;
    }

    public Criterion generateCriterion(String keyword) throws DataSourceException {
        try {
            return removeInvalidConditions(_datasource.generateCriterion(keyword));
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public Criterion generateCriterion(String keyword, String[] searchFields) throws DataSourceException {
        try {
            validateReadFields(searchFields);
            return _datasource.generateCriterion(keyword, searchFields);
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public Criterion generateCriterion(String[] keywords, int criterionType) throws DataSourceException {
        try {
            return removeInvalidConditions(_datasource.generateCriterion(keywords, criterionType));
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public Criterion generateCriterion(String[] keywords, int criterionType, String[] searchFields) throws DataSourceException {
        try {
            validateReadFields(searchFields);
            return _datasource.generateCriterion(keywords, criterionType, searchFields);
        } catch (SecureDataSourceException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, se.getMessage()));
            throw se;
        } catch (DataSourceException de) {
            _doctor.log(new LogEvent(this, LogEvent.CRIT, de.getMessage()));
            throw de;
        }
    }

    public void configure(Config config) throws ConfigException {
        configureDiagnostics(config.getChild(Diagnostics.CFG_BASE), config.getValue());

        if (config.hasChild(CFG_ENABLE_ADD))
            _addEnabled = config.getChild(CFG_ENABLE_ADD).getBooleanValue();

        if (config.hasChild(CFG_ENABLE_UPDATE))
            _updateEnabled = config.getChild(CFG_ENABLE_UPDATE).getBooleanValue();

        if (config.hasChild(CFG_ENABLE_DELETE))
            _deleteEnabled = config.getChild(CFG_ENABLE_DELETE).getBooleanValue();

        if (config.hasChild(CFG_ENABLE_SEARCH))
            _searchEnabled = config.getChild(CFG_ENABLE_SEARCH).getBooleanValue();

        if (config.hasChild(CFG_LOG_ADDS))
            _logAdds = config.getChild(CFG_LOG_ADDS).getBooleanValue();

        if (config.hasChild(CFG_LOG_UPDATES))
            _logUpdates = config.getChild(CFG_LOG_UPDATES).getBooleanValue();

        if (config.hasChild(CFG_LOG_DELETES))
            _logDeletes = config.getChild(CFG_LOG_DELETES).getBooleanValue();

        if (!config.hasChild(CFG_BASE))
            throw new ConfigException("No underlying DataSource config specified");

        _datasource = (DataSource) config.getChild(CFG_BASE).createAndConfigureInstance();

        Set fieldSet = getDatasourceFields();

        populateFieldIds(config.getChild(CFG_WRITABLE_FIELDS), _writableFields, fieldSet);
        populateFieldIds(config.getChild(CFG_READABLE_FIELDS), _readableFields, fieldSet);
        if (_readableFields.size() > 0)
            _readableFieldsArray = (String[]) _readableFields.toArray(new String[_readableFields.size()]);

        if (config.hasChild(CFG_QUOTA_CHECK)) {
            Config quotaConfig = config.getChild(CFG_QUOTA_CHECK);
            if (quotaConfig.hasChild(CFG_QUOTA_MAX)) {
                ArrayList fields = new ArrayList();
                _maxQuota = quotaConfig.getChild(CFG_QUOTA_MAX).getIntValue();
                populateFieldIds(quotaConfig.getChild(CFG_QUOTA_FIELDS), fields, _writableFields);
                if (fields.size() > 0)
                    _quotaFields = (String[]) fields.toArray(new String[fields.size()]);
            }
        }
    }

    protected void configureDiagnostics(Config config, String name) {
        String myClassName = getClass().getName();
        myClassName = myClassName.substring(myClassName.lastIndexOf(".") + 1);
        String[] diagFilter = name != null ? new String[] { CFG_BASE, myClassName, name }
        : new String[] { CFG_BASE, myClassName };

        _doctor = Diagnostics.getDoctor(diagFilter);

        if (config != null) {
            Config[] logConfigs = config.getChildren();
            for (int i = 0; logConfigs != null && i < logConfigs.length; i++) {
                if (logConfigs[i].getName().equals(Diagnostics.CFG_LOGLISTENER)) {
                    try {
                        String[] logFilter = diagFilter;
                        LogListener logListener = (LogListener) Class.forName(logConfigs[i].getType()).newInstance();
                        logListener.configure(logConfigs[i]);
                        if (logConfigs[i].getValue() != null) {
                            logFilter = new String[diagFilter.length + 1];
                            System.arraycopy(diagFilter, 0, logFilter, 0, diagFilter.length);
                            logFilter[diagFilter.length] = logConfigs[i].getValue();
                        }
                        Diagnostics.addLogListener(logListener, logFilter);
                    } catch (Exception e) {
                        String errmsg = "Failed to initialise LogListener " + logConfigs[i].getType() + ": " + e.getMessage();
                        _doctor.log(new LogEvent(this, LogEvent.WARN, errmsg));
                    }
                }
            }
        }
    }

    protected void populateFieldIds(Config config, AbstractCollection fieldIds, Set fields) throws ConfigException {
        if (config != null) {
            Config[] childConfigs = config.getChildren();
            for (int i = 0; i < childConfigs.length; i++) {
                if (childConfigs[i].getName().equals(CFG_FIELD_ID)) {
                    String field = childConfigs[i].getValue();
                    if (!fields.contains(field))
                        throw new ConfigException("Field not defined in underlying datasource: " + field);
                    fieldIds.add(field);
                }

            }
        }
    }

    protected Set getDatasourceFields() throws ConfigException {
        try {
            String[] fields = _datasource.getFields();
            Set fieldSet = new HashSet();
            for (int i = 0; i < fields.length; i++)
                fieldSet.add(fields[i]);
            return fieldSet;
        } catch (DataSourceException e) {
            throw new ConfigException("Failed to fields from underlying datasource", e);
        }
    }

    protected void validateReadFields(String[] fields) throws SecureDataSourceException {
        if (fields == null)
            return;

        for (int i = 0; i < fields.length; i++) {
            if (!_readableFields.contains(fields[i]))
                throw new SecureDataSourceException("Insufficient privileges read field: " + fields[i]);
        }
    }

    protected void validateWriteFields(Map rowValues) throws SecureDataSourceException {
        for (Iterator iter = rowValues.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            if (!_writableFields.contains(entry.getKey()))
                throw new SecureDataSourceException("Insufficient privileges to write to field: " + entry.getKey());
        }
    }

    protected Criterion removeInvalidConditions(Criterion criterion) throws SecureDataSourceException {
        if (criterion == null)
            return null;

        Condition[] conditions = criterion.getConditions();
        if (conditions != null) {
            ArrayList validConditions = new ArrayList(conditions.length);
            for (int i = 0; i < conditions.length; i++) {
                if (_readableFields.contains(conditions[i].getField()))
                    validConditions.add(conditions[i]);
            }
            conditions = validConditions.size() > 0 ? (Condition[]) validConditions.toArray(new Condition[validConditions.size()]) : null;
            criterion.setConditions(conditions);
        }

        Criterion[] subCriteria = criterion.getSubCriteria();
        if (subCriteria != null) {
            ArrayList validSubCriteria = new ArrayList(subCriteria.length);
            for (int i = 0; i < subCriteria.length; i++) {
                Criterion subCriterion = removeInvalidConditions(subCriteria[i]);
                if (subCriterion != null)
                    validSubCriteria.add(subCriterion);
            }
            subCriteria = validSubCriteria.size() > 0 ? (Criterion[]) validSubCriteria.toArray(new Criterion[validSubCriteria.size()]) : null;
            criterion.setSubCriteria(subCriteria);
        }

        return conditions != null || subCriteria != null ? criterion : null;
    }

    protected void validateCriterion(Criterion criterion) throws SecureDataSourceException {
        if (criterion == null)
            return;

        Condition[] conditions = criterion.getConditions();
        if (conditions != null) {
            for (int i = 0; i < conditions.length; i++) {
                if (!_readableFields.contains(conditions[i].getField()))
                    throw new SecureDataSourceException("Insufficient privileges read field: " + conditions[i].getField());
            }
        }

        Criterion[] subCriteria = criterion.getSubCriteria();
        if (subCriteria != null) {
            for (int i = 0; i < subCriteria.length; i++)
                validateCriterion(subCriteria[i]);
        }
    }

    protected void logValues(String msg, RowSet rs) throws SecureDataSourceException {
        try {
            HashMap values = new HashMap();
            ResultSetMetaData rsmd = rs.getMetaData();
            int numberOfColumns = rsmd.getColumnCount();

            while (rs.next()) {
                for (int i = 1; i <= numberOfColumns; i++)
                    values.put(rsmd.getColumnName(i), rs.getObject(i));
                _doctor.log(new LogEvent(this, LogEvent.INFO, msg, values));
            }

            rs.close();
        } catch (SQLException e) {
            throw new SecureDataSourceException("Failed to log values on " + msg + ": " + e.getMessage());
        }
    }
}