/*
 * SoapDataSource.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.datasource.soap;

import java.math.BigDecimal;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import javax.sql.RowSet;

import org.apache.soap.Constants;
import org.apache.soap.Fault;
import org.apache.soap.SOAPException;
import org.apache.soap.encoding.SOAPMappingRegistry;
import org.apache.soap.rpc.Call;
import org.apache.soap.rpc.Parameter;
import org.apache.soap.rpc.Response;
import org.apache.soap.transport.http.SOAPHTTPConnection;
import org.apache.soap.util.Bean;
import org.apache.soap.util.xml.Deserializer;
import org.apache.soap.util.xml.QName;
import org.apache.soap.util.xml.Serializer;
import org.w3c.dom.Node;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.datasource.Condition;
import com.eXtropia.datasource.Criterion;
import com.eXtropia.datasource.DataSource;
import com.eXtropia.datasource.DataSourceException;
import com.eXtropia.datasource.Query;
import com.eXtropia.soap.encoding.DecimalSerializer;
import com.eXtropia.soap.encoding.ExceptionSerializer;
import com.eXtropia.soap.encoding.StructDeserializer;
import com.eXtropia.soap.encoding.datasource.ConditionDeserializer;
import com.eXtropia.soap.encoding.datasource.CriterionDeserializer;
import com.eXtropia.soap.encoding.datasource.QueryDeserializer;
import com.eXtropia.soap.encoding.datasource.RowSetSerializer;
import com.eXtropia.util.ExtropiaException;

/**
 *
 * @author Carlo Ravagli
 * @version 0.1
 */
public class SoapDataSource extends DataSource {
    public static final String NS_URI_SOAP_DATASOURCE = "http://www.eXtropia.com/soap/encoding/datasource";

    public static final String CFG_SOAPSERVER_URL = "SoapServerUrl";
    public static final String CFG_SOAP_URN = "SoapUrn";
    public static final String CFG_TIMEOUT = "Timeout";
    public static final String CFG_PROXY_HOST = "ProxyHost";
    public static final String CFG_PROXY_PORT = "ProxyPort";
    public static final String CFG_USERNAME = "Username";
    public static final String CFG_PASSWORD = "Password";

    public static final String CFG_MAPPINGS = "Mappings";
    public static final String CFG_MAPPING = "Mapping";
    public static final String CFG_MAPPING_QNAME = "Qname";
    public static final String CFG_MAPPING_NAMESPACE = "Namespace";
    public static final String CFG_MAPPING_CLASS = "Class";
    public static final String CFG_MAPPING_SER = "Serializer";
    public static final String CFG_MAPPING_DESER = "Deserializer";

    private URL _url;
    private String _urn;
    private SOAPMappingRegistry _smr;

    private int _timeout;
    private String _proxyHost;
    private int _proxyPort;
    private String _username;
    private String _password;

    public SoapDataSource() {
        _smr = new SOAPMappingRegistry();

        com.eXtropia.soap.encoding.DateSerializer dateSer = new com.eXtropia.soap.encoding.DateSerializer();
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(Constants.NS_URI_1999_SCHEMA_XSD, "date"), java.util.Date.class, dateSer, dateSer);
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(Constants.NS_URI_1999_SCHEMA_XSD, "date"), java.sql.Date.class, dateSer, dateSer);
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(Constants.NS_URI_1999_SCHEMA_XSD, "time"), java.sql.Time.class, dateSer, dateSer);
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(Constants.NS_URI_1999_SCHEMA_XSD, "timeInstant"), java.sql.Timestamp.class, dateSer, dateSer);

        DecimalSerializer decimalSer = new DecimalSerializer();
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(Constants.NS_URI_1999_SCHEMA_XSD, "decimal"), BigDecimal.class, decimalSer, decimalSer);

        StructDeserializer mapSer = new StructDeserializer();
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(Constants.NS_URI_SOAP_ENC, "Struct"), Map.class, mapSer, mapSer);

        ConditionDeserializer conditionSer = new ConditionDeserializer();
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(NS_URI_SOAP_DATASOURCE, "Condition"), Condition.class, conditionSer, conditionSer);

        CriterionDeserializer criterionSer = new CriterionDeserializer();
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(NS_URI_SOAP_DATASOURCE, "Criterion"), Criterion.class, criterionSer, criterionSer);

        QueryDeserializer querySer = new QueryDeserializer();
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(NS_URI_SOAP_DATASOURCE, "Query"), Query.class, querySer, querySer);

        RowSetSerializer rowSetSer = new RowSetSerializer();
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(NS_URI_SOAP_DATASOURCE, "RowSet"), RowSet.class, rowSetSer, rowSetSer);

        ExceptionSerializer exceptionSer = new ExceptionSerializer();
        _smr.mapTypes(Constants.NS_URI_SOAP_ENC, new QName(NS_URI_SOAP_DATASOURCE, "Exception"), Exception.class, exceptionSer, exceptionSer);
    }

    private Call createCall(String method) {
        Call call = new Call();
        SOAPHTTPConnection st = new SOAPHTTPConnection();

        if (_proxyHost != null) {
            st.setProxyHost(_proxyHost);
            st.setProxyPort(_proxyPort);
        }

        if (_username != null) {
            st.setUserName(_username);
            st.setPassword(_password);
        }

        st.setTimeout(_timeout);

        call.setSOAPTransport(st);
        call.setSOAPMappingRegistry(_smr);
        call.setTargetObjectURI(_urn);
        call.setMethodName(method);
        call.setEncodingStyleURI(Constants.NS_URI_SOAP_ENC);

        return call;
    }

    private Object executeCall(Call call) throws DataSourceException {
        Response resp;
        try {
            resp = call.invoke(_url, call.getMethodName());
        } catch (SOAPException e) {
            throw new DataSourceException("SOAP Exception occured while performing " + call.getMethodName() + ": " + e.getMessage());
        }

        if (resp.generatedFault()) {
            Fault fault = resp.getFault();
            Vector detail = fault.getDetailEntries();
            /* if( detail != null )
            {
            	for( Iterator iter = detail.iterator(); iter.hasNext(); )
            	{
            		Object e = iter.next();
                                        String stringError = e.toString();
            
                                        //System.err.println("EXCEPTION THROWN: " + e);
                                        //System.err.println("EXCEPTION THROWN: " + e.getClass().getName() );
                                        if(stringError.indexOf("QuotaException") > -1)
                                          throw new com.eXtropia.datasource.secure.QuotaException(fault.getFaultString());
            
            		if( e instanceof DataSourceException)
            			throw (DataSourceException)e;
            
            		if( e instanceof RuntimeException )
            			throw (RuntimeException)e;
            	}
            }
            
            throw new DataSourceException( "SOAP Fault occured while performing " +
            	call.getMethodName() + ": " +
            	fault.getFaultString()); */

            /** Cassandra Pang 27/03/2002 
             *  The section above is replaced with the codes below. Instead of determining
             *  the exception explicitly and then decide on what to throw, SoapDataSource would 
             *  deserialize the exception from the fault object directly and throw it to the client.
             **/
            if (detail != null) {
                for (Iterator iter = detail.iterator(); iter.hasNext();) {
                    Object e = iter.next();
                    Bean exBean = (new ExceptionSerializer()).unmarshall(NS_URI_SOAP_DATASOURCE, new QName(NS_URI_SOAP_DATASOURCE, "Exception"), (Node) e, _smr, resp.getSOAPContext());
                    Object exBeanVal = exBean.value;
                    if (exBeanVal instanceof DataSourceException)
                        throw (DataSourceException) exBeanVal;
                    else
                        throw new DataSourceException("Wrapper", new ExtropiaException((Throwable) exBeanVal));
                }
            }
        }

        return resp.getReturnValue().getValue();
    }

    public int add(Map rowValues) throws DataSourceException {
        Call call = createCall("add");
        Vector params = new Vector();
        params.addElement(new Parameter("rowValues", Map.class, rowValues, null));
        call.setParams(params);

        return ((Number) executeCall(call)).intValue();
    }

    public int delete() throws DataSourceException {
        return ((Number) executeCall(createCall("delete"))).intValue();
    }

    public int delete(Condition condition) throws DataSourceException {
        Call call = createCall("delete");
        Vector params = new Vector();
        params.addElement(new Parameter("condition", Condition.class, condition, null));
        call.setParams(params);

        return ((Number) executeCall(call)).intValue();
    }

    public int delete(Criterion criterion) throws DataSourceException {
        Call call = createCall("delete");
        Vector params = new Vector();
        params.addElement(new Parameter("criterion", Criterion.class, criterion, null));
        call.setParams(params);

        return ((Number) executeCall(call)).intValue();
    }

    public int update(Map rowValues, Condition condition) throws DataSourceException {
        Call call = createCall("update");
        Vector params = new Vector();
        params.addElement(new Parameter("rowValues", Map.class, rowValues, null));
        params.addElement(new Parameter("condition", Condition.class, condition, null));
        call.setParams(params);

        return ((Number) executeCall(call)).intValue();
    }

    public int update(Map rowValues, Criterion criterion) throws DataSourceException {
        Call call = createCall("update");
        Vector params = new Vector();
        params.addElement(new Parameter("rowValues", Map.class, rowValues, null));
        params.addElement(new Parameter("criterion", Criterion.class, criterion, null));
        call.setParams(params);

        return ((Number) executeCall(call)).intValue();
    }

    public RowSet search(Query query) throws DataSourceException {
        Call call = createCall("search");
        Vector params = new Vector();
        params.addElement(new Parameter("query", Query.class, query, null));
        call.setParams(params);

        return (RowSet) executeCall(call);
    }

    public int count() throws DataSourceException {
        return ((Number) executeCall(createCall("count"))).intValue();
    }

    public int count(Condition condition) throws DataSourceException {
        Call call = createCall("count");
        Vector params = new Vector();
        params.addElement(new Parameter("condition", Condition.class, condition, null));
        call.setParams(params);

        return ((Number) executeCall(call)).intValue();
    }

    public int count(Criterion criterion) throws DataSourceException {
        Call call = createCall("count");
        Vector params = new Vector();
        params.addElement(new Parameter("criterion", Criterion.class, criterion, null));
        call.setParams(params);

        return ((Number) executeCall(call)).intValue();
    }

    public String[] getKeyFields() throws DataSourceException {
        return (String[]) executeCall(createCall("getKeyFields"));
    }

    public String[] getFields() throws DataSourceException {
        return (String[]) executeCall(createCall("getFields"));
    }

    public Criterion generateCriterion(String keyword) throws DataSourceException {
        Call call = createCall("generateCriterion");
        Vector params = new Vector();
        params.addElement(new Parameter("keyword", String.class, keyword, null));
        call.setParams(params);

        return (Criterion) executeCall(call);
    }

    public Criterion generateCriterion(String keyword, String[] searchFields) throws DataSourceException {
        Call call = createCall("generateCriterion");
        Vector params = new Vector();
        params.addElement(new Parameter("keyword", String.class, keyword, null));
        params.addElement(new Parameter("searchFields", String[].class, searchFields, null));
        call.setParams(params);

        return (Criterion) executeCall(call);
    }

    public Criterion generateCriterion(String[] keywords, int criterionType) throws DataSourceException {
        Call call = createCall("generateCriterion");
        Vector params = new Vector();
        params.addElement(new Parameter("keywords", String[].class, keywords, null));
        params.addElement(new Parameter("criterionType", Integer.class, new Integer(criterionType), null));
        call.setParams(params);

        return (Criterion) executeCall(call);
    }

    public Criterion generateCriterion(String[] keywords, int criterionType, String[] searchFields) throws DataSourceException {
        Call call = createCall("generateCriterion");
        Vector params = new Vector();
        params.addElement(new Parameter("keywords", String[].class, keywords, null));
        params.addElement(new Parameter("criterionType", Integer.class, new Integer(criterionType), null));
        params.addElement(new Parameter("searchFields", String[].class, searchFields, null));
        call.setParams(params);

        return (Criterion) executeCall(call);
    }

    public void configure(Config config) throws ConfigException {
        if (_url != null || _urn != null)
            throw new ConfigException("Soap DataSource is already configured");

        if (!config.hasChild(CFG_SOAPSERVER_URL) || !config.hasChild(CFG_SOAP_URN))
            throw new ConfigException("Soap DataSource requies soap server url and soap urn config");

        try {
            _url = new URL(config.getChild(CFG_SOAPSERVER_URL).getValue());
            _urn = config.getChild(CFG_SOAP_URN).getValue();
        } catch (MalformedURLException e) {
            throw new ConfigException("Invalid Soap URL: " + e.getMessage());
        }

        if (config.hasChild(CFG_TIMEOUT)) {
            _timeout = Integer.parseInt(config.getChild(CFG_TIMEOUT).getValue());
        }

        if (config.hasChild(CFG_PROXY_HOST)) {
            Config proxyConfig = config.getChild(CFG_PROXY_HOST);
            _proxyHost = proxyConfig.getValue();

            if (proxyConfig.hasChild(CFG_PROXY_PORT)) {
                _proxyPort = Integer.parseInt(proxyConfig.getChild(CFG_PROXY_PORT).getValue());
            }
        }

        if (config.hasChild(CFG_USERNAME)) {
            Config authConfig = config.getChild(CFG_USERNAME);
            _username = authConfig.getValue();

            if (authConfig.hasChild(CFG_PASSWORD)) {
                _password = authConfig.getChild(CFG_PASSWORD).getValue();
            }
        }

        configureSoapRegistry(config.getChild(CFG_MAPPINGS));
    }

    protected void configureSoapRegistry(Config config) throws ConfigException {
        if (config == null)
            return;

        try {
            Config[] mappings = config.getChildren();
            for (int i = 0; i < mappings.length; i++) {
                QName qname;
                Class javaType;
                String namespace = null;
                Serializer ser = null;
                Deserializer deser = null;

                if (!mappings[i].hasChild(CFG_MAPPING_QNAME))
                    throw new ConfigException("Invalid mapping, " + CFG_MAPPING_QNAME + " not specifed");

                Config qnameConfig = mappings[i].getChild(CFG_MAPPING_QNAME);

                if (qnameConfig.hasChild(CFG_MAPPING_NAMESPACE))
                    namespace = qnameConfig.getChild(CFG_MAPPING_NAMESPACE).getValue();

                qname = new QName(namespace != null ? namespace : NS_URI_SOAP_DATASOURCE, qnameConfig.getValue());

                javaType = Class.forName(mappings[i].getChildValue(CFG_MAPPING_CLASS, true));
                if (mappings[i].hasChild(CFG_MAPPING_SER))
                    ser = (Serializer) Class.forName(mappings[i].getChildValue(CFG_MAPPING_SER)).newInstance();

                if (mappings[i].hasChild(CFG_MAPPING_DESER))
                    deser = (Deserializer) Class.forName(mappings[i].getChildValue(CFG_MAPPING_DESER)).newInstance();

                _smr.mapTypes(Constants.NS_URI_SOAP_ENC, qname, javaType, ser, deser);
            }
        } catch (Exception e) {
            throw new ConfigException("Failed to configure soap registry: " + e.getMessage());
        }
    }

}