/* 
 * Diagnostics.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.diagnostics;

import java.util.ArrayList;
import java.util.StringTokenizer;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;

/**
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 0.1
 */
public class Diagnostics {
    public static final String CFG_BASE = "Diagnostics";
    public static final String CFG_DEFAULT = "DefaultTraceLevel";
    public static final String CFG_LOGLISTENER = "LogListener";
    public static final String CFG_TRACELISTENER = "TraceListener";
    public static final String CFG_TRACE_INLINE = "TraceInline";
    public static final String CFG_LOG_INLINE = "LogInline";

    /**
     * The singleton instance of diagnostics
     */
    //protected static Diagnostics _instance;

    protected static Doctor _rootNode = new Doctor(null, "", new DiagnosticsEventQueue(true, true));

    /** Creates new Diagnostics */
    private Diagnostics() {
    }

    public static void addLogListener(LogListener logListener) {
        addLogListener(logListener, new String[0]);
    }

    public static void addLogListener(LogListener logListener, String[] filters) {
        Doctor d = getDoctor(filters, false);
        if (d != null) {
            d.addLogListener(logListener);
        } // Maybe add a failure error msg??

    }

    public static void addTraceListener(TraceListener traceListener) {
        addTraceListener(traceListener, new String[0]);
    }

    public static void addTraceListener(TraceListener traceListener, String[] filters) {
        Doctor d = getDoctor(filters, false);
        if (d != null) {
            d.addTraceListener(traceListener);
        }
    }

    public static void configureFactory(Config config) throws ConfigException {
        Config subConfig;
        //check that the base of this is called Diagnostics
        if (config.getName().equals(CFG_BASE)) {
            boolean logInline, traceInline;
            //see if we should be diagnosing inline
            try {
                logInline = Boolean.valueOf(config.getChild(CFG_LOG_INLINE).getValue()).booleanValue();
            } catch (Exception e) {
                logInline = false;
            }

            try {
                traceInline = Boolean.valueOf(config.getChild(CFG_TRACE_INLINE).getValue()).booleanValue();
            } catch (Exception e) {
                traceInline = false;
            }
            _rootNode = new Doctor(null, "", new DiagnosticsEventQueue(logInline, traceInline));

            //check for a default level
            subConfig = config.getChild(CFG_DEFAULT);
            if (subConfig != null) {
                for (short i = 0; i < TraceEvent.LEVEL_NAMES.length; i++) {
                    if (subConfig.getValue().equals(TraceEvent.LEVEL_NAMES[i])) {
                        _rootNode.setTraceLevel(i, true, true);
                    }
                }
            }
            //else
            //ah well, no default specified

            //now go through each child, and configure listeners
            Config[] children = config.getChildren();
            for (int i = 0; i < children.length; i++) {
                if (children[i].getName().equals(CFG_LOGLISTENER)) {
                    configureLogListener(children[i]);
                } else if (children[i].getName().equals(CFG_TRACELISTENER)) {
                    configureTraceListener(children[i]);
                }
            }
        } //end if CFG_BASE
    }

    private static void configureLogListener(Config config) throws ConfigException {
        try {
            LogListener logListener = (LogListener) Class.forName(config.getType()).newInstance();
            logListener.configure(config);
            addLogListener(logListener);
        } catch (Exception e) {
            System.out.println("Failed to initialise LogListener: " + config.getName() + " of type " + config.getType());
            e.printStackTrace();
        }
    }

    private static void configureTraceListener(Config config) throws ConfigException {
        try {
            TraceListener traceListener = (TraceListener) Class.forName(config.getType()).newInstance();
            traceListener.configure(config);
            addTraceListener(traceListener);
        } catch (Exception e) {
            System.out.println("Failed to initialise TraceListener: " + config.getName() + " of type " + config.getName());
            e.printStackTrace();
        }
    }

    public static Doctor getDoctor(String[] filters) {
        if (_rootNode != null) {
            Doctor d = getDoctor(filters, true);
            d.setFilters(filters);
            return d;
        } else {
            throw new IllegalStateException("Diagnostics not configured");
        }
    }

    public static Doctor getDoctor(String path) {
        ArrayList filters = new ArrayList();
        StringTokenizer st = new StringTokenizer(path, "/");
        while (st.hasMoreTokens())
            filters.add(st.nextToken());

        return getDoctor((String[]) filters.toArray(new String[filters.size()]));
    }

    public static String[] getFilters() {
        return getFilters(new String[0]);
    }

    public static String[] getFilters(String[] filters) {
        Doctor d = getDoctor(filters, false);
        if (d != null) {
            //System.out.println("Got a result");
            //System.out.println("" + (d == _rootNode));
            return d.getChildDoctorNames();
        } else {
            //System.out.println("Can't find it");
            return new String[0];
        }
    }

    public static void removeLogListener(LogListener logListener) {
        removeLogListener(logListener, new String[0]);
    }

    public static void removeLogListener(LogListener logListener, String[] filters) {
        Doctor d = getDoctor(filters, false);
        if (d != null) {
            d.removeLogListener(logListener);
        }
    }

    public static void removeTraceListener(TraceListener traceListener) {
        removeTraceListener(traceListener, new String[0]);
    }

    public static void removeTraceListener(TraceListener traceListener, String[] filters) {
        Doctor d = getDoctor(filters, false);
        if (d != null) {
            d.removeTraceListener(traceListener);
        }
    }

    public static void setTraceLevel(short traceLevel) {
        setTraceLevel(traceLevel, new String[0]);
    }

    public static void setTraceLevel(short traceLevel, String[] filters) {
        Doctor d = getDoctor(filters, false);
        if (d != null) {
            d.setTraceLevel(traceLevel, true, false);
        }
    }

    public static void reset(short traceLevel) {
        reset(traceLevel, new String[0]);
    }

    public static void reset(short traceLevel, String[] filters) {
        Doctor d = getDoctor(filters, false);
        if (d != null) {
            d.setTraceLevel(traceLevel, true, true);
        }
    }

    protected static synchronized Doctor getDoctor(String[] filters, boolean add) {
        Doctor node = _rootNode;
        int i = -1;
        while (++i < filters.length && node != null) {
            //System.out.println("FINDING FILTER: " + filters[i]);
            node = node.findChildDoctor(filters[i], add);
        }
        return node;
    }
}