/* 
 * Doctor.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.diagnostics;

import java.lang.ref.WeakReference;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

/** 
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 0.1
 */
public class Doctor {
    private String[] _filters;

    private boolean _explicitLevel;

    private short _traceLevel = TraceEvent.NOT_SPECIFIED;

    private HashSet _traceListeners = new HashSet(1);
    private HashSet _logListeners = new HashSet(1);

    private Doctor _parentNode;
    private DiagnosticsEventQueue _eventQueue;

    Map _childNodes = new HashMap(); //Map of String to weak references to Doctors

    Doctor(Doctor parent, String name, DiagnosticsEventQueue eventQueue) {
        _parentNode = parent == null ? this : parent;
        _eventQueue = eventQueue;
        _filters = new String[] { name };
        _traceLevel = _parentNode._traceLevel;
    }

    void setFilters(String[] filters) {
        _filters = filters;
    }

    public String[] getFilters() {
        return _filters;
    }

    synchronized void addTraceListener(TraceListener traceListener) {
        synchronized (_traceListeners) {
            _traceListeners.add(traceListener);
        }
    }

    synchronized void addLogListener(LogListener logListener) {
        synchronized (_logListeners) {
            _logListeners.add(logListener);
        }
    }

    public boolean doTrace(int level) {
        return level <= _traceLevel;
    }

    synchronized Doctor findChildDoctor(String oneToFind, boolean add) {
        Doctor childNode = null;
        try {
            childNode = (Doctor) ((WeakReference) _childNodes.get(oneToFind)).get();
        } catch (Exception e) {
        }
        if (childNode == null && add) {
            childNode = new Doctor(this, oneToFind, _eventQueue);
            _childNodes.put(oneToFind, new WeakReference(childNode));
        }
        return childNode;
    }

    int getTraceLevel() {
        return _traceLevel;
    }

    String[] getChildDoctorNames() {
        Object[] childNames = _childNodes.keySet().toArray();
        String[] names = new String[childNames.length];
        System.arraycopy(childNames, 0, names, 0, childNames.length);
        return names;
    }

    void handleEvent(LogEvent logEvent) {
        HashSet s;
        synchronized (_logListeners) {
            s = (HashSet) _logListeners.clone();
        }
        for (Iterator iter = s.iterator(); iter.hasNext();) {
            ((LogListener) iter.next()).log(logEvent);
        }
        if (_parentNode != this) {
            _parentNode.handleEvent(logEvent);
        }

    }

    void handleEvent(TraceEvent traceEvent) {
        HashSet s;
        synchronized (_traceListeners) {
            s = (HashSet) _traceListeners.clone();
        }
        for (Iterator iter = s.iterator(); iter.hasNext();) {
            ((TraceListener) iter.next()).trace(traceEvent);
        }
        if (_parentNode != this) {
            _parentNode.handleEvent(traceEvent);
        }
    }

    public void log(LogEvent logEvent) {
        logEvent.setDoctor(this);
        _eventQueue.addEvent(logEvent);
        trace(new TraceEvent(logEvent.getSource(), TraceEvent.INFO, "LOG: [" + logEvent.getMessage() + "]"));
    }

    void removeLogListener(LogListener logListener) {
        synchronized (_logListeners) {
            _logListeners.remove(logListener);
        }
    }

    void removeTraceListener(TraceListener traceListener) {
        synchronized (_traceListeners) {
            _traceListeners.remove(traceListener);
        }
    }

    synchronized void setTraceLevel(short traceLevel, boolean explicitLevel, boolean resetChildren) {
        if (explicitLevel || !_explicitLevel) {
            _explicitLevel = explicitLevel;
            //now we can set our children levels
            //if we have a new level (explicit or not)
            if (traceLevel != _traceLevel || (explicitLevel && resetChildren)) {
                _traceLevel = traceLevel;
                for (Iterator iter = _childNodes.values().iterator(); iter.hasNext();) {
                    ((Doctor) ((WeakReference) iter.next()).get()).setTraceLevel(_traceLevel, explicitLevel && resetChildren, true);
                }
            }
        }
        //else we already have an explicit level of our own so don't do anything
    }

    public void trace(TraceEvent traceEvent) {
        if (doTrace(traceEvent.getLevel())) {
            traceEvent.setDoctor(this);
            _eventQueue.addEvent(traceEvent);
        }
    }

    protected void finalize() {
        //need to remove myself from my parents hash
        if (_parentNode != this) {
            _parentNode._childNodes.remove(_filters[_filters.length - 1]);
        }
    }
}