/* 
 * LogEvent.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.diagnostics;

import java.util.Date;
import java.util.Map;

/** 
 * A class defining an event that requires logging
 * <P>
 * This object is passed around to the various {@link LogListener} that are currently
 * registered with the {@link Diagnostics} object
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 0.1
 */
public class LogEvent extends DiagnosticsEvent implements java.io.Serializable {
    /**
     * The level of an informational <CODE>LogEvent</CODE>
     */
    public static final short INFO = 3;

    /**
     * The level of a warning <CODE>LogEvent</CODE>
     */
    public static final short WARN = 2;

    /**
     * The level of a critical <CODE>LogEvent</CODE>
     */
    public static final short CRIT = 1;

    /**
     * The default level of a <CODE>LogEvent</CODE>
     */
    public static final short NOT_SPECIFIED = 0;

    /**
     * The most verbose level of a<CODE>LogEvent</CODE>
     */
    public static final short MAX_LEVEL = 3;

    /**
     * The most critical level of a <CODE>LogEvent</CODE>
     */
    public static final short MIN_LEVEL = 1;

    /**
     * String representations of the various levels of a <CODE>LogEvent</CODE>
     */
    static final String[] LEVEL_NAMES = new String[] { "N/A ", "CRIT", "WARN", "INFO" };

    /**
     * The level of the log event
     */
    private short _level;

    /**
     * The message of the log event
     */
    private String _message;

    /**
     * A Map of values to be inserted into the message upon formatting
     */
    private Map _map;

    /** Creates new LogEvent */
    public LogEvent(Object source, short level, String message) {
        super(source, LOG);
        _level = level;
        _message = message;
    }

    /** Creates new LogEvent */
    public LogEvent(Object source, short level, Map map) {
        super(source, LOG);
        _level = level;
        _map = map;
    }

    /** Creates new LogEvent */
    public LogEvent(Object source, short level, String message, Map map) {
        super(source, LOG);
        _level = level;
        _message = message;
        _map = map;
    }

    /**
     * Gets the level of the <CODE>LogEvent</CODE>
     * @return the level
     */
    public short getLevel() {
        return _level;
    }

    /**
     * Gets a friendly name for the level of the <CODE>LogEvent</CODE>
     * @return the level
     */
    public String getLevelName() {
        return LEVEL_NAMES[_level];
    }

    /**
     * Gets the map of values to be used in the message
     * @return the map
     */
    public Map getMap() {
        return _map;
    }

    /**
     * Gets the message of the <CODE>LogEvent</CODE>
     * @return the message
     */
    public String getMessage() {
        return _message;
    }

    /**
     * Returns a description of the <CODE>LogEvent</CODE>
     */
    public String toString() {
        return "" + new Date() + " " + getLevelName() + ": [" + getFilterString() + "] " + getMessage();
    }
}