/*
 * SysLogListener.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.diagnostics;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.ice.syslog.Syslog;
import com.ice.syslog.SyslogDefs;

/**
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 0.1
 */
public class SysLogListener extends ListenerSupport implements LogListener, TraceListener {
    public static final String CFG_FACILITY = "Facility";
    public static final String CFG_PRIORITY = "Priority";
    public static final String CFG_HOSTNAME = "Host";
    public static final String CFG_PORT = "Port";

    // Normally the priority will be something like INFO, WARNING, EMERG etc.
    // But "AUTO" will automatically take the prioty of the log event and attempt
    // to map it to the syslog object's appropriate priority

    private String _facility = "USER";
    private String _priority = "AUTO";

    private static final String SYSLOG_NAME = "eXtropia ADT";
    private static final int SYSLOG_FLAGS = 0;

    private Syslog _syslog = null;

    public SysLogListener() {

    }

    public void configure(Config config) throws ConfigException {
        //get out the file name
        super.configure(config);
        Config portConfig = config.getChild(CFG_PORT);
        Config hostConfig = config.getChild(CFG_HOSTNAME);
        Config facilityConfig = config.getChild(CFG_FACILITY);
        Config priorityConfig = config.getChild(CFG_PRIORITY);

        int port = -1;
        if (portConfig != null) {
            port = portConfig.getIntValue();
        }

        String hostname = null;
        if (hostConfig != null) {
            hostname = hostConfig.getValue();
        }

        if (facilityConfig != null) {
            _facility = facilityConfig.getValue();
        }

        if (priorityConfig != null) {
            _priority = priorityConfig.getValue();
        }

        //
        try {
            if (hostname == null && port == -1) {
                _syslog = new Syslog(SYSLOG_NAME, SYSLOG_FLAGS);
            } else if (hostname != null && port == -1) {
                _syslog = new Syslog(hostname, SyslogDefs.DEFAULT_PORT, SYSLOG_NAME, SYSLOG_FLAGS);
            } else if (hostname != null && port != -1) {
                _syslog = new Syslog(hostname, port, SYSLOG_NAME, SYSLOG_FLAGS);
            }
        } catch (Exception e) {
            throw new ConfigException("Failed to create Syslog object");
        }
    }

    public void log(LogEvent logEvent) {
        if (okLogLevel(logEvent.getLevel())) {
            String priority = _priority;
            if (_priority.equals("AUTO")) {
                if (logEvent.getLevel() == LogEvent.CRIT) {
                    priority = "CRIT";
                } else if (logEvent.getLevel() == LogEvent.WARN) {
                    priority = "WARNING";
                } else {
                    priority = "INFO";
                }
            }
            int intPriority;
            int intFacility;
            try {
                intPriority = SyslogDefs.getPriority(priority);
                intFacility = SyslogDefs.getFacility(_facility);
            } catch (java.text.ParseException e) {
                e.printStackTrace();
                intFacility = SyslogDefs.LOG_USER;
                intPriority = SyslogDefs.LOG_INFO;
            }

            try {
                _syslog.syslog(intFacility, intPriority, formatMessage(logEvent.getMessage(), logEvent.getLevelName(), logEvent.getMap()));
            } catch (com.ice.syslog.SyslogException e) {
                e.printStackTrace();
            }
        }
    }

    public void trace(TraceEvent traceEvent) {
        String priority = _priority;
        if (_priority.equals("AUTO")) {
            if (traceEvent.getLevel() == TraceEvent.CRIT) {
                _priority = "CRIT";
            } else if (traceEvent.getLevel() == TraceEvent.WARN) {
                _priority = "WARNING";
            } else {
                _priority = "INFO";
            }
        }
        int intPriority;
        int intFacility;
        try {
            intPriority = SyslogDefs.getPriority(priority);
            intFacility = SyslogDefs.getFacility(_facility);
        } catch (java.text.ParseException e) {
            e.printStackTrace();
            intFacility = SyslogDefs.LOG_USER;
            intPriority = SyslogDefs.LOG_INFO;
        }

        try {
            _syslog.syslog(intFacility, intPriority, formatMessage(traceEvent.getMessage(), traceEvent.getLevelName(), null));
        } catch (com.ice.syslog.SyslogException e) {
            e.printStackTrace();
        }

        if (traceEvent.getException() != null)
            traceEvent.getException().printStackTrace();
    }
}
