/* 
 * TraceEvent.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.diagnostics;

import java.util.Date;

/** 
 * A class defining an event that requires tracing
 * <P>
 * This object is passed around to the various {@link TraceListener} that are currently
 * registered with the {@link Diagnostics} object
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 0.1
 */
public class TraceEvent extends DiagnosticsEvent implements java.io.Serializable {
    /**
     * The most verbose level of a <CODE>TraceEvent</CODE> for debugging at development time
     */
    public static final short DEVELOPER = 5;

    /**
     * The most verbose level of a <CODE>TraceEvent</CODE>
     */
    public static final short VERBOSE = 4;

    /**
     * The level of an information <CODE>TraceEvent</CODE>
    */
    public static final short INFO = 3;

    /**
     * The level of a warning <CODE>TraceEvent</CODE>
    */
    public static final short WARN = 2;

    /**
     * The level of a critical <CODE>TraceEvent</CODE>
    */
    public static final short CRIT = 1;

    /**
     * The default level of a <CODE>TraceEvent</CODE>
    */
    public static final short NOT_SPECIFIED = 0;

    /**
     * The most verbose level of a <CODE>TraceEvent</CODE>
     */
    public static final short MAX_LEVEL = 5;

    /**
     * The most important level of a <CODE>TraceEvent</CODE>
     */
    public static final short MIN_LEVEL = 1;

    /**
     * String representations of the various levels of a <CODE>TraceEvent</CODE>
     */
    static final String[] LEVEL_NAMES = new String[] { "NONE", "CRIT", "WARN", "INFO", "VERBOSE", "DEVELOPER" };

    /**
     * The level of the trace event
     */
    private short _level;

    /**
     * The message of the trace event
     */
    private String _message;

    /**
     * An exception associated with the trace event
     */
    private Exception _exception;

    /**
     * The Date that the trace event occured
     */
    private Date _date = new Date();

    /** Creates new TraceEvent */
    public TraceEvent(Object source, short level, String message) {
        super(source, TRACE);
        _level = level;
        _message = message;
    }

    /** Creates new TraceEvent */
    public TraceEvent(Object source, short level, String message, Exception e) {
        super(source, TRACE);
        _level = level;
        _message = message;
        _exception = e;
    }

    /**
     * Gets the level of the trace event
     * @return the specified level
     */
    public short getLevel() {
        return _level;
    }

    /**
     * Gets a friendly name for the level of the trace event
     * @return the level
     */
    public String getLevelName() {
        return LEVEL_NAMES[_level];
    }

    /**
     * Gets the message of the trace event
     * @return the message
     */
    public String getMessage() {
        return _message;
    }

    /**
     * Gets the associated exception of the trace event
     * @return the associated exception or null if not specified
     */
    public Exception getException() {
        return _exception;
    }

    /**
     * Returns a description of the <CODE>TraceEvent</CODE>
     */
    public String toString() {
        return "" + _date + " " + getLevelName() + ": [" + getFilterString() + "] " + getMessage();
    }
}