/*
 * WebPageChecker.java
 *
 * Created on October 12, 2000, 3:42 PM
 */

package com.eXtropia.diagnostics.checkers;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;
import com.eXtropia.diagnostics.LogListener;

/** 
 *
 * @author  Carlo Ravagli (carlo.ravagli@extropia.com)
 * @version 
 */
public class WebPageChecker implements Configurable, Runnable {
    public static final String CFG_BASE = "WebPageChecker";
    public static final String CFG_URL = "Url";
    public static final String CFG_CHECK_INTERVAL = "CheckInterval";
    public static final String CFG_FAILURE_INTERVAL = "FailureInterval";
    public static final String CFG_FAILURE_THRESHOLD = "FailureThreshold";

    private boolean _configured;

    private URL _url;
    private Doctor _doctor;

    private long _lastChecked;
    private long _checkInterval;
    private long _failureInterval;
    private long _failureThreshold;
    private int _failureCount;

    public synchronized void configure(Config config) throws ConfigException {
        if (!config.getName().equals(CFG_BASE))
            throw new ConfigException("Cannot configure WebPageCheck with configuartion: " + config.getName());

        if (_configured)
            throw new ConfigException("WebPagerCheck already configured");

        configureUrl(config.getChild(CFG_URL));

        _checkInterval = getLongFromConfig(config.getChild(CFG_CHECK_INTERVAL), 900000);
        _failureInterval = getLongFromConfig(config.getChild(CFG_FAILURE_INTERVAL), 30000);
        _failureThreshold = getLongFromConfig(config.getChild(CFG_FAILURE_THRESHOLD), 3);

        configureDiagnostics(config.getChild(Diagnostics.CFG_BASE));

        new Thread(this).start();
        _configured = true;
    }

    public void run() {
        boolean run = true;
        _failureCount = 0;
        do {
            HttpURLConnection conn = null;

            try {
                _lastChecked = (long) (System.currentTimeMillis() / 1000);
                conn = (HttpURLConnection) _url.openConnection();
                conn.connect();
                if (conn.getResponseCode() != HttpURLConnection.HTTP_OK)
                    throw new IOException("Http request to " + _url + " was unsuccessful - " + conn.getResponseMessage());

                BufferedReader content = new BufferedReader(new InputStreamReader(conn.getInputStream()));
                for (String line = content.readLine(); line != null; line = content.readLine()) {
                    if (line.indexOf("<H1>Software error:</H1>") != -1)
                        throw new IOException("Http request to " + _url + " was unsuccessful due to software error");
                }

                if (_failureCount >= _failureThreshold)
                    _doctor.log(new LogEvent(this, LogEvent.INFO, "WebPage check " + _url + " succeeded."));
                _failureCount = 0;
            } catch (IOException ioe) {
                String errmsg = "WebPage check failed: " + ioe.getMessage();
                _doctor.log(new LogEvent(this, ++_failureCount == _failureThreshold ? LogEvent.CRIT : LogEvent.WARN, errmsg));
            } finally {
                if (conn != null)
                    conn.disconnect();
            }

            try {
                Thread.sleep(_failureCount > 0 ? _failureInterval : _checkInterval);
            } catch (InterruptedException ie) {
                run = false;
            }
        } while (run);
    }

    private void configureUrl(Config config) throws ConfigException {
        if (config == null || config.getValue() == null)
            throw new ConfigException("Url not specifed for WebPageChecker");

        try {
            _url = new URL(config.getValue());
        } catch (MalformedURLException e) {
            throw new ConfigException("Invalid URL format: " + e.getMessage());
        }
    }

    private void configureDiagnostics(Config config) {
        String[] diagFilter = { CFG_BASE, _url.toString()};

        _doctor = Diagnostics.getDoctor(diagFilter);

        if (config != null) {
            Config[] logConfigs = config.getChildren();
            for (int i = 0; logConfigs != null && i < logConfigs.length; i++) {
                if (logConfigs[i].getName().equals(Diagnostics.CFG_LOGLISTENER)) {
                    try {
                        LogListener logListener = (LogListener) Class.forName(logConfigs[i].getType()).newInstance();
                        logListener.configure(logConfigs[i]);
                        Diagnostics.addLogListener(logListener, diagFilter);
                    } catch (Exception e) {
                        String errmsg = "Failed to initialise LogListener " + logConfigs[i].getType() + ": " + e.getMessage();
                        _doctor.log(new LogEvent(this, LogEvent.WARN, errmsg));
                    }
                }
            }
        }
    }

    private long getLongFromConfig(Config config, long defaultValue) {
        return config != null && config.getValue() != null ? Long.parseLong(config.getValue()) : defaultValue;
    }

    public int getFailureCount() {
        return _failureCount;
    }

    public long getLastChecked() {
        return _lastChecked;
    }

    public void setCheckInterval(long checkInterval) {
        _checkInterval = checkInterval;
    }

    public long getCheckInterval() {
        return _checkInterval;
    }

    public void setFailureInterval(long failureInterval) {
        _failureInterval = failureInterval;
    }

    public long getFailureInterval() {
        return _failureInterval;
    }

    public void setFailureThreshold(long failureThreshold) {
        _failureThreshold = failureThreshold;
    }

    public long getFailureThreshold() {
        return _failureThreshold;
    }

}