/* 
 * Mailer.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.mail;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.util.ExtropiaException;

/** 
 * An abstract class defining an eXtropia mailer.
 *
 * All subclasses of <CODE>Mailer</CODE> have to define {@link #handleSend}, a method
 * that sends the email given or returns false.
 * <P>
 * In order to use <CODE>Mailer</CODE>, first call the configureFactory with a {@link Config}
 * that contains enough detail to configure the required subclass e.g. {@link SmtpMailer}
 * Then, the static method {@link #send} can be called on the Mailer class.<BR><BR>
 * <CODE>
 * Config myConfig= new Config(new FileInputStream(myFileLocation));<BR>
 * Mailer.configureFactory(config);<BR>
 * Mailer.send(myMailMessage);<BR>
 * </CODE>
 * <P>
 * A sample config file to configure a mailer can be found in the {@link SmtpMailer} class
 * 
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 0.1
 */
public abstract class Mailer implements Configurable {
    /**
     * The expected config base
     */
    public static final String CFG_BASE = "Mailer";

    /**
     * The singleton instance of Mailer
     */
    private static Mailer _instance;

    /**
    	 * Protected constructor for Mailer
    	 */
    protected Mailer() {
    }

    /**
     * Instantiates a Mailer from the specified <CODE>Config</CODE>
     */
    public static void configureMailer(Config config) throws ConfigException {
        if (config.getName().equals(CFG_BASE)) {
            //choose the class to instantiate from the config and return it
            try {
                _instance = (Mailer) Class.forName(config.getType()).newInstance();
                _instance.configure(config);
            } catch (ConfigException e) {
                _instance = null;
                throw new ConfigException("Failed to configure Mailer: " + config.getType());
            } catch (Exception e) {
                e.printStackTrace();
                _instance = null;
                throw new ConfigException("Failed to create Mailer: " + config.getType());
            }
        } else {
            throw new ConfigException("Incorrect config base supplied: " + config.getName());
        }
    }

    /**
     * Abstract method that subclass must implement to send specified messages
     */
    protected abstract boolean handleSend(Message msg);

    /**
     * Sends the specified <CODE>Message</CODE> using the configured instance
     */
    public static boolean send(Message msg) throws ExtropiaException {
        try {
            return _instance.handleSend(msg);
        } catch (NullPointerException npe) {
            throw new ExtropiaException("Mailer: no mailer configured");
        }
    }
}