/*
 * Message.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.mail;

import java.io.File;
import java.util.Enumeration;
import java.util.Vector;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;

/** 
 * A class representing an email message
 * <P>
 * This class and package is not intended to compete with <CODE>javax.mail.Message</CODE> and the JavaMail API.
 * The purpose of the this class is to provide a simpler abstraction for the sending of email
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 0.1
 */
public class Message implements Configurable, Cloneable {
    public static final String CFG_BASE = "Message";
    public static final String CFG_FROM = "From";
    public static final String CFG_REPLYTO = "ReplyTo";
    public static final String CFG_RECIPIENT = "Recipient";
    public static final String CFG_TO = "To";
    public static final String CFG_CC = "Cc";
    public static final String CFG_BCC = "Bcc";
    public static final String CFG_SUBJECT = "Subject";
    public static final String CFG_CONTENT = "Content";
    public static final String CFG_SIGNATURE = "Signature";
    public static final String CFG_ATTACHMENT = "Attachment";
    /**
     * The address type of a recipient
     * @see #getRecipients
     * @see #setRecipients
     * @see #addRecipient
     */
    public static final short TO_RECIPIENT = 0;
    /**
     * The address type of a recipient
     * @see #getRecipients
     * @see #setRecipients
     * @see #addRecipient
     */
    public static final short CC_RECIPIENT = 1;
    /**
     * The address type of a recipient
     * @see #getRecipients
     * @see #setRecipients
     * @see #addRecipient
     */
    public static final short BCC_RECIPIENT = 2;
    /**
     * The email address of the sender
     */
    private String _from;
    /**
     * The email address any replies should be sent to
     */
    private String _replyTo;
    /**
     * An array of Vectors containing recipient addresses.
     *
     * The array is indexed by {@link #TO_RECIPIENT}, {@link #CC_RECIPIENT} and {@link #BCC_RECIPIENT}
     */
    private Vector[] _recipients = new Vector[3];
    /**
     * The subject of the email
     */
    private String _subject = "";
    /**
     * The content of the email message
     *
     * Each element in the array is treated as a paragraph
     */
    private String[] _content = new String[0];
    /**
     * The content appended to end of the email message
     *
     * Each element in the array is treated as a paragraph
     */
    private String[] _signature = new String[0];
    /**
     * A list of fully qualified file names to be attached to the email when sent
     */
    private Vector _attachments = new Vector(0);
    public Message() {
        for (short i = 0; i < _recipients.length; i++) {
            _recipients[i] = new Vector(1);
        }
    }
    /**
     * Gets the address of the sender of the email
     * @return the sender
     * @see #setFrom
     */
    public String getFrom() {
        return _from;
    }
    /**
     * sets the addess of the sender of the email
     * @param from the from address
     * @see #getFrom
     */
    public void setFrom(String from) {
        _from = from;
    }
    /**
     * Gets the address that any replies should be sent to
     * @return the reply address
     * @see #setReplyTo
     */
    public String getReplyTo() {
        return _replyTo;
    }
    /**
     * sets the addess that any reply emails should be sent to
     * @param the address to send any replies to
     * @see #setReplyTo	
     *
     */
    public void setReplyTo(String replyTo) {
        _replyTo = replyTo;
    }
    /**
     * Adds a message <CODE>recipient</CODE> of the specified <CODE>type</CODE>
     * @param recipient the address of the recipient
     * @param type the type of the recipient
     * @see #TO_RECIPIENT
     * @see #CC_RECIPIENT
     * @see #BCC_RECIPIENT
     * @see #setRecipients
     * @see #getRecipients
     */
    public boolean addRecipient(String recipient, short type) {
        try {
            Vector recipients = _recipients[type];
            recipients.addElement(recipient);
            return true;
        } catch (Exception e) {
            return false;
        }
    }
    /**
     * Sets the recipient list for the specified <CODE>type</CODE>
     * @param recipients the recipient addresses
     * @param type the type of the recipients
     * @see #TO_RECIPIENT
     * @see #CC_RECIPIENT
     * @see #BCC_RECIPIENT
     * @see #addRecipient
     * @see #getRecipients
     */
    public boolean setRecipients(String[] recipients, short type) {
        try {
            Vector v = new Vector();
            for (int i = 0; i < recipients.length; i++) {
                v.addElement(recipients[i]);
            }
            _recipients[type] = v;
            return true;
        } catch (Exception e) {
            return false;
        }
    }
    /**
     * Gets the recipient list for the specified <CODE>type</CODE>
     * @param type the type of the recipients
     * @return An array of the recipient addresses
     * @see #TO_RECIPIENT
     * @see #CC_RECIPIENT
     * @see #BCC_RECIPIENT
     * @see #addRecipient
     * @see #getRecipients
     */
    public String[] getRecipients(short type) {
        try {
            Vector recipients = _recipients[type];
            String[] result = new String[recipients.size()];
            recipients.copyInto(result);
            return result;
        } catch (Exception e) {
            return new String[0];
        }
    }
    /**
     * Sets the content of the email message
     * @param an array of paragraphs
     * @see #getContent
     * @see #addParagraph
     */
    public void setContent(String[] content) {
        _content = content;
    }
    /**
     * Adds a paragraph to the content of the email
     * @param a string containing the paragraph
     * @see #setContent
     * @see #getContent
     */
    public void addParagraph(String paragraph) {
        String[] newContent = new String[_content.length + 1];
        System.arraycopy(_content, 0, newContent, 0, _content.length);
        newContent[_content.length] = paragraph;
        _content = newContent;
    }
    /**
     * Returns the content of the email message
     * @return an array of paragraphs
     * @see #setContent
     * @see #addParagraph
     */
    public String[] getContent() {
        return _content;
    }
    /**
     * Sets the content of the email message
     * @param an array of paragraphs
     * @see #getContent
     * @see #addParagraph
     */
    public void setSignature(String[] content) {
        System.out.print("Message.setSignature(String [" + new Integer(content.length).toString() + "]{");
        for (int i = 0; i < content.length; i++)
            System.out.print(content[i] + ",");
        System.out.println("}");
        _signature = content;
    }
    /**
     * Adds a paragraph to the content of the email
     * @param a string containing the paragraph
     * @see #setContent
     * @see #getContent
     */
    public void addSignature(String paragraph) {
        String[] newContent = new String[_signature.length + 1];
        System.arraycopy(_signature, 0, newContent, 0, _signature.length);
        newContent[_signature.length] = paragraph;
        _signature = newContent;
    }
    /**
     * Returns the content of the email message
     * @return an array of paragraphs
     * @see #setContent
     * @see #addParagraph
     */
    public String[] getSignature() {
        System.out.print("Message.getSignature(String [" + new Integer(_signature.length).toString() + "]{");
        for (int i = 0; i < _signature.length; i++)
            System.out.print(_signature[i] + ",");
        System.out.println("}");
        return _signature;
    }
    /**
     * Sets the Subject of the email
     * @param subject the subject
     */
    public void setSubject(String subject) {
        _subject = subject;
    }
    /**
     * Gets the subject of the email
     * @return the subject
     */
    public String getSubject() {
        return _subject;
    }
    /**
     * Validates the email
     * @return <CODE>true</CODE> if the email is in a valid state to be sent
     */
    public boolean isValid() {
        System.out.println("Message: message.from = " + _from);
        for (short j = 0; j < 3; j++) {
            for (int i = 0; i < _recipients.length; i++)
                System.out.println("Message: message.recipients[" + new Integer(i).toString() + "] = " + _recipients[i]);
        }
        return _from != null && _recipients[TO_RECIPIENT].size() > 0;
    }
    /**
     * Sets the list of attachment filenames to those supplied
     * @param attachments the list of Files to attach
     * @see #addAttachment
     * @see #getAttachments
     */
    public void setAttachments(File[] attachments) {
        _attachments = new Vector(attachments.length);
        for (int i = 0; i < attachments.length; i++) {
            _attachments.addElement(attachments[i]);
        }
    }
    /**
     * Gets the list of attachments to be sent with the email
     * @return an array of <CODE>File</CODE> attachments
     * @see #setAttachments
     * @see #addAttachment
     */
    public File[] getAttachments() {
        File[] files = new File[_attachments.size()];
        _attachments.copyInto(files);
        return files;
    }
    /**
     * Adds the given filename to the list of attachments
     * @param attachment The <CODE>File</CODE> to attach
     * @see #setAttachments
     * @see #getAttachments
     */
    public void addAttachment(File attachment) {
        _attachments.addElement(attachment);
    }
    public void configure(Config config) throws ConfigException {
        Config[] children = config.getChildren();
        if (children == null)
            throw new ConfigException("Bad message configuration: " + config.getName() + " is Empty!");
        Vector attachments = new Vector();
        String value;
        for (int i = 0; i < children.length; i++) {
            value = children[i].getValue();
            if (value == null)
                continue;
            if (children[i].getName().equals(CFG_FROM)) {
                setFrom(value);
            } else if (children[i].getName().equals(CFG_REPLYTO)) {
                setReplyTo(value);
            } else if (children[i].getName().equals(CFG_TO)) {
                addRecipient(value, TO_RECIPIENT);
            } else if (children[i].getName().equals(CFG_CC)) {
                addRecipient(value, CC_RECIPIENT);
            } else if (children[i].getName().equals(CFG_BCC)) {
                addRecipient(value, BCC_RECIPIENT);
            } else if (children[i].getName().equals(CFG_SUBJECT)) {
                setSubject(value);
            } else if (children[i].getName().equals(CFG_CONTENT)) {
                setContent(new String[] { value });
            } else if (children[i].getName().equals(CFG_SIGNATURE)) {
                setSignature(new String[] { value });
            } else if (children[i].getName().equals(CFG_ATTACHMENT)) {
                attachments.addElement(value);
            } else if (children[i].getName().equals(CFG_RECIPIENT)) {
                configure(children[i]);
            }
        }
        String filename;
        File file;
        for (Enumeration enum = attachments.elements(); enum.hasMoreElements();) {
            filename = (String) enum.nextElement();
            file = new File(filename);
            if (file.exists()) {
                addAttachment(file);
            } else {
                throw new ConfigException("File does not exist");
            }
        }
    }
    public Object clone() throws CloneNotSupportedException {
        return super.clone();
    }
}