/* 
 * SmtpMailer.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.mail;

import java.io.File;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import javax.mail.Address;
import javax.mail.Multipart;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.util.Worker;

/** 
 * Implements the Mailer interface to send email by SMTP using the J2EE JavaMail
 * API
 * A sample confiuration file for SmtpMailer would be as follows:
 * <P>
 * <CODE>
 * &lt;Config name="Mailer"&gt;<BR>
 * &nbsp;&nbsp;&lt;Type&gt;com.eXtropia.mail.SmtpMailer&lt;/Type&gt;<BR>
 * &nbsp;&nbsp;&lt;Config name="SmtpHost"&gt;my.smtp.server.com&lt;/Config&gt;<BR>
 * &lt;/Config&gt;<BR>
 * </CODE>
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 0.1
 */
public class SmtpMailer extends Mailer {
    /**
     * Required Parameter in the {@link Config}
     */
    public static final String SMTP_HOST = "SmtpHost";
    /**
     * If set mails are sent on a separate thread
     */
    public static final String CFG_SEND_THREAD = "SendThread";
    /**
     * The filters used when logging to the {@link Diagnostics} API
     */
    public static final String[] DIAG_FILTERS = new String[] { "Mailer", "SmtpMailer" };
    /**
     * The doctor used to perform {@link Diagnostics}
     */
    protected Doctor _doctor;
    /**
     * The worker used for sending messages
     */
    protected Worker _worker;
    /**
     * The default message values used for sending messages
     */
    protected Message _messageDefaults;
    /**
     * The session used to send messages
     */
    protected javax.mail.Session session;
    /** Creates new SmtpMailer */
    public SmtpMailer() {
        _doctor = Diagnostics.getDoctor(DIAG_FILTERS);
        _messageDefaults = new Message();
    }
    /**
     * Configures the the SmtpMailer
     * <P>
     * Requires one child <CODE>Config</CODE> containing the value of the smtp host
     * @see #SMTP_HOST
     */
    public void configure(Config config) throws ConfigException {
        if (session != null) //we are already configured
            return;
        //    _doctor.log(new LogEvent(this, LogEvent.INFO, "SmtpMailer: configuring mailer."));
        Properties properties = new Properties();
        // need the smtp host
        Config host = config.getChild(SMTP_HOST);
        if (host != null) {
            //      _doctor.log(new LogEvent(this, LogEvent.INFO, "SmtpMailer: configuring mailer host."));
            properties.put("mail.smtp.host", host.getValue());
        }
        // get message defaults
        if (config.hasChild(Message.CFG_BASE)) {
            //      _doctor.log(new LogEvent(this, LogEvent.INFO, "SmtpMailer: configuring mailer message defaults."));
            Config child;
            if ((child = config.getChild(Message.CFG_BASE)) != null) {
                _messageDefaults = new Message();
                _messageDefaults.configure(child);
                /*
                String content [];
                if ( (content= _messageDefaults.getSignature()) != null )
                {
                	System.out.println("SmtpMailer.configure: _messageDefaults.getSignature()[" + content.length +"]={");
                	for(int i=0;i<content.length;i++)
                 		System.out.println( content[i]+",");
                	System.out.println("}");
                }
                */
            }
        }
        // get thread option
        if (config.hasChild(CFG_SEND_THREAD)) {
            //      _doctor.log(new LogEvent(this, LogEvent.INFO, "SmtpMailer: configuring mailer thread option."));
            boolean sendThread = config.getChild(CFG_SEND_THREAD).getBooleanValue();
            if (sendThread) {
                _worker = new com.eXtropia.util.Worker(Thread.NORM_PRIORITY - 1, false);
            }
        }
        //    _doctor.log(new LogEvent(this, LogEvent.INFO, "SmtpMailer: testing Session.getDefaultInstance."));
        session = Session.getDefaultInstance(properties, null);
        //    _doctor.log(new LogEvent(this, LogEvent.INFO, "SmtpMailer: configuration completed."));
    }
    /**
     * Defines {@link Mailer}
     */
    public boolean handleSend(Message message) {
        if (_worker != null) {
            _worker.addWork(new MailSender(message));
            return true;
        } else
            return realHandleSend(message);
    }
    public boolean realHandleSend(Message message, Message defaults) {
        _messageDefaults = defaults;
        return realHandleSend(message);
    }
    public boolean realHandleSend(Message message) {
        if (message.isValid()) {
            try {
                javax.mail.Message outMessage = new MimeMessage(session);
                setAddresses(message, outMessage);
                outMessage.setSubject(message.getSubject());
                setContent(message, outMessage);
                if (message.getFrom() != null)
                    outMessage.setFrom(new InternetAddress(message.getFrom()));
                if (message.getReplyTo() != null)
                    outMessage.setReplyTo(new Address[] { new InternetAddress(message.getReplyTo())});
                Transport.send(outMessage);
            } catch (Exception e) {
                if (_doctor.doTrace(TraceEvent.INFO)) {
                    _doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Caught exception sending mail", e));
                }
                return false;
            }
            if (_doctor.doTrace(TraceEvent.INFO)) {
                _doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Sent mail: \"" + message.getSubject() + "\" to \"" + message.getRecipients(Message.TO_RECIPIENT)[0] + "\" et al"));
            }
            return true;
        } else {
            if (_doctor.doTrace(TraceEvent.VERBOSE)) {
                _doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Can't send invalid com.eXtropia.mail.Message"));
            }
            return false;
        }
    }
    /**
     * Sets the addresses in the J2EE mail from the input mail
     */
    private void setAddresses(Message inMessage, javax.mail.Message outMessage) throws SmtpMailerException {
        try {
            outMessage.setRecipients(javax.mail.Message.RecipientType.TO, convertAddresses(inMessage, Message.TO_RECIPIENT));
            outMessage.setRecipients(javax.mail.Message.RecipientType.CC, convertAddresses(inMessage, Message.CC_RECIPIENT));
            outMessage.setRecipients(javax.mail.Message.RecipientType.BCC, convertAddresses(inMessage, Message.BCC_RECIPIENT));
        } catch (Exception e) {
            throw new SmtpMailerException();
        }
    }
    /**
     * Converts a String array into a javax.mail.Address array
     */
    private Address[] convertAddresses(Message inMessage, short recipientType) throws SmtpMailerException {
        String[] sa;
        Set accum = new HashSet();
        try {
            if ((sa = inMessage.getRecipients(recipientType)) != null)
                for (int i = 0; i < sa.length; i++)
                    accum.add(new InternetAddress(sa[i]));
            if ((sa = _messageDefaults.getRecipients(recipientType)) != null)
                for (int i = 0; i < sa.length; i++)
                    accum.add(new InternetAddress(sa[i]));
            Address[] addresses = (Address[]) accum.toArray(new Address[accum.size()]);
            return addresses;
        } catch (Exception e) {
            return new Address[0];
        }
    }
    /**
     * Generates the content of the mail
     */
    private String createContent(Message inMessage) {
        String[] content;
        StringBuffer entireContent = new StringBuffer();
        System.out.println("SmtpMailer.createContent(inMessage): _messageDefaults.getContent()...");
        if ((content = _messageDefaults.getContent()) != null)
            for (int i = 0; i < content.length; i++)
                entireContent.append(content[i] + "\n\n");
        System.out.println("SmtpMailer.createContent(inMessage): inMessage.getContent()...");
        if ((content = inMessage.getContent()) != null)
            for (int i = 0; i < content.length; i++)
                entireContent.append(content[i] + "\n\n");
        System.out.println("SmtpMailer.createContent(inMessage): inMessage.getSignature()...");
        if ((content = inMessage.getSignature()) != null)
            for (int i = 0; i < content.length; i++)
                entireContent.append(content[i] + "\n\n");
        System.out.println("SmtpMailer.createContent(inMessage): _messageDefaults.getSignature()...");
        if ((content = _messageDefaults.getSignature()) != null)
            for (int i = 0; i < content.length; i++)
                entireContent.append(content[i] + "\n\n");
        return entireContent.toString();
    }
    class SmtpMailerException extends Exception {
        public SmtpMailerException() {
        }
    }
    /**
     * Sets the content of the email to be sent
     */
    private void setContent(Message message, javax.mail.Message outMessage) throws SmtpMailerException {
        try {
            File[] attachments = message.getAttachments();
            if (attachments.length > 0) {
                MimeBodyPart part;
                Multipart multiPart = new MimeMultipart();
                part = new MimeBodyPart();
                part.setContent(createContent(message), "text/plain");
                multiPart.addBodyPart(part);
                FileDataSource fds;
                for (int i = 0; i < attachments.length; i++) {
                    part = new MimeBodyPart();
                    fds = new FileDataSource(attachments[i]);
                    part.setDataHandler(new DataHandler(fds));
                    part.setFileName(fds.getName());
                    multiPart.addBodyPart(part);
                }
                outMessage.setContent(multiPart);
            } else {
                outMessage.setContent(createContent(message), "text/plain");
            }
        } catch (Exception e) {
            e.printStackTrace();
            throw new SmtpMailerException();
        }
    }
    /**
     * Get message
     */
    public Message getMessageDefaults() {
        return _messageDefaults;
    }
    /**
     * set message
     */
    public void setMessageDefaults(Message m) {
        _messageDefaults = m;
    }
    private class MailSender implements Runnable {
        private Message _message;
        MailSender(Message message) {
            _message = message;
        }
        public void run() {
            realHandleSend(_message);
        }
    }
}