/*
 * CmdDispatcher.java
 * Copyright (C) 2000  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.sms.dispatcher;

import java.io.IOException;
import java.text.MessageFormat;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.sms.Sms;

/**
 * This class is a command line executor.
 *
 * @author <a href="mailto:carlo@eXtropia.com">Carlo Ravagli</a>
 * @author <a href="mailto:st_lim@eXtropia.com">Lim Swee Tat</a>
 * @version $Id: CmdDispatcher.java,v 1.10 2002/11/16 08:45:13 rayfd Exp $
 */
public class CmdDispatcher implements SmsDispatcher {

    public static final String CFG_CMD = "Command";
    public static final String CFG_ARG = "Argument";

    protected static final int TO_PARAM_INDEX = 0;
    protected static final int FROM_PARAM_INDEX = 1;
    protected static final int MSG_PARAM_INDEX = 2;
    protected static final int TYPE_PARAM_INDEX = 3;
    protected static final int MAX_PARAM_INDEX = TYPE_PARAM_INDEX;

    private MessageFormat[] _cmdFormat;

    /**
     * Creates a new <code>CmdDispatcher</code> instance.
     *
     */
    public CmdDispatcher() {
    }

    /**
     * <code>configure</code> the Dispatcher.
     *
     * @param config a <code>Config</code> value
     * @exception ConfigException if an error occurs
     */
    public void configure(Config config) throws ConfigException {
        Config cmdConfig = config.getChild(CFG_CMD);
        if (cmdConfig == null || cmdConfig.getValue() == null)
            throw new ConfigException("Failed to configure cmd displatcher: No command config found");

        Config[] argsConfig = cmdConfig.getChildren();
        _cmdFormat = new MessageFormat[argsConfig.length + 1];
        _cmdFormat[0] = new MessageFormat(reformatPattern(cmdConfig.getValue()));

        for (int i = 0; argsConfig != null && i < argsConfig.length; i++) {
            if (argsConfig[i].getName().equals(CFG_ARG) && argsConfig[i].getValue() != null)
                _cmdFormat[i + 1] = new MessageFormat(reformatPattern(argsConfig[i].getValue()));
        }
    }

    /**
     * Describe <code>send</code> method here.
     *
     * @param sms a <code>Sms</code> value
     * @exception SmsDispatchException if an error occurs
     */
    public void send(Sms sms) throws SmsDispatchException {
        Object[] cmdParams = getCmdParams(sms);
        String[] cmd = new String[_cmdFormat.length];

        for (int i = 0; i < cmd.length; i++)
            cmd[i] = _cmdFormat[i].format(cmdParams);

        try {
            Doctor doctor = Diagnostics.getDoctor(new String[] { CFG_CMD, "SMS CmdDispatcher" });
            String cmdLine = null;
            if (cmd[0] != null)
                cmdLine = cmd[0];
            if (cmd.length > 1) {
                for (int i = 1; i < cmd.length; i++)
                    cmdLine += " " + cmd[i];
            }
            doctor.trace(new TraceEvent(this, TraceEvent.DEVELOPER, "Command-Line Generated: " + cmdLine));
            Process p = Runtime.getRuntime().exec(cmd);
            if (p.waitFor() != 0)
                throw new SmsDispatchException("Failed to send SMS message: command failed");
        } catch (IOException e) {
            throw new SmsDispatchException("Failed to send SMS message: io error - " + e.getMessage());
        } catch (InterruptedException e) {
            throw new SmsDispatchException("Failed to send SMS message: process interrupted - " + e.getMessage());
        }
    }

    /**
     * Describe <code>getCmdParams</code> method here.
     *
     * @param sms a <code>Sms</code> value
     * @return an <code>Object[]</code> value
     */
    protected Object[] getCmdParams(Sms sms) {
        Object[] cmdParams = new Object[MAX_PARAM_INDEX + 1];
        cmdParams[TO_PARAM_INDEX] = sms.getTo();
        cmdParams[FROM_PARAM_INDEX] = sms.getFrom();
        cmdParams[MSG_PARAM_INDEX] = sms.getText();
        cmdParams[TYPE_PARAM_INDEX] = new Integer(sms.getType());
        return cmdParams;
    }

    /**
     * Describe <code>reformatPattern</code> method here.
     *
     * @param pattern a <code>String</code> value
     * @return a <code>String</code> value
     */
    protected String reformatPattern(String pattern) {
        pattern = replaceAll(pattern, "{to", "{" + TO_PARAM_INDEX);
        pattern = replaceAll(pattern, "{from", "{" + FROM_PARAM_INDEX);
        pattern = replaceAll(pattern, "{msg", "{" + MSG_PARAM_INDEX);
        pattern = replaceAll(pattern, "{type", "{" + TYPE_PARAM_INDEX);
        return pattern;
    }

    /**
     * Describe <code>replaceAll</code> method here.
     *
     * @param pattern a <code>String</code> value
     * @param what a <code>String</code> value
     * @param with a <code>String</code> value
     * @return a <code>String</code> value
     */
    protected String replaceAll(String pattern, String what, String with) {
        for (int i = pattern.indexOf(what); i >= 0; i = pattern.indexOf(what)) {
            StringBuffer newPattern = new StringBuffer(pattern);
            pattern = newPattern.replace(i, i + what.length(), with).toString();
        }
        return pattern;
    }
} // CmdDispatcher.java
