/*
 * ScheduledDispatcher.java
 * Copyright (C) 2000  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.sms.dispatcher;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.sms.ScheduledSms;
import com.eXtropia.sms.Sms;

/**
 * <p><code>ScheduledDispatcher</code> will read two dispatchers. One
 * performs for scheduled Sms, and the other sends immediately.
 * <p><code>ScheduledDispatcher</code> only deals with ScheduledSms.
 * Hence all Sms to be sent must have Scheduled information.
 *
 * @author <a href="mailto:carlo@eXtropia.com">Carlo Ravagli</a>
 * @author <a href="mailto:st_lim@eXtropia.com">Lim Swee Tat</a>
 * @version $Id: ScheduledDispatcher.java,v 1.5 2002/11/16 08:45:13 rayfd Exp $
 */
public class ScheduledDispatcher implements SmsDispatcher {

    public static final String CFG_SCHEDULER = "Scheduler";
    public static final String CFG_IMMEDIATE = "Immediate";

    protected SmsDispatcher _schedulerDispatcher;
    protected SmsDispatcher _immediateDispatcher;

    /**
     * Creates a new <code>ScheduledDispatcher</code> instance.
     *
     */
    public ScheduledDispatcher() {
    }

    /**
     * Describe <code>configure</code> method here.
     *
     * @param config a <code>Config</code> value
     * @exception ConfigException if an error occurs
     */
    public void configure(Config config) throws ConfigException {

        // Configures the Scheduler
        Config schedulerConfig = config.getChild(CFG_SCHEDULER);
        if (schedulerConfig == null || schedulerConfig.getValue() == null)
            throw new ConfigException("Failed to configure scheduled displatcher: No scheduler config found");
        try {
            _schedulerDispatcher = (SmsDispatcher) Class.forName(schedulerConfig.getType()).newInstance();
            _schedulerDispatcher.configure(schedulerConfig);
        } catch (ClassCastException cce) {
            throw new ConfigException("Unable to create SmsDispatcher. Invalid type: " + schedulerConfig.getType());
        } catch (InstantiationException ie) {
            throw new ConfigException("Failed to instantiate the specified SmsDispatcher: " + schedulerConfig.getType());
        } catch (IllegalAccessException iae) {
            throw new ConfigException("Illegal access to specified SmsDispatcher class: " + schedulerConfig.getType());
        } catch (ClassNotFoundException cnfe) {
            throw new ConfigException("Unable to find specified SmsDispatcher class: " + schedulerConfig.getType());
        }

        // Configures the Immediate Dispatcher
        Config immediateConfig = config.getChild(CFG_IMMEDIATE);
        if (immediateConfig == null || immediateConfig.getValue() == null)
            throw new ConfigException("Failed to configure immediate displatcher: No immediate config found");
        try {
            _immediateDispatcher = (SmsDispatcher) Class.forName(immediateConfig.getType()).newInstance();
            _immediateDispatcher.configure(immediateConfig);
        } catch (ClassCastException cce) {
            throw new ConfigException("Unable to create SmsDispatcher. Invalid type: " + immediateConfig.getType());
        } catch (InstantiationException ie) {
            throw new ConfigException("Failed to instantiate the specified SmsDispatcher: " + immediateConfig.getType());
        } catch (IllegalAccessException iae) {
            throw new ConfigException("Illegal access to specified SmsDispatcher class: " + immediateConfig.getType());
        } catch (ClassNotFoundException cnfe) {
            throw new ConfigException("Unable to find specified SmsDispatcher class: " + immediateConfig.getType());
        }
    }

    /**
     * Describe <code>send</code> method here.
     *
     * @param sms a <code>Sms</code> value
     * @exception SmsDispatchException if an error occurs
     */
    public void send(Sms sms) throws SmsDispatchException {
        if (!(sms instanceof ScheduledSms)) {
            throw new SmsDispatchException("Failed to send SMS message: wrong sms type used.");
        }

        ScheduledSms sSms = (ScheduledSms) sms;
        //if (sSms.isScheduled()) {
        //    _schedulerDispatcher.send(sms);
        //} else {
        _immediateDispatcher.send(sms);
        //}
    }

} // ScheduledDispatcher.java
