/*
 * SmsDispatcherManager.java
 * Copyright (C) 2000  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.sms.dispatcher;

import java.util.HashMap;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.sms.Sms;

/**
 * Describe class <code>SmsDispatcherManager</code> here.
 *
 * @author <a href="mailto:carlo@eXtropia.com">Carlo Ravagli</a>
 * @author <a href="mailto:st_lim@eXtropia.com">Lim Swee Tat</a>
 * @version $Id: SmsDispatcherManager.java,v 1.6 2002/11/16 08:45:13 rayfd Exp $
 */
public abstract class SmsDispatcherManager implements Configurable {

    public static final String CFG_BASE = "SmsDispatcherManager";
    public static final String CFG_DISPATCHERS = "Dispatchers";
    public static final String CFG_DEFAULT_DISPATCHER = "Default";

    protected HashMap _dispatchers = new HashMap();
    protected SmsDispatcher _default = null;

    /**
     * Creates a new <code>SmsDispatcherManager</code> instance.
     *
     */
    protected SmsDispatcherManager() {
    }

    /**
     * Describe <code>configure</code> method here.
     *
     * @param config a <code>Config</code> value
     * @exception ConfigException if an error occurs
     */
    public void configure(Config config) throws ConfigException {
        _dispatchers.clear();

        Config dispatcherConfig = config.getChild(CFG_DISPATCHERS);
        if (dispatcherConfig != null) {
            Config[] dispatchersConfig = dispatcherConfig.getChildren();
            for (int i = 0; dispatchersConfig != null && i < dispatchersConfig.length; i++) {
                try {
                    SmsDispatcher d = (SmsDispatcher) Class.forName(dispatchersConfig[i].getType()).newInstance();
                    d.configure(dispatchersConfig[i]);
                    if (_dispatchers.put(dispatchersConfig[i].getName(), d) != null)
                        throw new ConfigException("All SMS dispatchers must have a unique name");
                } catch (ClassCastException cce) {
                    throw new ConfigException("Unable to create SmsDispatcher. Invalid type: " + dispatchersConfig[i].getType());
                } catch (InstantiationException ie) {
                    throw new ConfigException("Failed to instantiate the specified SmsDispatcher: " + dispatchersConfig[i].getType());
                } catch (IllegalAccessException iae) {
                    throw new ConfigException("Illegal access to specified SmsDispatcher class: " + dispatchersConfig[i].getType());
                } catch (ClassNotFoundException cnfe) {
                    throw new ConfigException("Unable to find specified SmsDispatcher class: " + dispatchersConfig[i].getType());
                }
            }
        }

        // Putting it here means that the default is just outside the normal set of dispatcher array.
        Config defaultConfig = config.getChild(CFG_DEFAULT_DISPATCHER);
        if (defaultConfig == null) {
            _default = null;
        } else {
            try {
                _default = (SmsDispatcher) Class.forName(defaultConfig.getType()).newInstance();
                _default.configure(defaultConfig);
            } catch (ClassCastException cce) {
                throw new ConfigException("Unable to create SmsDispatcher. Invalid type: " + defaultConfig.getType());
            } catch (InstantiationException ie) {
                throw new ConfigException("Failed to instantiate the specified SmsDispatcher: " + defaultConfig.getType());
            } catch (IllegalAccessException iae) {
                throw new ConfigException("Illegal access to specified SmsDispatcher class: " + defaultConfig.getType());
            } catch (ClassNotFoundException cnfe) {
                throw new ConfigException("Unable to find specified SmsDispatcher class: " + defaultConfig.getType());
            }
        }
    }

    /**
     * Describe <code>send</code> method here.
     *
     * @param sms a <code>Sms</code> value
     * @exception SmsDispatchException if an error occurs
     */
    public void send(Sms sms) throws SmsDispatchException {
        SmsDispatcher dispatcher = getDispatcher(sms.getTo());
        if (dispatcher == null)
            throw new SmsDispatchException("Failed to send msg: could not get dispatcher for phone number - " + sms.getTo());
        dispatcher.send(sms);
    }

    /**
     * The child classes _must_ extend the
     * <code>getDispatcher</code> method.
     *
     * @param to an <code>String</code> value
     * @return a <code>SmsDispatcher</code> value
     */
    protected abstract SmsDispatcher getDispatcher(String to);

    /**
     * Describe <code>create</code> method here.
     *
     * @param config a <code>Config</code> value
     * @return a <code>SmsDispatcherManager</code> value
     * @exception ConfigException if an error occurs
     */
    public static final SmsDispatcherManager create(Config config) throws ConfigException {
        if (!config.getName().equals(CFG_BASE))
            throw new ConfigException("Cannot create a SmsDispatcherManager with configuration: " + config.getName());

        try {
            SmsDispatcherManager dm = (SmsDispatcherManager) Class.forName(config.getType()).newInstance();
            dm.configure(config);
            return dm;
        } catch (ClassCastException cce) {
            throw new ConfigException("Unable to create SmsDispatcherManager. Invalid type: " + config.getType());
        } catch (InstantiationException ie) {
            throw new ConfigException("Failed to instantiate the specified SmsDispatcherManager: " + config.getType());
        } catch (IllegalAccessException iae) {
            throw new ConfigException("Illegal access to specified SmsDispatcherManager class: " + config.getType());
        } catch (ClassNotFoundException cnfe) {
            throw new ConfigException("Unable to find specified SmsDispatcherManager class: " + config.getType());
        }
    }
} // SmsDispatcherManager.java
