/*
 * ConditionDeserializer.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.soap.encoding.datasource;

import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;

import org.apache.soap.Constants;
import org.apache.soap.encoding.soapenc.SoapEncUtils;
import org.apache.soap.rpc.SOAPContext;
import org.apache.soap.util.Bean;
import org.apache.soap.util.xml.DOMUtils;
import org.apache.soap.util.xml.Deserializer;
import org.apache.soap.util.xml.NSStack;
import org.apache.soap.util.xml.QName;
import org.apache.soap.util.xml.Serializer;
import org.apache.soap.util.xml.XMLJavaMappingRegistry;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.eXtropia.datasource.Condition;

/**
 *
 * @author  Carlo Ravagli (carlo.ravagli@extropia.com)
 * @version 1.0
 */
public class ConditionDeserializer implements Serializer, Deserializer {
    public static final String CONDITION_ELEMENT_NAME = "Condition";
    public static final String OPERATOR_ELEMENT_NAME = "Operand";
    public static final String FIELD_ELEMENT_NAME = "Field";
    public static final String VALUE_ELEMENT_NAME = "Value";

    public ConditionDeserializer() {
    }

    public void marshall(String inScopeEncStyle, Class javaType, Object src, Object context, Writer sink, NSStack nsStack, XMLJavaMappingRegistry xjmr, SOAPContext ctx) throws IOException {
        Condition cond = (Condition) src;
        nsStack.pushScope();

        if (cond == null) {
            SoapEncUtils.generateNullStructure(inScopeEncStyle, javaType, context, sink, nsStack, xjmr);
        } else {
            SoapEncUtils.generateStructureHeader(inScopeEncStyle, javaType, context, sink, nsStack, xjmr);

            sink.write('<' + FIELD_ELEMENT_NAME + '>');
            sink.write(cond.getField());
            sink.write("</" + FIELD_ELEMENT_NAME + '>');
            sink.write('<' + OPERATOR_ELEMENT_NAME + '>');

            switch (cond.getOperator()) {
                case Condition.NOT_EQUAL :
                    sink.write("NOT_EQUAL");
                    break;
                case Condition.GREATER_OR_EQUAL :
                    sink.write("GREATER_OR_EQUAL");
                    break;
                case Condition.GREATER_THAN :
                    sink.write("GREATER_THAN");
                    break;
                case Condition.LESS_OR_EQUAL :
                    sink.write("LESS_OR_EQUAL");
                    break;
                case Condition.LESS_THAN :
                    sink.write("LESS_THAN");
                    break;
                default :
                    sink.write("EQUAL");
            }

            sink.write("</" + OPERATOR_ELEMENT_NAME + '>');

            if (cond.getValue() != null) {
                xjmr.marshall(inScopeEncStyle, cond.getValue().getClass(), cond.getValue(), VALUE_ELEMENT_NAME, sink, nsStack, ctx);
            } else {
                SoapEncUtils.generateNullStructure(inScopeEncStyle, Object.class, VALUE_ELEMENT_NAME, sink, nsStack, xjmr);
            }
            sink.write("</" + context + '>');
        }

        nsStack.popScope();
    }

    public Bean unmarshall(String inScopeEncStyle, final QName elementType, final Node src, final XMLJavaMappingRegistry xjmr, final SOAPContext ctx) {
        if (SoapEncUtils.isNull((Element) src))
            return new Bean(Condition.class, null);

        Condition condition = new Condition();
        boolean specified[] = new boolean[3];
        Arrays.fill(specified, false);

        for (Node childNode = src.getFirstChild(); childNode != null; childNode = childNode.getNextSibling()) {
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                String name = childNode.getNodeName();
                if (name.equals(FIELD_ELEMENT_NAME)) {
                    if (specified[0])
                        throw new IllegalArgumentException("Failed to unmarshall Condition: <" + name + "> is specifed twice");

                    String field = DOMUtils.getChildCharacterData((Element) childNode);

                    condition.setField(field);
                    specified[0] = true;
                } else if (name.equals(OPERATOR_ELEMENT_NAME)) {
                    if (specified[1])
                        throw new IllegalArgumentException("Failed to unmarshall Condition: <" + name + "> is specifed twice");

                    String operator = DOMUtils.getChildCharacterData((Element) childNode);

                    try {
                        condition.setOperator(Condition.class.getDeclaredField(operator).getInt(null));
                    } catch (Exception e) {
                        throw new IllegalArgumentException(e.getMessage());
                    }
                    specified[1] = true;
                } else if (name.equals(VALUE_ELEMENT_NAME)) {
                    if (specified[2])
                        throw new IllegalArgumentException("Failed to unmarshall Condition: <" + name + "> is specifed twice");

                    Element valueElement = (Element) childNode;

                    if (!SoapEncUtils.isNull(valueElement)) {

                        QName soapType = SoapEncUtils.getAttributeValue(valueElement, Constants.NS_URI_1999_SCHEMA_XSI, Constants.ATTR_TYPE, VALUE_ELEMENT_NAME, true);

                        condition.setValue(xjmr.unmarshall(inScopeEncStyle, soapType, childNode, ctx).value);
                    }
                    specified[2] = true;
                } else
                    throw new IllegalArgumentException("Failed to unmarshall Condition: <" + name + "> is not a valid sub element");
            }
        }

        return new Bean(Condition.class, condition);
    }
}