/* 
 * QueryDeserializer.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.soap.encoding.datasource;

import java.io.IOException;
import java.io.Writer;

import org.apache.soap.encoding.soapenc.SoapEncUtils;
import org.apache.soap.rpc.SOAPContext;
import org.apache.soap.util.Bean;
import org.apache.soap.util.xml.DOMUtils;
import org.apache.soap.util.xml.Deserializer;
import org.apache.soap.util.xml.NSStack;
import org.apache.soap.util.xml.QName;
import org.apache.soap.util.xml.Serializer;
import org.apache.soap.util.xml.XMLJavaMappingRegistry;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.eXtropia.datasource.Criterion;
import com.eXtropia.datasource.FieldOrder;
import com.eXtropia.datasource.Query;

/** 
 *
 * @author  Carlo Ravagli (carlo.ravagli@extropia.com)
 * @version 1.0
 */
public class QueryDeserializer implements Serializer, Deserializer {
    public static final String QUERY_ELEMENT_NAME = "Query";
    public static final String OFFSET_ELEMENT_NAME = "Offset";
    public static final String LIMIT_ELEMENT_NAME = "Limit";
    public static final String ORDER_ELEMENT_NAME = "Order";
    public static final String SORTBY_ELEMENT_NAME = "SortBy";
    public static final String FIELD_ELEMENT_NAME = "Field";
    public static final String DIRECTION_ELEMENT_NAME = "Direction";

    public QueryDeserializer() {
    }

    public void marshall(String inScopeEncStyle, Class javaType, Object src, Object context, Writer sink, NSStack nsStack, XMLJavaMappingRegistry xjmr, SOAPContext ctx) throws IOException {
        Query query = (Query) src;
        nsStack.pushScope();

        if (query == null) {
            SoapEncUtils.generateNullStructure(inScopeEncStyle, javaType, context, sink, nsStack, xjmr);
        } else {
            SoapEncUtils.generateStructureHeader(inScopeEncStyle, javaType, context, sink, nsStack, xjmr);

            sink.write('<' + OFFSET_ELEMENT_NAME + '>');
            sink.write(Integer.toString(query.getOffset()));
            sink.write("</" + OFFSET_ELEMENT_NAME + '>');

            sink.write('<' + LIMIT_ELEMENT_NAME + '>');
            sink.write(Integer.toString(query.getLimit()));
            sink.write("</" + LIMIT_ELEMENT_NAME + '>');

            if (query.getCriterion() != null) {
                xjmr.marshall(inScopeEncStyle, Criterion.class, query.getCriterion(), CriterionDeserializer.CRITERION_ELEMENT_NAME, sink, nsStack, ctx);
            }

            FieldOrder[] orderBy = query.getFieldOrder();
            if (orderBy != null) {
                sink.write('<' + ORDER_ELEMENT_NAME + '>');
                for (int i = 0; i < orderBy.length; i++) {
                    sink.write('<' + SORTBY_ELEMENT_NAME + '>');
                    sink.write('<' + FIELD_ELEMENT_NAME + '>');
                    sink.write(orderBy[i].getField());
                    sink.write("</" + FIELD_ELEMENT_NAME + '>');
                    sink.write('<' + DIRECTION_ELEMENT_NAME + '>');
                    sink.write(orderBy[i].getOrder() == FieldOrder.ASCEND ? "ASCEND" : "DESCEND");
                    sink.write("</" + DIRECTION_ELEMENT_NAME + '>');
                    sink.write("</" + SORTBY_ELEMENT_NAME + '>');
                }
                sink.write("</" + ORDER_ELEMENT_NAME + '>');
            }

            sink.write("</" + context + '>');
        }
        nsStack.popScope();
    }

    public Bean unmarshall(String inScopeEncStyle, final QName elementType, final Node src, final XMLJavaMappingRegistry xjmr, final SOAPContext ctx) {
        if (SoapEncUtils.isNull((Element) src))
            return new Bean(Query.class, null);

        Query query = new Query();

        for (Node childNode = src.getFirstChild(); childNode != null; childNode = childNode.getNextSibling()) {
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                String name = childNode.getNodeName();
                if (name.equals(CriterionDeserializer.CRITERION_ELEMENT_NAME)) {
                    if (query.getCriterion() != null)
                        throw new IllegalArgumentException("Invalid Query element: more one criterion specified");

                    Criterion criterion = (Criterion) xjmr.unmarshall(inScopeEncStyle, xjmr.queryElementType(Criterion.class, inScopeEncStyle), childNode, ctx).value;

                    query.setCriterion(criterion);

                } else if (name.equals(ORDER_ELEMENT_NAME)) {
                    if (query.getFieldOrder() != null)
                        throw new IllegalArgumentException("Invalid Query element: more one sort order specified");

                    query.setFieldOrder(getOrder(childNode));

                } else if (name.equals(OFFSET_ELEMENT_NAME)) {
                    query.setOffset(Integer.parseInt(DOMUtils.getChildCharacterData((Element) childNode)));
                } else if (name.equals(LIMIT_ELEMENT_NAME)) {
                    query.setLimit(Integer.parseInt(DOMUtils.getChildCharacterData((Element) childNode)));
                } else
                    throw new IllegalArgumentException("Failed to unmarshall Query: <" + name + "> is not a valid sub element");
            }
        }

        return new Bean(Query.class, query);
    }

    private FieldOrder[] getOrder(Node node) {
        FieldOrder[] order = new FieldOrder[DOMUtils.countKids((Element) node, Node.ELEMENT_NODE)];
        int i = 0;

        for (Node sortByNode = node.getFirstChild(); sortByNode != null; sortByNode = sortByNode.getNextSibling()) {
            if (sortByNode.getNodeType() == Node.ELEMENT_NODE) {
                if (!sortByNode.getNodeName().equals(SORTBY_ELEMENT_NAME))
                    throw new IllegalArgumentException("Failed to unmarshall Order: <" + sortByNode.getNodeName() + "> is not a valid sub element");

                order[i++] = getSortBy(sortByNode);
            }
        }

        return order;
    }

    private FieldOrder getSortBy(Node node) {
        FieldOrder sortBy = new FieldOrder();

        for (Node childNode = node.getFirstChild(); childNode != null; childNode = childNode.getNextSibling()) {
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                String name = childNode.getNodeName();
                if (name.equals(FIELD_ELEMENT_NAME)) {
                    if (sortBy.getField() != null)
                        throw new IllegalArgumentException("Failed to unmarshall " + SORTBY_ELEMENT_NAME + ": <" + name + "> is specifed more than once");

                    sortBy.setField(DOMUtils.getChildCharacterData((Element) childNode));
                } else if (name.equals(DIRECTION_ELEMENT_NAME)) {
                    String direction = DOMUtils.getChildCharacterData((Element) childNode);
                    try {
                        sortBy.setOrder(FieldOrder.class.getDeclaredField(direction).getInt(null));
                    } catch (Exception e) {
                        throw new IllegalArgumentException(e.getMessage());
                    }
                } else
                    throw new IllegalArgumentException("Failed to unmarshall " + SORTBY_ELEMENT_NAME + ": <" + name + "> is not a valid sub element");
            }
        }

        return sortBy;
    }
}