/*
 * RowSetSerializer.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.soap.encoding.datasource;

import java.io.IOException;
import java.io.Writer;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;

import javax.sql.RowSet;

import org.apache.soap.Constants;
import org.apache.soap.encoding.soapenc.SoapEncUtils;
import org.apache.soap.rpc.SOAPContext;
import org.apache.soap.util.Bean;
import org.apache.soap.util.xml.DOMUtils;
import org.apache.soap.util.xml.Deserializer;
import org.apache.soap.util.xml.NSStack;
import org.apache.soap.util.xml.QName;
import org.apache.soap.util.xml.Serializer;
import org.apache.soap.util.xml.XMLJavaMappingRegistry;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import sun.jdbc.rowset.CachedRowSet;
import sun.jdbc.rowset.RowSetMetaDataImpl;

/**
 * A <code>RowSetSerializer</code> is used to serialize
 * ResultSets.
 *
 * @author Carlo Ravagli (carlo.ravagli@extropia.com)
 */
public class RowSetSerializer implements Serializer, Deserializer {
    public static final String ELEMENT_COLS = "Columns";
    public static final String ELEMENT_COL = "Column";

    public static final String ELEMENT_ROWS = "Rows";
    public static final String ELEMENT_ROW = "Row";

    public void marshall(String inScopeEncStyle, Class javaType, Object src, Object context, Writer sink, NSStack nsStack, XMLJavaMappingRegistry xjmr, SOAPContext ctx) throws IOException {
        RowSet rs = (RowSet) src;
        nsStack.pushScope();

        if (rs == null) {
            SoapEncUtils.generateNullStructure(inScopeEncStyle, javaType, context, sink, nsStack, xjmr);
        } else {
            try {
                ResultSetMetaData rsmd = rs.getMetaData();
                int numberOfColumns = rsmd != null ? rsmd.getColumnCount() : 0;

                SoapEncUtils.generateStructureHeader(inScopeEncStyle, javaType, context, sink, nsStack, xjmr);

                sink.write('<' + ELEMENT_COLS + '>');
                for (int i = 1; i <= numberOfColumns; i++) {
                    sink.write('<' + ELEMENT_COL + '>');
                    sink.write(rsmd.getColumnName(i));
                    sink.write("</" + ELEMENT_COL + '>');
                }
                sink.write("</" + ELEMENT_COLS + '>');

                sink.write('<' + ELEMENT_ROWS + '>');
                while (rs.next()) {
                    sink.write('<' + ELEMENT_ROW + '>');
                    for (int i = 1; i <= numberOfColumns; i++) {
                        Object value = rs.getObject(i);

                        if (rs.wasNull()) {
                            SoapEncUtils.generateNullStructure(inScopeEncStyle, Object.class, rsmd.getColumnName(i), sink, nsStack, xjmr);
                        } else {
                            xjmr.marshall(inScopeEncStyle, value.getClass(), value, rsmd.getColumnName(i), sink, nsStack, ctx);
                        }
                    }
                    sink.write("</" + ELEMENT_ROW + '>');
                }
                sink.write("</" + ELEMENT_ROWS + '>');

                rs.close();
            } catch (SQLException e) {
                throw new IOException("Failed to marshall RowSet due to database error: " + e.getMessage());
            } catch (Exception e) {
                e.printStackTrace();
            }

            sink.write("</" + context + '>');
        }
        nsStack.popScope();
    }

    public Bean unmarshall(String inScopeEncStyle, final QName elementType, final Node src, final XMLJavaMappingRegistry xjmr, final SOAPContext ctx) {
        if (SoapEncUtils.isNull((Element) src))
            return new Bean(RowSet.class, null);

        try {
            CachedRowSet crs = new CachedRowSet();
            crs.setWriter(null);

            for (Node childNode = src.getFirstChild(); childNode != null; childNode = childNode.getNextSibling()) {
                if (childNode.getNodeType() == Node.ELEMENT_NODE && childNode.getNodeName().equals(ELEMENT_COLS)) {
                    RowSetMetaDataImpl rsmd = new RowSetMetaDataImpl();
                    rsmd.setColumnCount(DOMUtils.countKids((Element) childNode, Node.ELEMENT_NODE));
                    int colCount = 1;
                    for (Node colNode = childNode.getFirstChild(); colNode != null; colNode = colNode.getNextSibling()) {
                        if (colNode.getNodeType() == Node.ELEMENT_NODE && colNode.getNodeName().equals(ELEMENT_COL))
                            rsmd.setColumnName(colCount++, DOMUtils.getChildCharacterData((Element) colNode));
                    }
                    crs.setMetaData(rsmd);
                } else if (childNode.getNodeType() == Node.ELEMENT_NODE && childNode.getNodeName().equals(ELEMENT_ROWS)) {
                    for (Node rowNode = childNode.getFirstChild(); rowNode != null; rowNode = rowNode.getNextSibling()) {
                        if (rowNode.getNodeType() == Node.ELEMENT_NODE && rowNode.getNodeName().equals(ELEMENT_ROW)) {
                            crs.moveToInsertRow();
                            for (Node colNode = rowNode.getFirstChild(); colNode != null; colNode = colNode.getNextSibling()) {
                                if (colNode.getNodeType() == Node.ELEMENT_NODE) {
                                    if (!SoapEncUtils.isNull((Element) colNode)) {
                                        QName soapType = SoapEncUtils.getAttributeValue((Element) colNode, Constants.NS_URI_1999_SCHEMA_XSI, Constants.ATTR_TYPE, colNode.getNodeName(), true);

                                        crs.updateObject(colNode.getNodeName(), xjmr.unmarshall(inScopeEncStyle, soapType, colNode, ctx).value);
                                    } else
                                        crs.updateNull(colNode.getNodeName());
                                }
                            }
                            crs.insertRow();
                        }
                    }
                }
            }
            crs.moveToCurrentRow();
            crs.acceptChanges();
            crs.beforeFirst();

            return new Bean(RowSet.class, crs);
        } catch (SQLException e) {
            throw new RuntimeException("Failed to unmarshall RowSet due to database error: " + e.getMessage());
        }
    }
}