/*
 * SoapServlet.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.soap.server;

import java.io.FileReader;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;

import org.apache.soap.server.DeploymentDescriptor;
import org.apache.soap.server.ServiceManager;
import org.apache.soap.server.http.RPCRouterServlet;
import org.apache.soap.server.http.ServerHTTPUtils;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.mail.Mailer;

/**
 *
 * @author  Carlo Ravagli (carlo.ravagli@eXtropia.com)
 * @version 0.1
 */
public class SoapServlet extends RPCRouterServlet {
    public static final String CFG_SERVICE = "Service";

    public static final String CONFIG_FILENAME = "ConfigFile";
    public static final String SERVICES_CONFIG_FILENAME = "ServicesConfig";

    private Doctor _doctor;

    public SoapServlet() {
    }

    public void init() throws javax.servlet.ServletException {
        super.init();
        ServletConfig servletConfig = getServletConfig();
        ServletContext context = servletConfig.getServletContext();
        String filename = servletConfig.getInitParameter(SERVICES_CONFIG_FILENAME);
        Config config = null;

        if (filename == null) {
            System.err.println(SERVICES_CONFIG_FILENAME + " parameter not specified");
            return;
        }

        filename = filename.trim();
        if (filename.length() == 0) {
            System.err.println(SERVICES_CONFIG_FILENAME + " parameter must specify a valid filename ");
            return;
        }

        try {
            config = new Config(new FileReader(filename));
        } catch (Exception e) {
            System.err.println("Failed to open specified config file (" + filename + "): " + e.getMessage());
            return;
        }

        if (!config.hasChild(Diagnostics.CFG_BASE)) {
            System.err.println("No diagnostics configuration specifed.");
            return;
        }

        try {
            Diagnostics.configureFactory(config.getChild(Diagnostics.CFG_BASE));
        } catch (ConfigException ce) {
            System.err.println("Failed to configure diagnostics: " + ce.getMessage());
            return;
        }

        _doctor = Diagnostics.getDoctor(new String[] { "SOAP", "Server" });
        _doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Initialising SOAP Server"));

        if (config.hasChild(Mailer.CFG_BASE)) {
            try {
                _doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Configuring Mailer"));
                Mailer.configureMailer(config.getChild(Mailer.CFG_BASE));
            } catch (ConfigException ce) {
                System.err.println("Failed to configure mailer: " + ce.getMessage());
            }
        }

        _doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Configuring SOAP objects"));

        filename = servletConfig.getInitParameter(CONFIG_FILENAME).trim();
        ServiceManager serviceManager = ServerHTTPUtils.getServiceManagerFromContext(context, filename);
        Config[] serviceConfigs = config.getChildren();

        for (int i = 0; i < serviceConfigs.length; i++) {
            if (serviceConfigs[i].getName().equals(CFG_SERVICE)) {
                String urn = serviceConfigs[i].getValue();

                try {
                    DeploymentDescriptor dd = serviceManager.query(urn);

                    if (dd.getScope() != DeploymentDescriptor.SCOPE_APPLICATION && dd.getProviderType() != DeploymentDescriptor.PROVIDER_JAVA && !dd.getIsStatic()) {
                        String errmsg = "Unable to configure deployed service " + urn;
                        _doctor.log(new LogEvent(this, LogEvent.CRIT, errmsg));
                        continue;
                    }

                    Object soapObject = context.getAttribute(urn);

                    if (soapObject == null) {
                        soapObject = Class.forName(dd.getProviderClass()).newInstance();
                        context.setAttribute(urn, soapObject);
                    }

                    if (soapObject instanceof Configurable) {
                        Config[] objectConfigs = serviceConfigs[i].getChildren();
                        for (int j = 0; j < objectConfigs.length; j++) {
                            String msg = "Configuring service " + urn + " -> " + objectConfigs[j].getName();
                            _doctor.trace(new TraceEvent(this, TraceEvent.INFO, msg));
                            ((Configurable) soapObject).configure(objectConfigs[j]);
                        }
                    }
                } catch (Exception e) {
                    _doctor.log(new LogEvent(this, LogEvent.CRIT, "Failed to configure service " + urn + ": " + e.getMessage()));
                    e.printStackTrace();
                }
            }
        }

        _doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Finished Initialisation"));
    }
}
