/*
 * Base64.java
 * Copyright (C) 2000 - 2002  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.util;

import java.util.StringTokenizer;

/**
 * A simple Base64 Coder/Decoder.
 *
 * Created on December 28, 2001, 1:59 PM
 * @author <a href="mailto:rayfd@eXtropia.com">Ray Djajadinata</a>
 * @version $Id: Base64.java,v 1.5 2002/11/15 15:45:53 rayfd Exp $
 */
public class Base64 {
	
    /**
     * Get off my constructor!
     */
    private Base64() {}
	
    /**
     * The default number of characters per line in the 
     * Base64 string. RFC 2045 says the string must be broken
     * into lines of no more than 76 chars each.
     */
    public static final int DEFAULT_CHARS_PER_LINE = 76;

    /**
     * The default string that separates each line
     */
    public static final String DEFAULT_LINE_SEPARATOR = "\n";

    public static String encode(byte[] toEncode) {
        return encode(toEncode, DEFAULT_CHARS_PER_LINE, false, DEFAULT_LINE_SEPARATOR);
    }

    public static String encode(byte[] toEncode, int charsPerLine) {
        return encode(toEncode, charsPerLine, false, DEFAULT_LINE_SEPARATOR);
    }

    public static String encode(byte[] toEncode, int charsPerLine, boolean wrapPrecisely) {
        return encode(toEncode, charsPerLine, wrapPrecisely, DEFAULT_LINE_SEPARATOR);
    }
	
    /**
     * Translates a byte array into a Base64 string as per RFC1521
     */
    public static String encode(byte[] toEncode, int charsPerLine, boolean wrapPrecisely, String lineSeparator) {
        // get the length of the whole thing and count
        // the number of 24-bit input group in the message.
        int len = toEncode.length;
        int numGroups = len / 3;

        // find out -- how many remaining bytes are there?
        int numRemainingBytes = len % 3;

        // calculate the result length. We're gonna get 4 for every
        // 3 bytes of the input, and the remaining bytes are padded,
        // so, divide aLen by 3, ceil() it, and * 4.
        int resultLen = 4 * ( (len + 2) / 3 );
        
        // adjust it for the separators. For each line there is one
        // separator, but no separator at the end of the last one,
        // so it's ok to floor() it.
        int numSeparators = resultLen / charsPerLine;
        resultLen += numSeparators * lineSeparator.length();
        
        StringBuffer result = new StringBuffer(resultLen);

        // Translate those from the full 24-bit groups first. Take
        // 3 bytes at a time, translates them to 4 characters.
        int toEncodeIndex = 0;
        int lineLength = 0;
        int insertAt = 0;
        int flooredCharsPerLine = (charsPerLine / 4) * 4;
        for (int i = 0; i < numGroups; i++, lineLength += 4) {
            // remember that we need to break the line everytime
            // the line length == charsPerLine
            if(wrapPrecisely) {
                while(lineLength >= charsPerLine) {
                    insertAt += charsPerLine;
                    result.insert(insertAt, lineSeparator);
                    insertAt += lineSeparator.length();
                    lineLength -= charsPerLine;
                }			        	
            } else {
                // use the much faster algorithm--just round down 
                // the charsPerLine to the nearest multiple of 4
                if(lineLength + 4 > flooredCharsPerLine) {
                    result.append(lineSeparator);
                    lineLength = 0;
                }			        	
            }

            int byte0 = toEncode[toEncodeIndex++] & 0xff;
            int byte1 = toEncode[toEncodeIndex++] & 0xff;
            int byte2 = toEncode[toEncodeIndex++] & 0xff;
            // get the first 6 bits of the first byte and index to the lookup table
            result.append(LOOKUP_TABLE_TO_BASE64[byte0 >> 2]);
            // get the remaining 2 bits plus 4 of the second byte
            result.append(LOOKUP_TABLE_TO_BASE64[(byte0 << 4) & 0x3f | (byte1 >> 4)]);
            // and so on...
            result.append(LOOKUP_TABLE_TO_BASE64[(byte1 << 2) & 0x3f | (byte2 >> 6)]);
            result.append(LOOKUP_TABLE_TO_BASE64[byte2 & 0x3f]);
        }

        if(!wrapPrecisely) {
            if(lineLength == flooredCharsPerLine) {
                result.append(lineSeparator);
            }
        }

        // Translate the remaining bytes if present and pad as necessary
        if (numRemainingBytes != 0) {
            int byte0 = toEncode[toEncodeIndex++] & 0xff;
            result.append(LOOKUP_TABLE_TO_BASE64[byte0 >> 2]);
            if (numRemainingBytes == 1) {
                result.append(LOOKUP_TABLE_TO_BASE64[(byte0 << 4) & 0x3f]);
                result.append("==");
            } else {
                int byte1 = toEncode[toEncodeIndex++] & 0xff;
                result.append(LOOKUP_TABLE_TO_BASE64[(byte0 << 4)&0x3f | (byte1 >> 4)]);
                result.append(LOOKUP_TABLE_TO_BASE64[(byte1 << 2)&0x3f]);
                result.append('=');
            }
        }
        
        // take care of the slow algorithm again...
        if(wrapPrecisely) {
            // 4 cos the remaining bytes always translates to four after padding
            lineLength += 4;
            while(lineLength >= charsPerLine) {
                insertAt += charsPerLine;
                result.insert(insertAt, lineSeparator);
                insertAt += lineSeparator.length();
                lineLength -= charsPerLine;
            }			        	
        }
        return result.toString();
    }

    /**
     * This is a lookup table that translates 6-bit positive integer
     * index values into Base64 characters.
     */
    private static final char LOOKUP_TABLE_TO_BASE64[] =
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".toCharArray();

    /**
     * Translates back a Base64 string to the corresponding byte array, as per RFC1521.
     *
     * @throw IllegalArgumentException if the argument is not a valid Base64
     *        string.
     */
    public static byte[] decode(String toDecode) {
        // first of all, remove all whitespace from string
        StringTokenizer tok = new StringTokenizer(toDecode, " \n\r\t", false);
        StringBuffer buf = new StringBuffer(toDecode.length());

        while (tok.hasMoreElements()) {
            buf.append(tok.nextToken());
        }
		
        toDecode = buf.toString();

        // find out: how many 4-byte groups are there? If not multiple of four,
        // then it's an invalid Base64 String, throw an exception.
        int len = toDecode.length();
        int numGroups = len / 4;
        if((4 * numGroups) != len) {
            throw new IllegalArgumentException(
                "Invalid Base64 string! Length must be a multiple of four.");
        }

        // how many bytes are padding bytes in the last group, actually?
        // if there's padding, then the number of full groups are 1 less than
        // numGroups.
        int paddingBytes = 0;
        int numFullGroups = numGroups;
        if (len != 0) {
            if (toDecode.charAt(len - 1) == '=') {
                paddingBytes++;
                numFullGroups--;
            }
            if (toDecode.charAt(len - 2) == '=') {
                paddingBytes++;
            }
        }
        byte[] result = new byte[(3 * numGroups) - paddingBytes];

        // Translate full groups from base64 to bytes.
        int toDecodeIndex = 0, resultIndex = 0;
        for (int i=0; i < numFullGroups; i++) {
            int ch0 = translateToByte(toDecode.charAt(toDecodeIndex++));
            int ch1 = translateToByte(toDecode.charAt(toDecodeIndex++));
            int ch2 = translateToByte(toDecode.charAt(toDecodeIndex++));
            int ch3 = translateToByte(toDecode.charAt(toDecodeIndex++));
            result[resultIndex++] = (byte) ((ch0 << 2) | (ch1 >> 4));
            result[resultIndex++] = (byte) ((ch1 << 4) | (ch2 >> 2));
            result[resultIndex++] = (byte) ((ch2 << 6) | ch3);
        }

        // Translate the partial group.
        if (paddingBytes > 0) {
            int ch0 = translateToByte(toDecode.charAt(toDecodeIndex++));
            int ch1 = translateToByte(toDecode.charAt(toDecodeIndex++));
            result[resultIndex++] = (byte) ((ch0 << 2) | (ch1 >> 4));

            if (paddingBytes == 1) {
                int ch2 = translateToByte(toDecode.charAt(toDecodeIndex++));
                result[resultIndex++] = (byte) ((ch1 << 4) | (ch2 >> 2));
            }
        }
        return result;
    }

    private static int translateToByte(char c) {
        int result = LOOKUP_TABLE_FROM_BASE64[c];
        if (result < 0) {
            throw new IllegalArgumentException("Illegal character '" + c + "'.");
        }
        return result;
    }

    /**
     * This is a lookup table that translates a Base64 alphabet character into
     * its byte value. For instance, 'A' is 65, but it's assigned 0 in Base64,
     * so at position 65 in this lookup table, we have 0. 'B' is 66 so we have
     * 1 there, and so on and so on.
     */
    private static final byte LOOKUP_TABLE_FROM_BASE64[] = {
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63, 52, 53, 54,
        55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1, -1, 0, 1, 2, 3, 4,
        5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23,
        24, 25, -1, -1, -1, -1, -1, -1, 26, 27, 28, 29, 30, 31, 32, 33, 34,
        35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51
    };
}
