/*
 * ExtropiaException.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.util;

import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;

/**
 * A nested exception structure
 * <P>
 * This class allows the nesting of exceptions i.e. if an application catches
 * an exception and then throws it's own, it may store the caught exception(s)
 * for interested parties up the chain.
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 0.1
 */
public class ExtropiaException extends Exception {
    private ExtropiaException _parent;
    private ArrayList _children = new ArrayList(1);
    private Throwable _nestedException;

    /**
     * Creates new <code>ExtropiaException</code> without detail message.
     */
    public ExtropiaException() {
    }

    /**
     * Creates new <code>ExtropiaException</code> with the specified child
     * @param child the child exception
     */
    public ExtropiaException(ExtropiaException child) {
        addChild(child);
    }

    /**
     * Constructs an <code>ExtropiaException</code> with the specified detail message.
     * @param msg the detail message.
     */
    public ExtropiaException(String msg) {
        super(msg);
    }

    /**
    * Constructs an <code>ExtropiaException</code> with the specified
    * nested exception.
    * @param nestedException is a Throwable object.
    */
    public ExtropiaException(Throwable nestedException) {
        _nestedException = nestedException;
    }

    /**
     * Creates new <code>ExtropiaException</code> with the specified detail and child
         * @param msg the detail message.
     * @param child the child exception
     */
    public ExtropiaException(String msg, ExtropiaException child) {
        this(msg);
        addChild(child);
    }

    /**
     * Creates new <code>ExtropiaException</code> with the specified
         * nested exception and other child ExtropiaExceptions
         * @param nestedException is a throwable object
     * @param child the child exception
     */
    public ExtropiaException(Throwable nestedException, ExtropiaException child) {
        _nestedException = nestedException;
        addChild(child);
    }

    /**
     * Adds a new child <code>ExtropiaException</code>
     */
    public void addChild(ExtropiaException child) {
        child._parent = this;
        _children.add(child);
    }

    /**
     * @return The parent exception or null if no parent is specified
     */
    public ExtropiaException getParentException() {
        return _parent;
    }

    /**
     * @return An array of child exceptions
     */
    public ExtropiaException[] getChildren() {
        Object[] objs = _children.toArray();
        ExtropiaException[] exceptions = new ExtropiaException[objs.length];
        System.arraycopy(objs, 0, exceptions, 0, objs.length);
        return exceptions;
    }

    /**
     * @return <code>true</code> if the <code>ExtropiaException</code> has a child exception(s)
     */
    public boolean hasChildren() {
        return !_children.isEmpty();
    }

    public void setNestedException(ExtropiaException nestedException) {
        _nestedException = nestedException;
    }
    /**
    * @return <code>String</code> indicating the messages of all nested eXtropia exceptions
    */
    public String getExtendedMessage() {
        StringBuffer buffer = new StringBuffer();
        buffer.append(getMessage());
        ExtropiaException[] ees = getChildren();
        for (int i = 0; i < ees.length; i++) {
            buffer.append("\n-- " + ees[i].getMessage());
        }
        return buffer.toString();
    }

    /**
     * @return <code>Throwable</code> if the <code>ExtropiaException</code>
         * encapsulates a nested exception
     */
    public Throwable getNestedException() {
        return _nestedException;
    }

    /**
     * @return <code>String</code> indicating the message of the given
         * message at construction time or the nested exception's message
         * if this instance was constructed with a Throwable instead of
         * a message
     */
    public String getMessage() {
        if (_nestedException != null) {
            return _nestedException.getMessage();
        }

        return super.getMessage();
    }

    /**
     * Prints the stack trace of the nested exception or just
         * prints the stack trace of the original exception if no
         * nested exception is found.
     */
    public void printStackTrace() {
        if (_nestedException != null) {
            System.err.println("Nested Exception: ");
            _nestedException.printStackTrace();
        }

        super.printStackTrace();
    }

    /**
     * Prints the stack trace of the nested exception or just
         * prints the stack trace of the original exception if no
         * nested exception is found.
     */
    public void printStackTrace(PrintStream ps) {
        if (_nestedException != null) {
            ps.println("Nested Exception: ");
            _nestedException.printStackTrace(ps);
        }
        super.printStackTrace(ps);
    }

    /**
     * Prints the stack trace of the nested exception or just
         * prints the stack trace of the original exception if no
         * nested exception is found.
     */
    public void printStackTrace(PrintWriter pw) {
        // Cassandra updated 7/6/2002 Corrected an error here. If the nestedException is null printStackTrace method would give a NullPointerException.

        if (_nestedException != null) {
            //if (_nestedException == null) {
            pw.println("Nested Exception: ");
            _nestedException.printStackTrace(pw);
        }

        super.printStackTrace(pw);

    }

    /**
     * Prints the stack trace of the child ExtropiaExceptions
         * that may have been passed to this one
     */
    public void printChildrenStackTrace() {
        ExtropiaException[] ee = getChildren();

        if (ee != null)
            for (int i = 0; i < ee.length; i++) {
                System.err.println("Child Exception " + i + ":");
                ee[i].printStackTrace();
            }

    }

    /**
     * Prints the stack trace of the child ExtropiaExceptions
         * that may have been passed to this one
     */
    public void printChildrenStackTrace(PrintStream ps) {
        ExtropiaException[] ee = getChildren();

        if (ee != null)
            for (int i = 0; i < ee.length; i++) {
                ps.println("Child Exception " + i + ":");
                ee[i].printStackTrace(ps);
            }

    }

    /**
     * Prints the stack trace of the child ExtropiaExceptions
         * that may have been passed to this one
     */
    public void printChildrenStackTrace(PrintWriter pw) {
        ExtropiaException[] ee = getChildren();

        if (ee != null)
            for (int i = 0; i < ee.length; i++) {
                pw.println("Child Exception " + i + ":");
                ee[i].printStackTrace(pw);
            }

    }

}
