/*
 * JndiBeanBinder.java 
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.util;

import java.beans.Beans;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.InitialContext;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;

/**
 * Created on December 27, 2001, 11:56 AM
 *
 * @author Carlo Ravagli
 * @version 1.0
 */
public class JndiBeanBinder implements Configurable {
    public static final String CFG_BASE = "JndiBeanBinder";
    public static final String CFG_INITIAL_CONTEXT_FACTORY = "InitialContextFactory";
    public static final String CFG_PROVIDER_URL = "ProviderUrl";
    public static final String CFG_BEAN = "Bean";

    private Doctor _doctor;

    public JndiBeanBinder() throws ConfigException {
        try {
            String className = getClass().getName();
            className = className.substring(className.lastIndexOf('.') + 1);
            InputStream xmlConfigFile = getClass().getResourceAsStream(className + ".xml");
            if (xmlConfigFile != null)
                configure(new Config(new InputStreamReader(xmlConfigFile)));
        } catch (ConfigException ce) {
            String msg = "Failed to start due to configuration error: " + ce.getMessage();
            _doctor.log(new LogEvent(this, LogEvent.CRIT, msg));
            throw ce;
        }
    }

    public JndiBeanBinder(String configFile) throws ConfigException {
        try {
            configure(new Config(new FileReader(configFile)));
        } catch (FileNotFoundException e) {
            String msg = "Unable to find config file: " + configFile;
            _doctor.log(new LogEvent(this, LogEvent.CRIT, msg));
            throw new ConfigException(msg);
        } catch (ConfigException ce) {
            String msg = "Configuration error: " + ce.getMessage();
            _doctor.log(new LogEvent(this, LogEvent.CRIT, msg));
            throw ce;
        }
    }

    public synchronized void configure(Config config) throws ConfigException {
        if (config.hasChild(Diagnostics.CFG_BASE))
            Diagnostics.configureFactory(config.getChild(Diagnostics.CFG_BASE));

        _doctor = Diagnostics.getDoctor(new String[0]);
        Context ctx = null;

        try {
            Properties props = null;

            if (config.hasChild(CFG_INITIAL_CONTEXT_FACTORY)) {
                String contextFactory = config.getChildValue(CFG_INITIAL_CONTEXT_FACTORY);
                props = new Properties();
                props.setProperty(Context.INITIAL_CONTEXT_FACTORY, contextFactory);
            }

            if (config.hasChild(CFG_PROVIDER_URL)) {
                String providerUrl = config.getChildValue(CFG_PROVIDER_URL);
                if (props == null)
                    props = new Properties();
                props.setProperty(Context.PROVIDER_URL, providerUrl);
                if (providerUrl.startsWith("rmi:"))
                    checkRegistry(providerUrl);
            }

            ctx = new InitialContext(props);
        } catch (Exception e) {
            e.printStackTrace();
            throw new ConfigException("Unable to get JNDI initial context: " + e.getMessage());
        }

        Config[] childConfigs = config.getChildren();

        for (int i = 0; i < childConfigs.length; i++) {
            if (childConfigs[i].getName().equals(CFG_BEAN)) {
                String type = childConfigs[i].getType();
                String name = childConfigs[i].getValue();
                Object bean = null;

                if (type == null)
                    throw new ConfigException("Invalid bean config: No type attribute specifed");

                if (name == null)
                    throw new ConfigException("Invalid bean config: Bind name not specifed");

                try {
                    bean = Beans.instantiate(null, type);
                } catch (Exception e) {
                    String msg = "Failed to instantiate bean " + name + " of type " + type + ": " + e.getMessage();
                    throw new ConfigException(msg);
                }

                Config[] beanPropsConfig = childConfigs[i].getChildren();
                for (int p = 0; p < beanPropsConfig.length; p++) {
                    String paramTypeName = beanPropsConfig[p].getType();

                    if (paramTypeName == null) {
                        String msg = "Invalid bean config " + name + " property " + beanPropsConfig[p].getName() + " has no type attribute specifed";
                        throw new ConfigException(msg);
                    }

                    try {
                        Class paramType = paramTypeName.equals("boolean") ? Boolean.TYPE : paramTypeName.equals("char") ? Character.TYPE : paramTypeName.equals("byte") ? Byte.TYPE : paramTypeName.equals("short") ? Short.TYPE : paramTypeName.equals("int") ? Integer.TYPE : paramTypeName.equals("long") ? Long.TYPE : paramTypeName.equals("float") ? Float.TYPE : paramTypeName.equals("double") ? Double.TYPE : Class.forName(paramTypeName);
                        Method method = bean.getClass().getMethod("set" + beanPropsConfig[p].getName(), new Class[] { paramType });

                        if (paramType.isPrimitive()) {
                            paramType = paramType.equals(Boolean.TYPE) ? Boolean.class : paramType.equals(Character.TYPE) ? Character.TYPE : paramType.equals(Byte.TYPE) ? Byte.class : paramType.equals(Short.TYPE) ? Short.class : paramType.equals(Integer.TYPE) ? Integer.class : paramType.equals(Long.TYPE) ? Long.class : paramType.equals(Float.TYPE) ? Float.class : paramType.equals(Double.TYPE) ? Double.class : Void.class;
                        }
                        Constructor constructor = paramType.getConstructor(new Class[] { String.class });
                        Object param = constructor.newInstance(new Object[] { beanPropsConfig[p].getValue()});
                        method.invoke(bean, new Object[] { param });
                    } catch (Exception e) {
                        String msg = "Invalid bean config " + name + " for property " + beanPropsConfig[p].getName();
                        throw new ConfigException(msg);
                    }
                }

                try {
                    ctx.rebind(name, bean);
                } catch (Exception e) {
                    String msg = "Failed to bind bean " + name + " of type " + type + " to JNDI Provider: " + e.getMessage();
                    throw new ConfigException(msg);
                }
            }
        }
    }

    private void checkRegistry(String url) throws MalformedURLException, UnknownHostException, RemoteException {
        String host = "";
        int port = Registry.REGISTRY_PORT;

        // remove the approved protocol
        if (!url.startsWith("rmi://")) {
            throw new MalformedURLException("Invalid/Unspecifed URL protocol for registry JNDI provider: " + url);
        }

        url = url.substring(6);

        if (url.endsWith("/"))
            url = url.substring(0, url.length());

        int colon = url.indexOf(":");
        if (colon > 0) {
            host = url.substring(0, colon);
            // explicit port supplied
            try {
                port = Integer.parseInt(url.substring(colon + 1));
            } catch (NumberFormatException e) {
                throw new MalformedURLException("Invalid port number: " + url);
            }
        } else
            host = url;

        if (host.length() == 0)
            throw new MalformedURLException("Hostname unspecifed: " + url);

        try {
            LocateRegistry.getRegistry(host, port).list();
        } catch (RemoteException e) {
            boolean createRegistry = host.equals("127.0.0.1") || host.equals("localhost");
            // Could not find registry therefore create one
            if (!createRegistry) {
                // Check if the host specfied was the local host
                InetAddress[] hostIPs = InetAddress.getAllByName(host);
                InetAddress localhost = InetAddress.getLocalHost();
                for (int i = 0; !createRegistry && i < hostIPs.length; i++)
                    createRegistry = hostIPs[i].equals(localhost);
            }

            if (createRegistry)
                LocateRegistry.createRegistry(port);
            else
                throw e;
        }
    }

    public static void main(String args[]) throws Exception {
        if (args.length > 0)
            new JndiBeanBinder(args[0]);
        else
            new JndiBeanBinder();
    }
}
