/*
 * XMLParser.java
 * Copyright (C) 2000 - 2002  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.util;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.Properties;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class XMLParser {

    protected Element _element;

    public XMLParser(String fileName) throws FileNotFoundException, XMLParseException {
        this(new FileReader(fileName));
    }

    public XMLParser(Reader xmldocument) throws XMLParseException {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        Document document = null;

        try {
            builder = factory.newDocumentBuilder();
            document = builder.parse(new InputSource(xmldocument));
        } catch (ParserConfigurationException pce) {
            throw new IllegalArgumentException("Failed to parser File provided");
        } catch (IOException ioe) {
            throw new IllegalArgumentException("Unable to use File provided");
        } catch (SAXException se) {
            throw new XMLParseException("File specified is not of the correct format: " + se.getMessage());
        }

        _element = (Element) document.getFirstChild();

    }

    /**
     * Creates a new Properties object referencing a node in a DOM created using
     * one of the public methods.
     */
    private XMLParser(Element element) {
        _element = element;
    }

    public void addChild(XMLParser parser) {
        _element.appendChild(parser._element);
    }

    public XMLParser addChild(String name, String value) {
        Document owner = _element.getOwnerDocument();
        Element newElement = owner.createElement(name);
        newElement.appendChild(owner.createTextNode(value));
        _element.appendChild(newElement);
        return new XMLParser(newElement);
    }

    /**
     * @return The FIRST XMLParser tag matching the specified key
     */
    public XMLParser getChild(String key) {
        NodeList nodes = _element.getChildNodes();
        Node node;
        for (int i = 0; i < nodes.getLength(); i++) {

            node = nodes.item(i);

            if ((node.getNodeType() == Node.ELEMENT_NODE) && node.getNodeName().equals(key))
                return new XMLParser((Element) node);
        }

        return null;
    }

    /**
     * @return An iterator over the child Properties objects.
     */
    public XMLParser[] getChildren() {
        NodeList nodes = _element.getChildNodes();
        Node node = null;
        Vector v = new Vector();
        for (int i = 0; i < nodes.getLength(); i++) {
            node = nodes.item(i);

            if (node.getNodeType() == Node.ELEMENT_NODE)
                v.addElement(new XMLParser((Element) node));
        }

        //now copy the Vector into a XMLParser []
        Object[] objs = v.toArray();
        XMLParser[] children = new XMLParser[objs.length];
        System.arraycopy(objs, 0, children, 0, objs.length);
        return children;
    }

    public String getName() {
        return _element.getNodeName();
    }

    private String getText(Element element, boolean trim) {
        NodeList nodes = element.getChildNodes();
        Node node;
        String res = "";
        for (int j = 0; j < nodes.getLength(); j++) {
            node = nodes.item(j);
            if (node.getNodeType() == Node.TEXT_NODE) {
                res += node.getNodeValue();
            }
        }
        if (trim)
            res = res.trim();
        return res.equals("") ? null : res;
    }

    private String getCData(Element element, boolean trim) {

        NodeList nodes = element.getChildNodes();
        Node node;
        String res = "";
        for (int j = 0; j < nodes.getLength(); j++) {
            node = nodes.item(j);
            if (node.getNodeType() == Node.CDATA_SECTION_NODE) {
                res += node.getNodeValue();
            }
        }
        if (trim)
            res = res.trim();
        return res.equals("") ? null : res;

    }

    public String getValue() {
        return getText(_element, true);
    }

    public String getCDataValue() {
        return getCData(_element, false);
    }

    public boolean getBooleanValue() {
        return Boolean.valueOf(getValue()).booleanValue();
    }

    public int getIntValue() throws XMLParseException {
        try {
            return Integer.parseInt(getValue());
        } catch (NumberFormatException nfe) {
            throw new XMLParseException(getName() + " should be an integer");
        }
    }

    public String getRealValue() {
        return getText(_element, false);
    }

    public String getChildValue(String child) throws XMLParseException {
        return getChildValue(child, true);
    }

    public String getChildCDValue(String child) throws XMLParseException {
        XMLParser childParser = getChild(child);
        String value = null;

        if (childParser != null) {
            value = childParser.getCDataValue();
        }

        return value;

    }

    public String getChildValue(String child, boolean required) throws XMLParseException {
        XMLParser childParser = getChild(child);
        String value = null;

        if (required && childParser == null)
            throw new XMLParseException("No <" + child + "> specifed in " + getName() + " config");

        if (childParser != null) {
            value = childParser.getValue();
            if (required && value == null)
                throw new XMLParseException("No <" + child + "> value specifed in " + getName() + " config");
        }

        return value;
    }

    public boolean hasChild(String key) {
        return !(getChild(key) == null);
    }

    public boolean hasChildren() {
        NodeList nodes = _element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++) {
            if (nodes.item(i).getNodeType() == Node.ELEMENT_NODE)
                return true;
        }
        return false;

    }

    public void removeChild(XMLParser oneToDelete) {
        _element.removeChild(oneToDelete._element);
    }

    public void save(Writer w) throws IOException {
        try {
            TransformerFactory transformerFactory = TransformerFactory.newInstance();
            Transformer transformer = transformerFactory.newTransformer();

            DOMSource source = new DOMSource(_element.getOwnerDocument());
            StreamResult result = new StreamResult(w);
            transformer.transform(source, result);

        } catch (TransformerConfigurationException e) {
            e.printStackTrace();
        } catch (TransformerException e) {
            e.printStackTrace();
        }
    }

    private void setAttribute(String attr, String value) {
        //get out the attribute and setValue
        if (value == null || value.equals("")) {
            if (_element.getAttributeNode(attr) != null) {
                _element.removeAttribute(attr);
            }
        } else {
            _element.setAttribute(attr, value);
        }
    }

    public void setValue(String value) {
        //remove all the old text nodes
        NodeList nodes = _element.getChildNodes();
        Node node;
        for (int i = 0; i < nodes.getLength(); i++) {
            node = nodes.item(i);
            if (node.getNodeType() == Node.TEXT_NODE) {
                _element.removeChild(node);
            }
        }
        Document owner = _element.getOwnerDocument();

        _element.appendChild(owner.createTextNode(value));
    }

    /**
     * @return The XMLParser object expressed as a Properties object
     */
    public Properties toProperties() {
        Properties properties = new Properties();
        XMLParser[] children = getChildren();
        for (int i = 0; i < children.length; i++) {
            properties.put(children[i].getName(), children[i].getValue());
        }
        return properties;
    }

    public String toString() {
        return getName();
    }

}
