/*
 * X3DESCrypt.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Created on December 28, 2001, 1:59 PM
 */
package com.eXtropia.util.crypt;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.spec.SecretKeySpec;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.util.Base64;

/**
 * @author  Ray Djajadinata (rayfd@eXtropia.com)
 * @version 0.1
 */
public class X3DESCrypt implements XCrypt {
    public static final String CFG_3DES_KEY = "Key";
    private SecretKeySpec key = null;

    public synchronized void configure(Config config) throws ConfigException {
        try {
            RandomAccessFile raf = new RandomAccessFile(config.getChildValue(CFG_3DES_KEY), "r");
            // 3DES key is 24 bytes long.
            int len = (int) raf.length();
            if (len != 24) {
                throw new ConfigException("Invalid 3DES key in file " + config.getChildValue(CFG_3DES_KEY));
            }

            byte[] rawKey = new byte[len];
            raf.read(rawKey);
            raf.close();

            // import the key
            key = new SecretKeySpec(rawKey, "DESede");
        } catch (IOException ioEx) {
            throw new ConfigException(ioEx.getMessage());
        }
    }

    /**
     * Encrypts plaintext with the key obtained in configure(). The algorithm is
     * 3DES.
     *
     * @param plainText   the plain text to be encrypted
     * @return            the cipher text
     * @throws NoSuchPaddingException
     * @throws NoSuchAlgorithmException
     * @throws InvalidKeyException
     * @throws BadPaddingException
     * @throws IllegalBlockSizeException
     */
    public byte[] encrypt(byte[] plainText) throws NoSuchPaddingException, NoSuchAlgorithmException, InvalidKeyException, BadPaddingException, IllegalBlockSizeException {
        Cipher desCipher = Cipher.getInstance("DESede/ECB/PKCS5Padding");
        desCipher.init(Cipher.ENCRYPT_MODE, key);

        return desCipher.doFinal(plainText);
    }

    /**
     * Decrypts cipher text with the key obtained in configure(). The algorithm is
     * 3DES.
     *
     * @param cipherText    the cipher text to be decrypted
     * @return              the plain text
     * @throws NoSuchPaddingException
     * @throws NoSuchAlgorithmException
     * @throws InvalidKeyException
     * @throws BadPaddingException
     * @throws IllegalBlockSizeException
     */
    public byte[] decrypt(byte[] cipherText) throws NoSuchPaddingException, NoSuchAlgorithmException, InvalidKeyException, BadPaddingException, IllegalBlockSizeException {
        Cipher desCipher = Cipher.getInstance("DESede/ECB/PKCS5Padding");
        desCipher.init(Cipher.DECRYPT_MODE, key);

        return desCipher.doFinal(cipherText);
    }

    /**
     * Encrypts plaintext with the key obtained in configure() and Base64 encode
     * the result. The algorithm is 3DES.
     *
     * @param plainText   the plain text to be encrypted
     * @return            the cipher text
     * @throws NoSuchPaddingException
     * @throws NoSuchAlgorithmException
     * @throws InvalidKeyException
     * @throws BadPaddingException
     * @throws IllegalBlockSizeException
     */
    public String encryptAndBase64Encode(byte[] plainText) throws NoSuchPaddingException, NoSuchAlgorithmException, InvalidKeyException, BadPaddingException, IllegalBlockSizeException {

        Cipher desCipher = Cipher.getInstance("DESede/ECB/PKCS5Padding");
        desCipher.init(Cipher.ENCRYPT_MODE, key);

        return Base64.encode(desCipher.doFinal(plainText));
    }

    /**
     * Base64-decodes a string and decrypts the resulting cipher text with the
     * key obtained in configure(). The algorithm is 3DES.
     *
     * @param cipherText    the cipher text to be decrypted
     * @return              the plain text
     * @throws NoSuchPaddingException
     * @throws NoSuchAlgorithmException
     * @throws InvalidKeyException
     * @throws BadPaddingException
     * @throws IllegalBlockSizeException
     */
    public byte[] base64DecodeAndDecrypt(String base64CipherText) throws NoSuchPaddingException, NoSuchAlgorithmException, InvalidKeyException, BadPaddingException, IllegalBlockSizeException {

        byte[] cipherText = Base64.decode(base64CipherText);

        Cipher desCipher = Cipher.getInstance("DESede/ECB/PKCS5Padding");
        desCipher.init(Cipher.DECRYPT_MODE, key);

        return desCipher.doFinal(cipherText);
    }
}