/*
 * ViewManager.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.view;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.datahandler.DataHandler;
import com.eXtropia.datahandler.Rule;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.view.widget.Widget;

/**
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 1.0
 */
public class ViewManager implements Configurable {
    public static final String CFG_BASE = "ViewManager";
    public static final String CFG_WIDGETS = "Widgets";
    public static final String CFG_VIEWS = "Views";
    public static final String CFG_VIEW = "View";
    public static final String CFG_HEADER_LOCATION = "HeaderLocation";
    public static final String CFG_FOOTER_LOCATION = "FooterLocation";
    public static final String CFG_RESOURCE_BUNDLE = "ResourceBundle";
    public static final String CFG_VIEW_TITLE = "Title";
    public static final String CFG_WIDGET_ORDER = "WidgetOrder";
    public static final String CFG_WIDGET_ORDER_NAME = "Name";

    public static final String CFG_VIEW_LOCATION = "Location";
    public static final String CFG_VIEW_ID = "Id";

    public static final String VIEW_ATTRIBUTE_NAME = "view";

    //TODO: allow widgets to be tied to accessors
    //public static final String VIEW_ATTRIBUTE_DATA_ACCESSORS= "DataAccessors";

    public static final String REQUEST_ATTRIBUTE_VIEW_ATTRIBUTES = "ViewAttributes";
    public static final String REQUEST_ATTRIBUTE_NAME = "com.eXtropia.view.ViewManager";

    private Map _views = new HashMap(); //view name -> View Objects
    protected String _resourceIdentifier;
    protected String _resourceBundleLocation;
    protected Map _widgets;
    protected Map _rules;
    protected HashMap _viewAttributes = new HashMap();
    protected Doctor _doctor;
    protected String _headerLocation;
    protected String _footerLocation;

    public ViewManager(String appName, String appInstance) {
        _doctor = Diagnostics.getDoctor(new String[] { appInstance, appName, "ViewManager" });
    }

    public void addViewAttribute(String key, Object value) {
        _viewAttributes.put(key, value);
    }

    public void configure(Config config) throws ConfigException {
        HashMap defaultWidgets = null;

        //try and get out the ResouceBundle name
        if (config.hasChild(CFG_RESOURCE_BUNDLE))
            _resourceBundleLocation = config.getChild(CFG_RESOURCE_BUNDLE).getValue();
        else
            throw new ConfigException("No " + CFG_RESOURCE_BUNDLE + " specified in the configuration");

        if (!config.hasChild(CFG_WIDGETS))
            throw new ConfigException("No " + CFG_WIDGETS + " specified in " + CFG_BASE);

        _widgets = createWidgets(config.getChild(CFG_WIDGETS));

        String[] defaultWidgetOrder = getWidgetOrder(config);
        if (defaultWidgetOrder == null)
            throw new ConfigException("No " + CFG_WIDGET_ORDER + " specified in " + CFG_BASE);

        if (!config.hasChild(CFG_VIEWS))
            throw new ConfigException("No " + CFG_VIEWS + " specified in the configuration");
        Config[] viewConfigs = config.getChild(CFG_VIEWS).getChildren();

        String[] widgetOrder;
        String currentViewName;
        Config currentViewConfig;
        Map currentWidgets;

        for (int i = 0; i < viewConfigs.length; i++) {
            if (!viewConfigs[i].getName().equals(CFG_VIEW)) {
                if (viewConfigs[i].getName().equals(CFG_HEADER_LOCATION))
                    _headerLocation = viewConfigs[i].getValue();
                else if (viewConfigs[i].getName().equals(CFG_FOOTER_LOCATION))
                    _footerLocation = viewConfigs[i].getValue();
                continue;
            } else
                currentViewConfig = viewConfigs[i];
            //get the widgetorder for the current view
            widgetOrder = null;
            widgetOrder = getWidgetOrder(currentViewConfig);
            if (widgetOrder == null) {
                widgetOrder = (String[]) defaultWidgetOrder.clone();
            }
            currentWidgets = createWidgetMapFromOrder(widgetOrder);

            if (currentViewConfig.hasChild(CFG_VIEW_LOCATION)) {
                if (_doctor.doTrace(TraceEvent.INFO))
                    _doctor.trace(new TraceEvent(this, TraceEvent.INFO, currentViewConfig.getValue() + " => " + currentViewConfig.getChild(CFG_VIEW_LOCATION).getValue()));
            } else
                throw new ConfigException("No " + CFG_VIEW_LOCATION + " specified for view: " + currentViewConfig.getValue());
            _views.put(currentViewConfig.getValue(), new View(currentViewConfig.getValue(), currentViewConfig.getChild(CFG_VIEW_LOCATION).getValue(), _resourceBundleLocation, currentWidgets, widgetOrder, _viewAttributes));
        }
    }

    public Map createWidgetMapFromOrder(String[] widgetOrder) {
        Map map = new HashMap();
        Widget widget;
        for (int i = 0; i < widgetOrder.length; i++) {
            widget = (Widget) _widgets.get(widgetOrder[i]);
            if (widget != null) {
                widgetOrder[i] = widget.getId();
                //anderson - support index from both the 'id' & the 'name'
                //           This is just a temorary solution
                //map.put(widget.getId(), widget);
                map.put("id_" + widget.getId(), widget);
                map.put("name_" + widget.getName(), widget);
            }
        }
        return map;
    }

    protected Map createWidgets(Config config) throws ConfigException {
        Map widgets = new HashMap();
        Config[] children = config.getChildren();
        Widget widget;
        Rule rule;
        DataHandler[] datahandlers;
        String[] ids;
        for (int i = 0; i < children.length; i++) {
            if (children[i].getName().equals(Widget.CFG_BASE)) {
                try {
                    widget = (Widget) children[i].createAndConfigureInstance();
                    rule = widget.getRule();
                    datahandlers = rule.getDataHandlers();
                    for (int j = 0; j < datahandlers.length; j++) {
                        ids = datahandlers[j].getIds();
                        for (int k = 0; k < ids.length; k++) {
                            if (ids[k] == null)
                                throw new ConfigException("Widget " + widget.getName() + " has a datahandler (" + datahandlers[j].getName() + ") that required a null id");
                        }
                    }
                    widget.setResourceLocation(_resourceBundleLocation);
                    widgets.put(widget.getName(), widget);
                } catch (ClassCastException e) {
                    throw new ConfigException(config.getType() + " is not of type Widget");
                }
            }
        }
        return widgets;
    }

    protected String[] getWidgetOrder(Config config) throws ConfigException {
        if (config.hasChild(CFG_WIDGET_ORDER)) {
            Config[] children = config.getChild(CFG_WIDGET_ORDER).getChildren();
            List order = new ArrayList();
            for (int i = 0; i < children.length; i++) {
                if (children[i].getName().equals(CFG_WIDGET_ORDER_NAME)) {
                    if (_widgets.containsKey(children[i].getValue())) {
                        order.add(children[i].getValue());
                    } else
                        throw new ConfigException("No Widget specified by " + Widget.CFG_NAME + ": " + children[i].getValue());
                }
            }
            return (String[]) order.toArray(new String[order.size()]);
        } else
            return null;
    }

    public void showView(String viewName, ServletRequest request, ServletResponse response) throws NoSuchViewException, ServletException, IOException {
        if (_doctor.doTrace(TraceEvent.INFO))
            _doctor.trace(new TraceEvent(this, TraceEvent.VERBOSE, "Showing view: " + viewName));
        View view = (View) _views.get(viewName);
        if (view != null) {
            //TODO: allow widgets to be tied to accessors
            /*Class [] requiredAccessors= view.getRequiredDataAccessors();
            Map accessors= null;
            if(requiredAccessors != null)
            {
              accessors= new HashMap();
              //instantiate the accessors, set the request, UserGesture and UserGestureResponse and set
              //as a view attribute
              for(int i=0;i<requiredAccessors.length;i++)
              {
                try
                {
                  WidgetDataAccessor accessor= (WidgetDataAccessor)requiredAccessors[i].newInstance();
                  accessor.setServletRequest(request);
                  accessors.put(requiredAccessors[i], accessor);
                }
                catch(Exception e)
                {
                  //exceptions that are thrown here should already have been checked in the view constructor
                }
              }
            }*/
            request.setAttribute(VIEW_ATTRIBUTE_NAME, view);
            request.setAttribute(REQUEST_ATTRIBUTE_NAME, this);
            request.setAttribute(REQUEST_ATTRIBUTE_VIEW_ATTRIBUTES, _viewAttributes);
            /*
            if(accessors != null)
            {
              Map myAttributes= (Map)_viewAttributes.clone();
              myAttributes.put(ViewManager.VIEW_ATTRIBUTE_DATA_ACCESSORS, accessors);
              request.setAttribute(REQUEST_ATTRIBUTE_VIEW_ATTRIBUTES, myAttributes);
            }
            else
            {
              request.setAttribute(REQUEST_ATTRIBUTE_VIEW_ATTRIBUTES, _viewAttributes);
            }*/
            request.getRequestDispatcher(view.getLocation()).include(request, response);
        } else {
            throw new NoSuchViewException(viewName);
        }
    }

    public View getView(String viewName) {
        return (View) _views.get(viewName);
    }

    public String getHeaderLocation() {
        return _headerLocation;
    }

    public String getFooterLocation() {
        return _footerLocation;
    }
}