/*
 * Widget.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.view.widget;

import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.Vector;

import javax.servlet.ServletRequest;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.datahandler.DataHandler;
import com.eXtropia.datahandler.Rule;
import com.eXtropia.datahandler.handler.CopyDataHandler;
import com.eXtropia.datahandler.handler.RequiredHandler;
import com.eXtropia.view.WidgetDataAccessor;

/**
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 0.1
 */
public abstract class Widget implements Configurable, Cloneable {
    public static final String WIDGET_RESOURCE_BASE = "widget.";
    public static final String WIDGET_RESOURCE_LABEL = ".label";

    public static final String CFG_BASE = "Widget";
    public static final String CFG_ID = "Id";
    public static final String CFG_NAME = "Name";
    public static final String CFG_REQUIRED = "Required";
    public static final String CFG_ACCESSOR = "Accessor";
    public static final String CFG_DEFAULT = "Default";

    public static final String RULE_NAME_BASE = "rule";

    protected String _id;
    protected String[] _submitIds;
    protected String _name;
    protected boolean _default;
    protected String _resourceBase;
    protected String _resourceLocation;
    //anderson - additional attrs
    protected String _additionalAttrs;
    //anderson - real name
    protected String _realName;

    private Class _accessor;

    private Rule _rule;
    private boolean _required = false;

    public Widget() {

    }

    public void addDataHandler(DataHandler dataHandler) {
        if (_rule == null) {
            _rule = new Rule(_id);
            _rule.setMaxErrors(1);
        }
        //need to make sure the required handler is always executed last
        if (_required) {
            setRequired(false);
            _rule.add(dataHandler);
            setRequired(true);
        } else
            _rule.add(dataHandler);
    }

    protected void configureWidgetBase(Config config) throws ConfigException {
        if (config.getName().equals(CFG_BASE)) {
            if (config.hasChild(CFG_ID))
                setId(config.getChild(CFG_ID).getValue());
            else
                throw new ConfigException("Invalid widget id: " + config.getChild(CFG_ID).getValue());

            if (config.hasChild(CFG_NAME))
                setName(config.getChild(CFG_NAME).getValue());

            Config[] children = config.getChildren();
            for (int i = 0; i < children.length; i++) {
                if (children[i].getName().equals(DataHandler.CFG_BASE)) {
                    //System.out.println("Adding datahandler: " + getName());
                    addDataHandler((DataHandler) children[i].createAndConfigureInstance());
                }
            }

            if (config.hasChild(CFG_ACCESSOR)) {
                try {
                    _accessor = Class.forName(config.getChild(CFG_ACCESSOR).getValue());
                    WidgetDataAccessor instance = (WidgetDataAccessor) _accessor.newInstance();
                } catch (Exception e) {
                    throw new ConfigException(CFG_ACCESSOR + " specified for widget " + _name + " is invalid: " + e.getMessage());
                }
            }

            //add required handler last if required
            if (config.hasChild(CFG_REQUIRED))
                setRequired(Boolean.valueOf(config.getChild(CFG_REQUIRED).getValue()).booleanValue());
        } else
            throw new ConfigException("Invalid Config passed to " + CFG_BASE + ": " + config.getName());

        if (config.hasChild(CFG_DEFAULT))
            _default = config.getChild(CFG_DEFAULT).getBooleanValue();
    }

    public void configure(Config config) throws ConfigException {
        configureWidgetBase(config);
    }

    public static Map createWidgets(Config config) throws ConfigException {
        Map widgets = new HashMap();
        Config[] children = config.getChildren();
        Widget widget;
        for (int i = 0; i < children.length; i++) {
            if (children[i].getName().equals(CFG_BASE)) {
                try {
                    widget = (Widget) Class.forName(children[i].getType()).newInstance();
                    widget.configure(children[i]);
                    widgets.put(widget.getName(), widget);
                } catch (ConfigException ce) {
                    throw ce;
                } catch (Exception e) {
                    throw new ConfigException("Failed to create new widget: " + e.getMessage());
                }
            }
        }
        return widgets;
    }

    public Class getDataAccessorClass() {
        return _accessor;
    }

    protected Object getDefaultValue(Locale locale) {
        ResourceBundle bundle = ResourceBundle.getBundle(_resourceLocation, locale);
        try {
            return bundle.getObject(_resourceBase + ".default");
        } catch (MissingResourceException mre) {
            return _resourceBase + ".default";
        }
    }

    public static String[] getWidgetOrder(Config config) throws ConfigException {
        Config[] children = config.getChildren();
        Vector order = new Vector();
        for (int i = 0; i < children.length; i++) {
            if (children[i].getName().equals(CFG_BASE)) {
                try {
                    String name = children[i].getChild(CFG_NAME).getValue();
                    order.addElement(name);
                } catch (Exception e) {
                    throw new ConfigException("No name specified");
                }
            }
        }
        return (String[]) order.toArray(new String[order.size()]);
    }

    public String getId() {
        return _id;
    }

    public String getName() {
        return _name;
    }

    public String getLabel() {
        return getLabel(Locale.getDefault());
    }

    public String getLabel(Locale locale) {
        ResourceBundle bundle = ResourceBundle.getBundle(_resourceLocation, locale);
        try {
            return bundle.getString(_resourceBase + ".label");
        } catch (MissingResourceException mre) {
            return _resourceBase + ".label";
        }
    }

    public void setId(String id) {
        _id = id;
        _resourceBase = WIDGET_RESOURCE_BASE + id;
    }

    public void setName(String name) {
        _name = name;
    }

    public void setResourceLocation(String resourceLocation) {
        _resourceLocation = resourceLocation;
    }

    public void setRequired(boolean required) {
        if (_required = required) {
            RequiredHandler handler = new RequiredHandler();
            handler.setIds(new String[] { _id });
            if (_rule == null)
                _rule = new Rule(_id);
            _rule.add(handler);
        } else {
            _rule.removeHandler(RequiredHandler.NAME);
        }
    }

    protected Object selectValueToUse(Object value, ServletRequest request) {
        Object requestValue = request.getAttribute(_id);
        if (requestValue == null)
            requestValue = request.getParameter(_id);
        if (requestValue == null)
            return value;
        else
            return requestValue;
    }

    public abstract String toString(Locale locale, Object value, ServletRequest request);
    public abstract String toString(Locale locale, Object value, ServletRequest request, String idSuffix);

    /**
     * Returns the DataHandler rule that may be used to validate the Widget
     */
    public Rule getRule() {
        if (_rule == null) {
            DataHandler handler = new CopyDataHandler();
            handler.setIds(new String[] { _id });
            addDataHandler(handler);
        }
        return _rule;
    }

    /**
     * @return the ids that will be submitted to the server for this widget
     */
    public String[] getSubmitIds() {
        if (_submitIds == null)
            _submitIds = new String[] { _id };
        return _submitIds;
    }

    //anderson - additional attrs
    public void setAdditionalAttrs(String attrs) {
        this._additionalAttrs = attrs;
    }

    //anderson - real name
    public void setRealName(String name) {
        this._realName = name;
    }
}
