/*
 * ComboDateTime.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.view.widget.html;

import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Locale;

import javax.servlet.ServletRequest;

import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.datahandler.handler.DateTimeHandler;
import com.eXtropia.view.widget.Widget;

/**
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 1.0
 */
public class ComboDateTime extends Widget {
    public static final String CFG_ALLOW_NULL = "AllowNull";
    public static final String CFG_TYPE = "Type";

    public static final String CFG_MONTH_FORMAT = "MonthFormat";
    public static final String CFG_LOW_YEAR_BOUND = "LowYearBound";
    public static final String CFG_HIGH_YEAR_BOUND = "HighYearBound";
    public static final String CFG_RELATIVE_LOW_YEAR_BOUND = "RelativeLowYearBound";
    public static final String CFG_RELATIVE_HIGH_YEAR_BOUND = "RelativeHighYearBound";
    public static final String CFG_MINUTE_INTERVAL = "MinuteInterval";

    public static final String CFG_TYPE_DATE = "date";
    public static final String CFG_TYPE_TIME = "time";

    protected static final NumberFormat NUMBER_FORMAT = new DecimalFormat("00");
    protected static final NumberFormat YEAR_NUMBER_FORMAT = new DecimalFormat("0000");

    private static final int TYPE_DATE = 1;
    private static final int TYPE_TIME = 2;
    private static final int TYPE_DATE_TIME = 0;

    protected boolean _allowNull = false;

    private int _type = TYPE_DATE_TIME;

    protected int _minuteInterval = 10;

    protected int _lowYearBound;
    protected int _highYearBound;

    private boolean highRelative;
    private boolean lowRelative;

    protected String _monthFormat;

    /** Creates new ComboDateTime */
    public ComboDateTime() {
        setMonthFormat("MMMMM");
        Calendar cal = Calendar.getInstance();
        setLowYearBound(cal.get(Calendar.YEAR));
        setHighYearBound(cal.get(Calendar.YEAR) + 5);
    }

    public void configure(Config config) throws ConfigException {
        super.configure(config);

        if (config.hasChild(CFG_TYPE)) {
            String type = config.getChild(CFG_TYPE).getValue().toLowerCase();
            if (type.equals(CFG_TYPE_DATE))
                _type = TYPE_DATE;
            else if (type.equals(CFG_TYPE_TIME))
                _type = TYPE_TIME;
            else
                _type = TYPE_DATE_TIME;
        }
        if (_type == TYPE_DATE_TIME) {
            DateTimeHandler handler = new DateTimeHandler(DateTimeHandler.TYPE_DATE_TIME);
            handler.setTransformId(_id);
            String[] ids = new String[5];
            ids[DateTimeHandler.YEAR_INDEX] = _id + "_y";
            ids[DateTimeHandler.MONTH_INDEX] = _id + "_M";
            ids[DateTimeHandler.DATE_INDEX] = _id + "_d";
            ids[DateTimeHandler.HOUR_INDEX] = _id + "_h";
            ids[DateTimeHandler.MINUTE_INDEX] = _id + "_m";
            handler.setIds(ids);
            addDataHandler(handler);
        } else if (_type == TYPE_DATE) {
            DateTimeHandler handler = new DateTimeHandler(DateTimeHandler.TYPE_DATE);
            handler.setTransformId(_id);
            String[] ids = new String[3];
            ids[DateTimeHandler.YEAR_INDEX] = _id + "_y";
            ids[DateTimeHandler.MONTH_INDEX] = _id + "_M";
            ids[DateTimeHandler.DATE_INDEX] = _id + "_d";
            handler.setIds(ids);
            addDataHandler(handler);
        }

        if (config.hasChild(CFG_LOW_YEAR_BOUND)) {
            try {
                setLowYearBound(Integer.parseInt(config.getChild(CFG_LOW_YEAR_BOUND).getValue()));
            } catch (NumberFormatException nfe) {
                throw new ConfigException(CFG_LOW_YEAR_BOUND + " is not valid: " + config.getChild(CFG_LOW_YEAR_BOUND).getValue());
            }
        }

        if (config.hasChild(CFG_HIGH_YEAR_BOUND)) {
            try {
                setHighYearBound(Integer.parseInt(config.getChild(CFG_HIGH_YEAR_BOUND).getValue()));
            } catch (NumberFormatException nfe) {
                throw new ConfigException(CFG_HIGH_YEAR_BOUND + " is not valid: " + config.getChild(CFG_HIGH_YEAR_BOUND).getValue());
            }
        }

        Calendar cal = Calendar.getInstance();
        if (config.hasChild(CFG_RELATIVE_LOW_YEAR_BOUND)) {
            try {
                setLowYearBound(cal.get(Calendar.YEAR) + Integer.parseInt(config.getChild(CFG_RELATIVE_LOW_YEAR_BOUND).getValue()));
                lowRelative = true;
            } catch (NumberFormatException nfe) {
                throw new ConfigException(CFG_LOW_YEAR_BOUND + " is not valid: " + config.getChild(CFG_LOW_YEAR_BOUND).getValue());
            }
        }

        if (config.hasChild(CFG_RELATIVE_HIGH_YEAR_BOUND)) {
            try {
                setHighYearBound(cal.get(Calendar.YEAR) + Integer.parseInt(config.getChild(CFG_RELATIVE_HIGH_YEAR_BOUND).getValue()));
                highRelative = true;
            } catch (NumberFormatException nfe) {
                throw new ConfigException(CFG_HIGH_YEAR_BOUND + " is not valid: " + config.getChild(CFG_HIGH_YEAR_BOUND).getValue());
            }
        }

        if (config.hasChild(CFG_MONTH_FORMAT))
            setMonthFormat(config.getChild(CFG_MONTH_FORMAT).getValue());

        if (config.hasChild(CFG_ALLOW_NULL))
            setAllowNull(Boolean.valueOf(config.getChild(CFG_ALLOW_NULL).getValue()).booleanValue());

        if (config.hasChild(CFG_MINUTE_INTERVAL))
            setMinuteInterval(Integer.parseInt(config.getChild(CFG_MINUTE_INTERVAL).getValue()));
    }

    public void setAllowNull(boolean allowNone) {
        _allowNull = allowNone;
    }

    public void setHighYearBound(int highYearBound) {
        _highYearBound = highYearBound;
    }

    public void setLowYearBound(int lowYearBound) {
        _lowYearBound = lowYearBound;
    }

    public void setMinuteInterval(int minutes) {
        _minuteInterval = minutes;
    }

    public void setMonthFormat(String format) {
        _monthFormat = format;
    }

    public String toString(Locale locale, Object value, ServletRequest request) {
        return toString(locale, value, request, null);
    }

    public String toString(Locale locale, Object value, ServletRequest request, String idSuffix) {
        Date dateObj = null;
        value = selectValueToUse(value, request);
        if (value == null) {
            if (!_allowNull) {
                dateObj = new Date();
            }
        } else
            dateObj = (Date) value;

        int hour = -1;
        int minute = -1;
        int date = -1;
        int month = -1;
        int year = -1;
        try {
            date = Integer.parseInt(request.getParameter(_id + "_d"));
        } catch (Exception e) {
            date = -1;
        }

        try {
            month = Integer.parseInt(request.getParameter(_id + "_M")) - 1;
        } catch (Exception e) {
            month = -1;
        }

        try {
            year = Integer.parseInt(request.getParameter(_id + "_y"));
        } catch (Exception e) {
            year = -1;
        }

        try {
            hour = Integer.parseInt(request.getParameter(_id + "_h"));
        } catch (Exception e) {
            hour = -1;
        }

        try {
            minute = Integer.parseInt(request.getParameter(_id + "_m"));
        } catch (Exception e) {
            minute = -1;
        }

        StringBuffer buffer = new StringBuffer(1000);

        if (_type == TYPE_DATE || _type == TYPE_DATE_TIME) {
            if (date == -1 && month == -1 && year == -1 && dateObj != null) {
                Calendar cal = Calendar.getInstance();
                cal.setTime(dateObj);
                date = cal.get(Calendar.DATE);
                month = cal.get(Calendar.MONTH);
                year = cal.get(Calendar.YEAR);
            }
            int lowBound = _lowYearBound, highBound = _highYearBound;
            if (year > -1) {
                lowBound = year < _lowYearBound ? year : _lowYearBound;
                highBound = year > _highYearBound ? year : _highYearBound;
            }
            if (lowRelative) {
                lowBound = _lowYearBound;
            }
            if (highRelative) {
                highBound = _highYearBound;
                if (year > highBound) {
                    //so that the default value would be the highBound
                    year = highBound;
                }
            }
            if (idSuffix == null) {
                generateIntSelect(_id + "_d", NUMBER_FORMAT, date, 1, 31, 1, buffer);
                generateMonthSelect(_id + "_M", month, locale, buffer);
                generateIntSelect(_id + "_y", YEAR_NUMBER_FORMAT, year, lowBound, highBound, 1, buffer);
            } else {
                generateIntSelect(_id + "_d" + idSuffix, NUMBER_FORMAT, date, 1, 31, 1, buffer);
                generateMonthSelect(_id + "_M" + idSuffix, month, locale, buffer);
                generateIntSelect(_id + "_y" + idSuffix, YEAR_NUMBER_FORMAT, year, lowBound, highBound, 1, buffer);
            }
        }

        if (_type == TYPE_DATE_TIME) {
            buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;");
        }

        if (_type == TYPE_TIME || _type == TYPE_DATE_TIME) {
            if (hour == -1 && minute == -1 && dateObj != null) {
                Calendar cal = Calendar.getInstance(locale);
                cal.setTime(dateObj);
                hour = cal.get(Calendar.HOUR_OF_DAY);
                minute = cal.get(Calendar.MINUTE);
            }
            if (idSuffix == null) {
                generateIntSelect(_id + "_h", NUMBER_FORMAT, hour, 0, 23, 1, buffer);
                generateIntSelect(_id + "_m", NUMBER_FORMAT, minute, 0, 59, _minuteInterval, buffer);
            } else {
                generateIntSelect(_id + "_h" + idSuffix, NUMBER_FORMAT, hour, 0, 23, 1, buffer);
                generateIntSelect(_id + "_m" + idSuffix, NUMBER_FORMAT, minute, 0, 59, _minuteInterval, buffer);
            }
        }

        return buffer.toString();
    }

    protected void generateIntSelect(String name, NumberFormat format, int selected, int lowBound, int highBound, int step, StringBuffer buffer) {
        generateSelectHeader(name, buffer);
        if (_allowNull)
            buffer.append("<OPTION>");
        for (int i = lowBound; i <= highBound; i += step) {
            buffer.append("<OPTION value=\"");
            buffer.append(i);
            buffer.append('\"');
            if (i == selected) {
                buffer.append(" SELECTED");
            }

            buffer.append('>');
            buffer.append(format.format(i));
        }

        generateSelectFooter(buffer);
    }

    protected void generateMonthSelect(String name, int selected, Locale locale, StringBuffer buffer) {
        generateSelectHeader(name, buffer);
        Calendar cal = Calendar.getInstance(locale);
        DateFormat monthFormat = new SimpleDateFormat(_monthFormat, locale);
        if (_allowNull)
            buffer.append("<OPTION>");
        for (int i = 0; i < 12; i++) {
            buffer.append("<OPTION value=\"");
            buffer.append(i + 1);
            buffer.append('\"');
            if (i == selected) {
                buffer.append(" SELECTED");
            }

            buffer.append('>');
            cal.set(Calendar.MONTH, i);
            cal.set(Calendar.DAY_OF_MONTH, 1); //sf:+ reset day to 1st, otherwise you'll get double-MARCH when day>=29
            buffer.append(monthFormat.format(cal.getTime()));
        }

        generateSelectFooter(buffer);
    }

    protected void generateSelectHeader(String name, StringBuffer buffer) {
        buffer.append("<SELECT name=\"");
        buffer.append(name);
        buffer.append("\"");
        if (_additionalAttrs != null)
            buffer.append(" " + _additionalAttrs);
        buffer.append(">");
    }

    protected void generateSelectFooter(StringBuffer buffer) {
        buffer.append("</SELECT>");
    }

    public String[] getSubmitIds() {
        if (_submitIds == null) {
            if (_type == TYPE_DATE_TIME)
                _submitIds = new String[] { _id + "_d", _id + "_M", _id + "_y", _id + "_h", _id + "_m" };
            else if (_type == TYPE_DATE)
                _submitIds = new String[] { _id + "_d", _id + "_M", _id + "_y" };
            else
                _submitIds = new String[] { _id + "_h", _id + "_m" };
        }
        return _submitIds;
    }
}
