/* 
 * Task.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.webcal;

import java.util.Date;

import com.eXtropia.auth.User;

/** 
 * A <CODE>Task</CODE> object represents something that a user of the WebCal must do.
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 1.0
 */
public class Task {
    protected User _owner;

    protected Object _id;
    protected String _subject;
    protected String _description;
    protected Date _startDate;
    protected Date _endDate;
    protected int _status;
    protected int _priority;

    protected User _lastModifiedBy;
    protected Date _lastModified;

    public Task(User owner, String subject, Date startDate) {
        this(null, owner, subject, startDate);
    }

    /**
     * Constructs a new <CODE>Task</CODE> object with the specified <CODE>subject</CODE> and <CODE>startDate</CODE>.
     * <P>
     * Defaults:
     * <UL>
     *    <LI> {@link #_priority} to {@link Task.Priority#NORMAL} 
     *    <LI> {@link #_status} to {@link Task.Status#NOT_STARTED}
     *    <LI> {@link #_lastModifiedBy} to {@link #_owner}
     *    <LI> {@link #_lastModified} to the current system time
     * </UL>
     * @param subject The subject of the <CODE>Task</CODE>
     * @param startDate The Start Date of the <CODE>Task</CODE>
     * @param owner The owner of the <CODE>Task</CODE>
     */
    public Task(Object id, User owner, String subject, Date startDate) {
        _id = id;
        _subject = subject;
        _startDate = startDate;
        _owner = owner;

        _lastModifiedBy = owner;
        _lastModified = new Date();
        _status = Task.Status.NOT_STARTED;
        _priority = Task.Priority.NORMAL;
    }

    /**
     * @return the {@link #_description} of the <CODE>Task</CODE>
    */
    public String getDescription() {
        return _description;
    }

    /**
     * @return the {@link #_endDate} of the <CODE>Task</CODE>
     */
    public Date getEndDate() {
        return _endDate;
    }

    /**
     * @return the unique identifer of this <CODE>Task</CODE>
     */
    public Object getId() {
        return _id;
    }

    /**
     * @return the date that the <CODE>Task</CODE> was last modified
     */
    public Date getLastModified() {
        return _lastModified;
    }

    /**
     * @return the <CODE>User</CODE> that last modified the <CODE>Task</CODE>
     */
    public User getLastModifiedBy() {
        return _lastModifiedBy;
    }

    /**
     * @return the {@link #_owner} of the <CODE>Task</CODE>
     */
    public User getOwner() {
        return _owner;
    }

    /**
    	 * @return the {@link #_priority} of the <CODE>Task</CODE>
     */
    public int getPriority() {
        return _priority;
    }

    /**
    	 * @return the {@link #_startDate} of the <CODE>Task</CODE>
     */
    public Date getStartDate() {
        return _startDate;
    }

    /**
     * @return the {@link #_status} of the <CODE>Task</CODE>
     */
    public int getStatus() {
        return _status;
    }

    /**
     * @return the {@link #_subject} of the <CODE>Task</CODE>
     */
    public String getSubject() {
        return _subject;
    }

    /**
     * Sets the {@link #_description} of the <CODE>Task</CODE>
     * @param description The new Description of the <CODE>Task</CODE>
     */
    public void setDescription(String description) {
        _description = description;
    }

    /**
     * Sets the {@link #_endDate} of the <CODE>Task</CODE>
     * @param endDate The new End Date of the <CODE>Task</CODE>
     */
    public void setEndDate(Date endDate) {
        _endDate = endDate;
    }

    /**
     * Sets the {@link #_lastModified} of the <CODE>Task</CODE>
     * @param lastModified The last time the <CODE>Task</CODE> was modified <CODE>Task</CODE>
     */
    public void setLastModified(Date lastModified) {
        _lastModified = lastModified;
    }

    /**
     * Sets the {@link #_lastModified} of the <CODE>Task</CODE>
     * @param lastModified The last time the <CODE>Task</CODE> was modified <CODE>Task</CODE>
     */
    public void setLastModifiedBy(User lastModifiedBy) {
        _lastModifiedBy = lastModifiedBy;
    }

    /**
     * Sets the {@link #_priority} of the <CODE>Task</CODE>
     * @param priority The new {@link Task.Priority} of the <CODE>Task</CODE>
     */
    public void setPriority(int priority) {
        _priority = priority;
    }

    /**
     * Sets the {@link #_startDate} of the <CODE>Task</CODE>
     * @param startDate The new Start Date of the <CODE>Task</CODE>
     */
    public void setStartDate(Date startDate) {
        _startDate = startDate;
    }

    /**
     * Sets the {@link #_status} of the <CODE>Task</CODE>
     * @param status The new {@link Task.Status} of the <CODE>Task</CODE>
     */
    public void setStatus(int status) {
        _status = status;
    }

    /**
     * Sets the {@link #_subject} of the <CODE>Task</CODE>
    * @param subject The new Subject of the <CODE>Task</CODE>
     */
    public void setSubject(String subject) {
        _subject = subject;
    }

    /**
     * An interface specifying the priorities a {@link Task} may have.
     *
     * @version 1.0
     */
    public interface Priority {
        /**
         * A <CODE>LOW</CODE> priority {@link Task}
         */
        public static final int LOW = 0;

        /**
         * The <CODE>NORMAL</CODE> priority of a {@link Task}
         */
        public static final int NORMAL = 1;

        /**
         * A <CODE>HIGH</CODE> priority {@link Task}
         */
        public static final int HIGH = 2;
    }

    /**
     * An interface specifying the various status' a {@link Task} may be in.
     *
     * @version 1.0
     */
    public interface Status {
        /**
         * The <CODE>Status</CODE> of a {@link Task} yet to be started.
         */
        public static final int NOT_STARTED = 0;

        /**
         * The <CODE>Status</CODE> of a {@link Task} yet to be completed.
         */
        public static final int IN_PR0GRESS = 1;

        /**
         * The <CODE>Status</CODE> of a <CODE>COMPLETE</CODE> {@link Task}.
         */
        public static final int COMPLETED = 2;
    }
}