/* 
 * TaskDataSource.java
 * Copyright (C) 2000 - 2001  eXtropia.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.eXtropia.webcal.datasource;

import java.sql.SQLException;
import java.util.Date;

import javax.sql.RowSet;

import com.eXtropia.auth.AuthManager;
import com.eXtropia.auth.User;
import com.eXtropia.config.Config;
import com.eXtropia.config.ConfigException;
import com.eXtropia.config.Configurable;
import com.eXtropia.datasource.Condition;
import com.eXtropia.datasource.Criterion;
import com.eXtropia.datasource.DataSource;
import com.eXtropia.datasource.FieldOrder;
import com.eXtropia.datasource.Query;
import com.eXtropia.diagnostics.Diagnostics;
import com.eXtropia.diagnostics.Doctor;
import com.eXtropia.diagnostics.LogEvent;
import com.eXtropia.diagnostics.TraceEvent;
import com.eXtropia.util.ExtropiaException;
import com.eXtropia.webcal.Holiday;
import com.eXtropia.webcal.Task;

/** 
 *
 * @author  Tim Parker (tim.parker@eXtropia.com)
 * @version 1.0
 */
public class TaskDataSource implements Configurable {
    public static final String CFG_BASE = "TaskDataSource";
    public static final String CFG_GLOBAL_UID = "GlobalUid";

    private AuthManager _authManager;
    private DataSource _dataSource;
    private Doctor _doctor;

    public TaskDataSource(String appName, String diagBase, AuthManager authManager) {
        _doctor = Diagnostics.getDoctor(new String[] { diagBase, appName, "WebCal", "TaskDataSource" });
        _authManager = authManager;
    }

    public void configure(Config config) throws ConfigException {
        if (!config.getName().equals(CFG_BASE))
            throw new ConfigException("Invalid Config passed to " + CFG_BASE + ": " + config.getName());

        try {
            _doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Initialising DataSource"));
            if (config.hasChild(DataSource.CFG_BASE))
                _dataSource = (DataSource) config.getChild(DataSource.CFG_BASE).createAndConfigureInstance();
            else
                throw new ConfigException("No " + Diagnostics.CFG_BASE + " specified in Config");
            _doctor.trace(new TraceEvent(this, TraceEvent.INFO, "Finished Initialising DataSource"));
        } catch (ClassCastException cce) {
            throw new ConfigException("" + config.getChild(DataSource.CFG_BASE) + "is not of type DataSource" + cce.getMessage());
        } catch (ConfigException e) {
            throw new ConfigException("Failed to create datasource", e);
        }
    }

    public void delete(Object id) throws ExtropiaException {
        try {
            _dataSource.delete(new Condition(DataSourceColumns.ID, Condition.EQUAL, id));
        } catch (ExtropiaException ee) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, "Failed to delete task: " + ee.getMessage()));
            throw ee;
        }
    }

    public Task[] load(User[] users, int types, FieldOrder order) throws ExtropiaException {
        return load(users, types, null, order);
    }

    public Task[] load(User[] users, String keyword) throws ExtropiaException {
        return load(users, DataSourceColumns.TYPE_ALL, keyword, null);
    }

    public Task[] load(User[] users, int types, String keyword, FieldOrder order) throws ExtropiaException {
        Criterion searchCriterion = new Criterion(Criterion.AND);

        Criterion usernameCriterion = generateUserCriterion(users);
        searchCriterion.add(usernameCriterion);

        if (keyword != null) {
            Criterion keywordCriterion = _dataSource.generateCriterion("*" + keyword + "*");
            searchCriterion.add(keywordCriterion);
        }

        if (types != DataSourceColumns.TYPE_ALL) {
            Criterion typesCriterion = new Criterion(Criterion.OR);
            if ((types & DataSourceColumns.TYPE_TASK) == DataSourceColumns.TYPE_TASK)
                typesCriterion.add(new Condition(DataSourceColumns.TYPE, Condition.EQUAL, new Integer(DataSourceColumns.TYPE_TASK)));
            if ((types & DataSourceColumns.TYPE_EVENT) == DataSourceColumns.TYPE_EVENT)
                typesCriterion.add(new Condition(DataSourceColumns.TYPE, Condition.EQUAL, new Integer(DataSourceColumns.TYPE_EVENT)));
            if ((types & DataSourceColumns.TYPE_APPOINTMENT) == DataSourceColumns.TYPE_APPOINTMENT)
                typesCriterion.add(new Condition(DataSourceColumns.TYPE, Condition.EQUAL, new Integer(DataSourceColumns.TYPE_APPOINTMENT)));

            //DONT SEARCH OVER HOLIDAYS BY DEFAULT
            //if((types & DataSourceColumns.TYPE_HOLIDAY) == DataSourceColumns.TYPE_HOLIDAY)
            //typesCriterion.add(new Condition(DataSourceColumns.TYPE, Condition.EQUAL, new Integer(DataSourceColumns.TYPE_HOLIDAY)));
            searchCriterion.add(typesCriterion);
        }

        Query q = new Query(searchCriterion);
        if (order != null)
            q.setFieldOrder(new FieldOrder[] { order });

        RowSet rowSet = _dataSource.search(q);

        try {
            Task[] tasks = TaskFactory.getTasks(rowSet, _authManager);
            rowSet.close();
            return tasks;
        } catch (ExtropiaException ee) {
            if (_doctor.doTrace(TraceEvent.INFO))
                _doctor.trace(new TraceEvent(this, TraceEvent.WARN, ee.getMessage()));
            throw ee;
        } catch (SQLException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, "Error closing rowset: " + se.getMessage()));
            throw new ExtropiaException("Error closing RowSet: " + se.getMessage());
        }
    }

    public Task[] load(User[] users, Date startDate, Date endDate) throws ExtropiaException {
        Criterion noRecurrenceRangeCriterion = new Criterion(Criterion.AND);
        noRecurrenceRangeCriterion.add(new Condition(DataSourceColumns.TYPE, Condition.NOT_EQUAL, new Integer(DataSourceColumns.TYPE_TASK)));
        noRecurrenceRangeCriterion.add(new Condition(DataSourceColumns.RECURRENCE, Condition.EQUAL, new Integer(DataSourceColumns.RECURRENCE_NONE)));
        noRecurrenceRangeCriterion.add(new Condition(DataSourceColumns.END_DATE, Condition.GREATER_OR_EQUAL, startDate));
        noRecurrenceRangeCriterion.add(new Condition(DataSourceColumns.START_DATE, Condition.LESS_OR_EQUAL, endDate));

        Criterion recurrenceEndDateCriterion = new Criterion(Criterion.OR);
        recurrenceEndDateCriterion.add(new Condition(DataSourceColumns.RECURRENCE_UNTIL_DATE, Condition.GREATER_OR_EQUAL, startDate));
        recurrenceEndDateCriterion.add(new Condition(DataSourceColumns.RECURRENCE_UNTIL_DATE, Condition.EQUAL, null));

        Criterion recurrenceRangeCriterion = new Criterion(Criterion.AND);
        recurrenceRangeCriterion.add(new Condition(DataSourceColumns.TYPE, Condition.NOT_EQUAL, new Integer(DataSourceColumns.TYPE_TASK)));
        recurrenceRangeCriterion.add(new Condition(DataSourceColumns.RECURRENCE, Condition.NOT_EQUAL, new Integer(DataSourceColumns.RECURRENCE_NONE)));
        recurrenceRangeCriterion.add(new Condition(DataSourceColumns.START_DATE, Condition.LESS_OR_EQUAL, endDate));
        recurrenceRangeCriterion.add(recurrenceEndDateCriterion);

        Criterion rangeCriterion = new Criterion(Criterion.OR);
        rangeCriterion.add(recurrenceRangeCriterion);
        rangeCriterion.add(noRecurrenceRangeCriterion);

        Criterion taskCriterion = new Criterion(Criterion.AND);
        taskCriterion.add(new Condition(DataSourceColumns.TYPE, Condition.EQUAL, new Integer(DataSourceColumns.TYPE_TASK)));
        taskCriterion.add(new Condition(DataSourceColumns.STATUS, Condition.NOT_EQUAL, new Integer(Task.Status.COMPLETED)));
        taskCriterion.add(new Condition(DataSourceColumns.START_DATE, Condition.LESS_THAN, endDate));

        Criterion usernameCriterion = generateUserCriterion(users);

        Criterion searchCriterion = new Criterion(Criterion.OR);
        searchCriterion.add(rangeCriterion);
        searchCriterion.add(taskCriterion);

        Criterion overallCriterion = new Criterion(Criterion.AND);
        overallCriterion.add(searchCriterion);
        overallCriterion.add(usernameCriterion);

        Query q = new Query(overallCriterion);
        q.setFieldOrder(new FieldOrder[] { new FieldOrder(DataSourceColumns.SUBJECT)});
        RowSet rowSet = _dataSource.search(q);

        try {
            Task[] tasks = TaskFactory.getTasks(rowSet, _authManager);
            rowSet.close();
            return tasks;
        } catch (ExtropiaException ee) {
            if (_doctor.doTrace(TraceEvent.INFO))
                _doctor.trace(new TraceEvent(this, TraceEvent.WARN, ee.getMessage()));
            throw ee;
        } catch (SQLException se) {
            _doctor.log(new LogEvent(this, LogEvent.WARN, "Error closing rowset: " + se.getMessage()));
            throw new ExtropiaException("Error closing RowSet: " + se.getMessage());
        }
    }

    public Task loadById(Object id) throws ExtropiaException {
        Criterion idCriterion = new Criterion(Criterion.AND);
        idCriterion.add(new Condition(DataSourceColumns.ID, Condition.EQUAL, id));
        RowSet rowSet = _dataSource.search(new Query(idCriterion));
        try {
            return TaskFactory.getTask(rowSet, _authManager);
        } catch (ExtropiaException ee) {
            if (_doctor.doTrace(TraceEvent.WARN))
                _doctor.trace(new TraceEvent(this, TraceEvent.WARN, ee.getMessage()));
            throw ee;
        }
    }

    public Holiday[] loadHolidays(Date startDate, Date endDate) throws ExtropiaException {
        Criterion noRecurrenceRangeCriterion = new Criterion(Criterion.AND);
        noRecurrenceRangeCriterion.add(new Condition(DataSourceColumns.TYPE, Condition.EQUAL, new Integer(DataSourceColumns.TYPE_HOLIDAY)));
        noRecurrenceRangeCriterion.add(new Condition(DataSourceColumns.RECURRENCE, Condition.EQUAL, new Integer(DataSourceColumns.RECURRENCE_NONE)));
        noRecurrenceRangeCriterion.add(new Condition(DataSourceColumns.END_DATE, Condition.GREATER_OR_EQUAL, startDate));
        noRecurrenceRangeCriterion.add(new Condition(DataSourceColumns.START_DATE, Condition.LESS_OR_EQUAL, endDate));

        Criterion recurrenceEndDateCriterion = new Criterion(Criterion.OR);
        recurrenceEndDateCriterion.add(new Condition(DataSourceColumns.RECURRENCE_UNTIL_DATE, Condition.GREATER_OR_EQUAL, startDate));
        recurrenceEndDateCriterion.add(new Condition(DataSourceColumns.RECURRENCE_UNTIL_DATE, Condition.EQUAL, null));

        Criterion recurrenceRangeCriterion = new Criterion(Criterion.AND);
        recurrenceRangeCriterion.add(new Condition(DataSourceColumns.TYPE, Condition.EQUAL, new Integer(DataSourceColumns.TYPE_HOLIDAY)));
        recurrenceRangeCriterion.add(new Condition(DataSourceColumns.RECURRENCE, Condition.NOT_EQUAL, new Integer(DataSourceColumns.RECURRENCE_NONE)));
        recurrenceRangeCriterion.add(new Condition(DataSourceColumns.START_DATE, Condition.LESS_OR_EQUAL, endDate));
        recurrenceRangeCriterion.add(recurrenceEndDateCriterion);

        Criterion holidayCriterion = new Criterion(Criterion.OR);
        holidayCriterion.add(recurrenceRangeCriterion);
        holidayCriterion.add(noRecurrenceRangeCriterion);

        RowSet rowSet = _dataSource.search(new Query(holidayCriterion));
        Task[] tasks = TaskFactory.getTasks(rowSet, _authManager);
        Holiday[] holidays = new Holiday[tasks.length];
        System.arraycopy(tasks, 0, holidays, 0, tasks.length);
        return holidays;
    }

    public Task[] loadTasks(User[] users, FieldOrder order) throws ExtropiaException {
        return load(users, DataSourceColumns.TYPE_TASK, order);
    }

    public void saveTask(Task task) throws ExtropiaException {
        //save all content in this task
        if (task.getId() != null) {
            Criterion updateCriterion = new Criterion(Criterion.AND);
            updateCriterion.add(new Condition(DataSourceColumns.ID, Condition.EQUAL, task.getId()));
            try {
                _dataSource.update(TaskFactory.generateUpdateHash(task), updateCriterion);
            } catch (ExtropiaException ee) {
                _doctor.log(new LogEvent(this, LogEvent.WARN, "Failed to update task: " + ee.getMessage()));
                throw ee;
            }
        } else {
            try {
                _dataSource.add(TaskFactory.generateUpdateHash(task));
            } catch (ExtropiaException ee) {
                _doctor.log(new LogEvent(this, LogEvent.WARN, "Failed to add task: " + ee.getMessage()));
                throw ee;
            }
        }
    }

    protected Criterion generateUserCriterion(User[] users) {
        Criterion c = new Criterion(Criterion.OR);
        for (int i = 0; i < users.length; i++) {
            c.add(new Condition(DataSourceColumns.OWNER, Condition.EQUAL, users[i].getId()));
        }
        return c;
    }
}